/*
 * Copyright 2005 Takao Sumitomo. All rights reserved.
 * 
 * Redistribution and use in source and binary forms,
 * with or without modification, are permitted provided 
 * that the following conditions are met:
 *
 *   1. Redistributions of source code must retain the 
 *      above copyright notice, this list of conditions
 *      and the following disclaimer.
 *   2. Redistributions in binary form must reproduce 
 *      the above copyright notice, this list of conditions
 *      and the following disclaimer in the documentation
 *      and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software
 * and documentation are those of the authors and should not be
 * interpreted as representing official policies,
 * either expressed or implied.
 * 
 * @Author: Takao Sumitomo
 */

/*
 * 作成日: 2005/01/14
 */
package bakusen.common.math;

/**
 * @author cattaka
 */
public class BMath
{
	private static final float max_distance = (float)Long.MAX_VALUE / 1000f;

	public static float getIntersection(BVector4f p, BVector2f fx, BVector2f fy)
	{
		float a = p.a * fx.x + p.b * fy.x + p.c;
		float b = p.a * fx.y + p.b * fy.y + p.d;
		if (a == 0)
			return Float.NaN;
		return -b/a;
	}
	public static float getClosestIntersection(BVector2f start, BVector2f end, BVector2f pos, BPolygon pg)
	{
		BVector2f[][] vss = pg.getVertices();
		float d = Float.MAX_VALUE;
		for (int i=0;i<vss.length;i++)
		{
			float t = getClosestIntersection(start,end,pos,vss[i]);
			if (!(Float.isNaN(t)) && t<d)
			{
				d=t;
			}
		}
		if (d > max_distance)
			return Float.NaN;
		else
			return d;
	}
	public static float getClosestIntersection(BVector2f start, BVector2f end, BVector2f pos, BVector2f[] vs)
	{
		BVector2f pos1 = new BVector2f();
		BVector2f pos2 = new BVector2f();
		BVector2f v = new BVector2f();
		pos1.x = pos.x + vs[vs.length-1].x;
		pos1.y = pos.y + vs[vs.length-1].y;
		float distance = Float.MAX_VALUE;
		for (int i=0;i<vs.length;i++)
		{
			pos2.x = pos.x + vs[i].x;
			pos2.y = pos.y + vs[i].y;
			if(getIntersection(v,start,end,pos1,pos2))
			{
				float d = (v.x-start.x)*(v.x-start.x)+(v.y-start.y)*(v.y-start.y);
				if (d<distance)
				{
					distance = d;
				}
			}
			pos1.set(pos2);
		}
		if (distance > max_distance)
		{
			return Float.NaN;
		}
		else
		{
			return (float)Math.sqrt(distance);
		}
	}
	public static boolean getIntersection(BVector2f dist, BVector2f p11, BVector2f p12, BVector2f p21, BVector2f p22)
	{
		float a1 = -(p12.y - p11.y);
		float a2 = -(p22.y - p21.y);
		float b1 = p12.x - p11.x;
		float b2 = p22.x - p21.x;
		float c1 = p11.x*p12.y - p12.x*p11.y;
		float c2 = p21.x*p22.y - p22.x*p21.y;
		float c = a1*b2 - a2*b1;
		if (c == 0)
		{
			return false;
		}else
		{
			BVector2f v = new BVector2f();
			v.x = (b1*c2 - b2*c1)/c;
			v.y = (c1*a2 - c2*a1)/c;
			if (isIntersect(v,p11,p12) && isIntersect(v,p21,p22))
			{
				dist.set(v);
				return true;
			}else
			{
				return false;
			}
		}
	}
	private static boolean isIntersect(BVector2f a, BVector2f b1, BVector2f b2)
	{
		//if ((p1.x < c.x && p2.x < c.x) || ( p1.x > c.x && p2.x > c.x)
		//	|| (p1.y < c.y && p2.y < c.y) || ( p1.y > c.y && p2.y > c.y))
		if (((a.x-b1.x)*(b2.x-b1.x) + (a.y-b1.y)*(b2.y-b1.y) >= 0)
				&& ((a.x-b2.x)*(b1.x-b2.x) + (a.y-b2.y)*(b1.y-b2.y) >= 0))

		{
			return true;
		}else
		{
			return false;
		}
	}
	private static boolean isIntersect(BVector3f a, BVector3f b1, BVector3f b2)
	{
		// exclude point intersect
		if (((a.x-b1.x)*(b2.x-b1.x) + (a.y-b1.y)*(b2.y-b1.y) + (a.z-b1.z)*(b2.z-b1.z) >= 0)
			&& ((a.x-b2.x)*(b1.x-b2.x) + (a.y-b2.y)*(b1.y-b2.y) + (a.z-b2.z)*(b1.z-b2.z) >= 0))
		{
			return true;
		}else
		{
			return false;
		}
		/*
		if ((b1.x < a.x && b2.x < a.x) || (a.x < b1.x && a.x < b2.x)
			|| (b1.y < a.y && b2.y < a.y) || (a.y < b1.y && a.y < b2.y)
			|| (b1.z < a.z && b2.z < a.z) || (a.z < b1.z && a.z < b2.z)
			)
		{
			return false;
		}
		else
		{
			return true;
		}
		*/
		
		
	}
	private static float isIntersect(BVector3f a1, BVector3f a2, BVector3f b1, BVector3f b2)
	{
		// exclude point intersect
		if ((a1.x < a2.x && ((b1.x <= a1.x && b2.x <= a1.x) || (a2.x <= b1.x && a2.x <= b2.x)))
				|| (a2.x < a1.x && ((b1.x <= a2.x && b2.x <= a2.x) || (a1.x <= b1.x && a1.x <= b2.x)))
				|| (a1.y < a2.y && ((b1.y <= a1.y && b2.y <= a1.y) || (a2.y <= b1.y && a2.y <= b2.y)))
				|| (a2.y < a1.y && ((b1.y <= a2.y && b2.y <= a2.y) || (a1.y <= b1.y && a1.y <= b2.y)))
				|| (a1.z < a2.z && ((b1.z <= a1.z && b2.z <= a1.z) || (a2.z <= b1.z && a2.z <= b2.z)))
				|| (a2.z < a1.z && ((b1.z <= a2.z && b2.z <= a2.z) || (a1.z <= b1.z && a1.z <= b2.z)))
				)
		{
			return Float.NaN;
		}
		else
		{
			float a = Math.max(Math.min(a1.z, a2.z),Math.min(b1.z,b2.z));
			float b = Math.min(Math.max(a1.z, a2.z),Math.max(b1.z,b2.z));
			return Math.min(a,b);
		}
		
	}
	public static void createPlane(BVector4f dist, BVector2f p1, BVector2f p2, BVector2f v)
	{
		float x1 = p1.x;// - v.x * t;
		float x2 = p2.x;// - v.x * t;
		float y1 = p1.y;// - v.y * t;
		float y2 = p2.y;// - v.y * t;
		dist.a = y1 - y2;
		dist.b = -(x1 - x2);
		dist.c = -dist.a*v.x - dist.b*v.y;
		dist.d = x1*y2 - x2*y1;
	}
	public static float nextCollidPoint(BVector2f p11, BVector2f p12, BVector2f v1,BVector2f p21, BVector2f p22, BVector2f v2)
	{
		BVector4f pl1 = new BVector4f();
		BVector4f pl2 = new BVector4f();
		BVector2f fx11 = new BVector2f();
		BVector2f fy11 = new BVector2f();
		BVector2f fx12 = new BVector2f();
		BVector2f fy12 = new BVector2f();
		BVector2f fx21 = new BVector2f();
		BVector2f fy21 = new BVector2f();
		BVector2f fx22 = new BVector2f();
		BVector2f fy22 = new BVector2f();
		BVector3f c1 = new BVector3f();
		BVector3f c2 = new BVector3f();
		BVector3f cp = new BVector3f();
		float d = Float.MAX_VALUE;
		createPlane(pl1, p11, p12, v1);
		createPlane(pl2, p21, p22, v2);
		fx11.x = v1.x;
		fx11.y = p11.x;// - v1.x * t;
		fy11.x = v1.y;
		fy11.y = p11.y;// - v1.y * t;
		fx12.x = v1.x;
		fx12.y = p12.x;// - v1.x * t;
		fy12.x = v1.y;
		fy12.y = p12.y;// - v1.y * t;
		fx21.x = v2.x;
		fx21.y = p21.x;// - v2.x * t;
		fy21.x = v2.y;
		fy21.y = p21.y;// - v2.y * t;
		fx22.x = v2.x;
		fx22.y = p22.x;// - v2.x * t;
		fy22.x = v2.y;
		fy22.y = p22.y;// - v2.y * t;

		cp.z = getIntersection(pl2,fx11,fy11);
		if (!(Float.isNaN(cp.z)))
		{
			cp.x = cp.z * fx11.x + fx11.y;
			cp.y = cp.z * fy11.x + fy11.y;
			c1.z = cp.z;
			c1.x = cp.z * fx21.x + fx21.y;
			c1.y = cp.z * fy21.x + fy21.y;
			c2.z = cp.z;
			c2.x = cp.z * fx22.x + fx22.y;
			c2.y = cp.z * fy22.x + fy22.y;
			if (isIntersect(cp,c1,c2))
			{
				d = Math.min(cp.z,d);
			}
		}

		cp.z = getIntersection(pl2,fx12,fy12);
		if (!(Float.isNaN(cp.z)))
		{
			cp.x = cp.z * fx12.x + fx12.y;
			cp.y = cp.z * fy12.x + fy12.y;
			c1.z = cp.z;
			c1.x = cp.z * fx21.x + fx21.y;
			c1.y = cp.z * fy21.x + fy21.y;
			c2.z = cp.z;
			c2.x = cp.z * fx22.x + fx22.y;
			c2.y = cp.z * fy22.x + fy22.y;
			if (isIntersect(cp,c1,c2))
			{
				d = Math.min(cp.z,d);
			}
		}
		
		cp.z = getIntersection(pl1,fx21,fy21);
		if (!(Float.isNaN(cp.z)))
		{
			cp.x = cp.z * fx21.x + fx21.y;
			cp.y = cp.z * fy21.x + fy21.y;
			c1.z = cp.z;
			c1.x = cp.z * fx11.x + fx11.y;
			c1.y = cp.z * fy11.x + fy11.y;
			c2.z = cp.z;
			c2.x = cp.z * fx12.x + fx12.y;
			c2.y = cp.z * fy12.x + fy12.y;
			if (isIntersect(cp,c1,c2))
			{
				d = Math.min(cp.z,d);
			}
		}

		cp.z = getIntersection(pl1,fx22,fy22);
		if (!(Float.isNaN(cp.z)))
		{
			cp.x = cp.z * fx22.x + fx22.y;
			cp.y = cp.z * fy22.x + fy22.y;
			c1.z = cp.z;
			c1.x = cp.z * fx11.x + fx11.y;
			c1.y = cp.z * fy11.x + fy11.y;
			c2.z = cp.z;
			c2.x = cp.z * fx12.x + fx12.y;
			c2.y = cp.z * fy12.x + fy12.y;
			if (isIntersect(cp,c1,c2))
			{
				d = Math.min(cp.z,d);
			}
		}

		if (d > max_distance)
		{
			return Float.NaN;
		}else
		{
			return d;
		}
	}
	/*
	public static float nextCollidPoint(BVector2d p11, BVector2d p12, BVector2d v1,BVector2d p21, BVector2d p22, BVector2d v2)
	{
		BVector4d pl1 = new BVector4d();
		BVector4d pl2 = new BVector4d();
		createPlane(pl1, p11, p12, v1);
		createPlane(pl2, p21, p22, v2);
		BVector2d fx = new BVector2d();
		BVector2d fy = new BVector2d();
		BVector3d[] cp = new BVector3d[]{new BVector3d(),new BVector3d(),new BVector3d(),new BVector3d()};
		fx.x = v1.x;
		fx.y = p11.x;// - v1.x * t;
		fy.x = v1.y;
		fy.y = p11.y;// - v1.y * t;
		cp[0].z = getIntersection(pl2,fx,fy);
		if (Float.isNaN(cp[0].z))
			return Float.NaN;
		cp[0].x = cp[0].z * fx.x + fx.y;
		cp[0].y = cp[0].z * fy.x + fy.y;
		
		fx.x = v1.x;
		fx.y = p12.x;// - v1.x * t;
		fy.x = v1.y;
		fy.y = p12.y;// - v1.y * t;
		cp[1].z = getIntersection(pl2,fx,fy);
		if (Float.isNaN(cp[1].z))
			return Float.NaN;
		cp[1].x = cp[1].z * fx.x + fx.y;
		cp[1].y = cp[1].z * fy.x + fy.y;

		fx.x = v2.x;
		fx.y = p21.x;// - v2.x * t;
		fy.x = v2.y;
		fy.y = p21.y;// - v2.y * t;
		cp[2].z = getIntersection(pl1,fx,fy);
		if (Float.isNaN(cp[2].z))
			return Float.NaN;
		cp[2].x = cp[2].z * fx.x + fx.y;
		cp[2].y = cp[2].z * fy.x + fy.y;
		
		fx.x = v2.x;
		fx.y = p22.x;// - v2.x * t;
		fy.x = v2.y;
		fy.y = p22.y;// - v2.y * t;
		cp[3].z = getIntersection(pl1,fx,fy);
		if (Float.isNaN(cp[3].z))
			return Float.NaN;
		cp[3].x = cp[3].z * fx.x + fx.y;
		cp[3].y = cp[3].z * fy.x + fy.y;
		
		return isIntersect(cp[0],cp[1],cp[2],cp[3]);
	}
	 */
	public static boolean isIntersect(BVector2f p1, BPolygon pg1, BVector2f p2, BPolygon pg2)
	{
		for (int i=0;i<pg1.getVertices().length;i++)
		{
			BVector2f[] vs1 = pg1.getVertices()[i];
			for (int j=0;j<pg2.getVertices().length;j++)
			{
				BVector2f[] vs2 = pg2.getVertices()[j];
				if (isIntersect(p1,vs1,p2,vs2))
				{
					return true;
				}
			}
		}
		return false;
	}
	public static boolean isIntersect(BVector2f p1, BVector2f[] vertices1, BVector2f p2, BVector2f[] vertices2)
	{
		BVector2f pos1 = new BVector2f();
		BVector2f pos2 = new BVector2f();
		pos1.x = p1.x + vertices1[vertices1.length-1].x;
		pos1.y = p1.y + vertices1[vertices1.length-1].y;
		outer:for (int i=0;i<vertices1.length;i++)
		{
			pos2.x = p1.x + vertices1[i].x;
			pos2.y = p1.y + vertices1[i].y;
			float a = -(pos2.y - pos1.y);
			float b = pos2.x - pos1.x;
			float c = pos1.x*pos2.y - pos2.x*pos1.y;
			for (int j=0;j<vertices2.length;j++)
			{
				if (a*(p2.x+vertices2[j].x) + b*(p2.y+vertices2[j].y) + c <= 0)
				{
					pos1.set(pos2);
					continue outer;
				}
			}
			return false;
		}

		pos1.x = p2.x + vertices2[vertices2.length-1].x;
		pos1.y = p2.y + vertices2[vertices2.length-1].y;
		outer:for (int i=0;i<vertices2.length;i++)
		{
			pos2.x = p2.x + vertices2[i].x;
			pos2.y = p2.y + vertices2[i].y;
			float a = -(pos2.y - pos1.y);
			float b = pos2.x - pos1.x;
			float c = pos1.x*pos2.y - pos2.x*pos1.y;
			for (int j=0;j<vertices1.length;j++)
			{
				if (a*(p1.x+vertices1[j].x) + b*(p1.y+vertices1[j].y) + c <= 0)
				{
					pos1.set(pos2);
					continue outer;
				}
			}
			return false;
		}

		return true;
	}
	public static float nextCollidPoint(BVector2f normaldist1,BVector2f normaldist2, BVector2f p1, BPolygon pg1, BVector2f v1, BVector2f p2, BPolygon pg2, BVector2f v2)
	{
		float d = Float.MAX_VALUE;
		BVector2f n1 = new BVector2f();
		BVector2f n2 = new BVector2f();
		for(int i=0;i<pg1.getVertices().length;i++)
		{
			BVector2f[] vs1 = pg1.getVertices()[i];
			for(int j=0;j<pg2.getVertices().length;j++)
			{
				BVector2f[] vs2 = pg2.getVertices()[j];
				float t = nextCollidPoint(n1,n2,p1,vs1,v1,p2,vs2,v2);
				if (!(Float.isNaN(t)))
				{
					if (t<d)
					{
						d = t;
						normaldist1.set(n1);
						normaldist2.set(n2);
					}else if(t == d)
					{
						normaldist1.x += n1.x;
						normaldist1.y += n1.y;

						normaldist2.x += n2.x;
						normaldist2.y += n2.y;
					}
				}
			}
		}
		//if (d == Float.MAX_VALUE)// || d > max_distance)
		if (d > max_distance)
		{
			return Float.NaN;
		}
		else
		{
			normaldist1.normalize();
			normaldist2.normalize();
			return d;
		}
	}
	public static float nextCollidPoint(BVector2f normaldist1,BVector2f normaldist2, BVector2f p1, BVector2f[] vertices1, BVector2f v1, BVector2f p2, BVector2f[] vertices2, BVector2f v2)
	{
		normaldist1.set(0,0);
		normaldist2.set(0,0);
		BVector2f p11 = new BVector2f();
		BVector2f p12 = new BVector2f();
		BVector2f p21 = new BVector2f();
		BVector2f p22 = new BVector2f();
		
		p12.x = p1.x + vertices1[vertices1.length-1].x;
		p12.y = p1.y + vertices1[vertices1.length-1].y;
		p22.x = p2.x + vertices2[vertices2.length-1].x;
		p22.y = p2.y + vertices2[vertices2.length-1].y;
		
		float cp = Float.MAX_VALUE;
		for (int i=0;i<vertices1.length;i++)
		{
			p11.set(p12);
			p12.x = p1.x + vertices1[i].x;
			p12.y = p1.y + vertices1[i].y;

			for (int j=0;j<vertices2.length;j++)
			{
				p21.set(p22);
				p22.x = p2.x + vertices2[j].x;
				p22.y = p2.y + vertices2[j].y;
				float d = nextCollidPoint(p11,p12,v1,p21,p22,v2);
				if (!(Float.isNaN(d)) && d >= 0)
				{
					if (d < cp)
					{
						BVector2f n = new BVector2f();
						n.x = -(p12.y - p11.y);
						n.y = (p12.x - p11.x);
						n.normalize();
						normaldist1.set(n);
						
						n.x = -(p22.y - p21.y);
						n.y = (p22.x - p21.x);
						n.normalize();
						normaldist2.set(n);
						
						cp = d;
					}else if(d == cp)
					{
						BVector2f n = new BVector2f();
						n.x = -(p12.y - p11.y);
						n.y = (p12.x - p11.x);
						n.normalize();
						normaldist1.x += n.x; 
						normaldist1.y += n.y; 
						
						n.x = -(p22.y - p21.y);
						n.y = (p22.x - p21.x);
						n.normalize();
						normaldist2.x += n.x; 
						normaldist2.y += n.y; 
					}
					//cp = Math.min(cp,d);
				}
			}
		}
		if (cp > max_distance)
		{
			return Float.NaN;
		}else
		{
			normaldist1.normalize();
			normaldist2.normalize();
			return cp;
		}
	}
	public static void getClosestPoint(BVector2f dist, BVector2f p1, BVector2f v, BVector2f p2)
	{
		float d = v.x * v.x + v.y * v.y;
		if (d == 0)
		{
			dist.x = Float.NaN;
			dist.y = Float.NaN;
			return;
		}
		d = 1 / d;
		//dist.x = d * (p1.y - p2.y + (v.x/v.y)*p1.x + (v.y/v.x)*p2.x);
		//dist.y = d * (p1.x - p2.x + (v.y/v.x)*p1.y + (v.x/v.y)*p2.y);
		dist.x = d * ((p2.y - p1.y)*v.x*v.y + (v.x*v.x)*p2.x + (v.y*v.y)*p1.x);
		dist.y = d * ((p2.x - p1.x)*v.x*v.y + (v.y*v.y)*p2.y + (v.x*v.x)*p1.y);
	}
}
