/*
 * Copyright 2005 Takao Sumitomo. All rights reserved.
 * 
 * Redistribution and use in source and binary forms,
 * with or without modification, are permitted provided 
 * that the following conditions are met:
 *
 *   1. Redistributions of source code must retain the 
 *      above copyright notice, this list of conditions
 *      and the following disclaimer.
 *   2. Redistributions in binary form must reproduce 
 *      the above copyright notice, this list of conditions
 *      and the following disclaimer in the documentation
 *      and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software
 * and documentation are those of the authors and should not be
 * interpreted as representing official policies,
 * either expressed or implied.
 * 
 * @Author: Takao Sumitomo
 */

/*
 * 作成日: 2005/02/07
 */
package bakusen.game.arena;

import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashSet;
import java.util.Iterator;

import bakusen.client.BUserInterface;
import bakusen.common.BPickIntersection;
import bakusen.common.BPickResult;
import bakusen.common.BStage;
import bakusen.common.Body;
import bakusen.common.DataManager;
import bakusen.common.EventHandler;
import bakusen.common.event.BodyAddEvent;
import bakusen.common.event.BodyDeleteEvent;
import bakusen.common.event.BodyEvent;
import bakusen.common.event.BodyMessageEvent;
import bakusen.common.event.BodyStateEvent;
import bakusen.common.math.BPolygon;
import bakusen.common.math.BVector2f;
import bakusen.common.util.BAudio;
import bakusen.game.deathmatch.BodyBom;
import bakusen.game.deathmatch.BodyItem;

/**
 * @author cattaka
 */
public class BodyMan extends Body
{
	public static final int ACTION_PUTBOM = 0;
	public static final int ACTION_PUNCH = 1;

	public static final int OPTION_FIRE = 0;
	public static final int OPTION_BOMS = 1;
	public static final int OPTION_SPEED = 2;
	public static final int OPTION_WEAPON = 3;

	public static final int WEAPON_KICK = 1<<1;
	public static final int WEAPON_GLOVE = 1<<2;
	public static final int WEAPON_PUNCH = 1<<3;

	public static final int DEFAULT_FIRE = 4;
	public static final int DEFAULT_BOMS = 4;
	public static final int DEFAULT_SPEED = 16;

	public static final int MAX_FIRE = 8;
	public static final int MAX_BOMS = 8;
	public static final int MAX_SPEED = 8;

	public static final long MINIMUM_THROW_INTERVAL = 1000;
	public static final long MINIMUM_PUNCH_INTERVAL = 500;

	private HashSet boms = new HashSet();
	private long lastThrow = 0;
	private long lastPunch = 0;
	
	public BodyMan()
	{
		setType(BStageArena.MAN1);
		setOptions(new int[]{0,0,0,0});
		setOption(BodyMan.OPTION_WEAPON, 0);
		setPolygon(new BPolygon(BPolygon.DIAGONAL_SQUARE, BStageArena.MAN_WIDTH));
		enableState(Body.LIFE);
		enableState(Body.DEAD);
		enableState(Body.WIN);
		enableState(Body.INVISIBLE);
	}
	
	public boolean hasName()
	{
		return true;
	}
	public boolean hasVariableBPolygon()
	{
		return false;
	} 
	public boolean hasScore()
	{
		return true;
	}

	public void actionForClient(BodyEvent bodyEvent, BAudio baudio, BUserInterface ui, long currenttime)
	{
		if (bodyEvent instanceof BodyAddEvent)
		{
			baudio.play(BStageArena.SOUND_LOGIN, 0);
		}
		else if (bodyEvent instanceof BodyDeleteEvent)
		{
			baudio.play(BStageArena.SOUND_LOGOUT, 0);
		}
		else if (bodyEvent instanceof BodyStateEvent)
		{
			BodyStateEvent bse = (BodyStateEvent)bodyEvent;
			if ((bse.getType() == Body.SET
					|| bse.getType() == Body.ENABLE)
					&& (bse.getState() & Body.DEAD) == Body.DEAD)
			{
				baudio.play(BStageArena.SOUND_DEAD, 0);
				ui.outputMessage(ui.translate("is_down", new String[]{getName()}));
			}
		}
		else if (bodyEvent instanceof BodyMessageEvent)
		{
			BodyMessageEvent bme = (BodyMessageEvent)bodyEvent;
			if (bme.getType() == DataManager.MESSAGE_CHAT)
			{
				baudio.play(BStageArena.SOUND_CHAT, 0);
			}
		}
	}
	public void eventAction(EventHandler eventHandler, BStage bstage, BVector2f position, int action)
	{
		if (isState(Body.DEAD) || !isState(Body.LIFE))
			return;
		
		if (action == ACTION_PUTBOM)
		{
			Object[] objs = boms.toArray();
			int num = 0;
			for (int i=0;i<objs.length;i++)
			{
				BodyBom bom = (BodyBom)objs[i];
				if (bom.getBstage() != null)
					num++;
				else
					boms.remove(bom);
			}
			if (boms.size() == 0)
			{
				boms.clear();
			}
			
			Body body = new BodyBom();
			body.setID(bstage.getNextID());
			float x = 0.5f + (int)(position.x);
			float y = 0.5f + (int)(position.y);
			body.setPosition(new BVector2f(x,y));
			body.setOption(BodyBom.OPTION_FIRE, getOption(OPTION_FIRE) + DEFAULT_FIRE);
			
			BPickResult pr = bstage.pickAllByType(body.getPosition(), body.getPolygon(), BStageArena.BOM1);
			long currenttime = bstage.getCurrentTime();
			if (pr.getSize() == 0)
			{
				if (num >= getOption(OPTION_BOMS) + DEFAULT_BOMS)
					return;
				
				boms.add(body);
				//eventHandler.sound(currenttime,currenttime,BStageArena.SOUND_PUTBOM);
				eventHandler.add(currenttime,currenttime,body);
				eventHandler.action(currenttime, currenttime + 3000, body, body.getPosition(), BodyBom.ACTION_BOMB);
			}else if ((getOption(OPTION_WEAPON) & WEAPON_GLOVE) == WEAPON_GLOVE)
			{
				if (currenttime - lastThrow < MINIMUM_THROW_INTERVAL)
					return;
				lastThrow = currenttime;
				
				BVector2f vec = new BVector2f(getDirection());
				if (vec.getLength() == 0)
					return;
				
				Iterator ite = pr.getBPickIntersections();
				BPickIntersection pi = (BPickIntersection)ite.next();
				BodyBom bom = (BodyBom)pi.getBody();
				if (Math.abs(vec.x) > Math.abs(vec.y))
					vec.y = 0;
				else
					vec.x = 0;
				vec.normalize();
				vec.scale(BStageArena.THROW_DISTANCE);
				vec.add(bom.getPosition());
				bom.jump(currenttime, vec, eventHandler);
			}
		}else if(action == ACTION_PUNCH
				&& (getOption(OPTION_WEAPON) & WEAPON_PUNCH) == WEAPON_PUNCH)
		{
			long currenttime = bstage.getCurrentTime();

			if (currenttime - lastPunch < MINIMUM_PUNCH_INTERVAL)
				return;
			lastPunch = currenttime;
			
			if (getDirection().getLength() == 0)
			{
				return;
			}
			BVector2f vec = new BVector2f(getDirection());
			if (Math.abs(vec.x) > Math.abs(vec.y))
				vec.y = 0;
			else
				vec.x = 0;
			vec.normalize();
			
			BVector2f pos = new BVector2f(getPosition());
			pos.add(vec);
			
			BPickResult pr = bstage.pickAllByType(pos, this.getPolygon(), BStageArena.BOM1);
			if (pr.getSize() == 0)
				return;
			
			Iterator ite = pr.getBPickIntersections();
			BPickIntersection pi = (BPickIntersection)ite.next();
			BodyBom bom = (BodyBom)pi.getBody();
			
			vec.scale(BStageArena.PUNCH_DISTANCE);
			vec.add(bom.getPosition());
			bom.jump(currenttime, vec, eventHandler);
		}
	}
	public float getMaxSpeed()
	{
		return 0.25f * (getOption(BodyMan.OPTION_SPEED) + DEFAULT_SPEED);
	}
	public void eventCollisition(EventHandler eventHandler, BStage bstage, Body target)
	{
		/*
		if (target.isState(Body.COLLIDABLE)
			&& !(BMath.isIntersect(getPosition(), getPolygon(), target.getPosition(), target.getPolygon())))
		{
			long currenttime = bstage.getCurrentTime();
			eventHandler.move(currenttime, currenttime, this, this.getPosition(), new BVector2d(0,0), this.getDirection());
		}
		*/
	}
	public void eventStateChange(EventHandler eventHandler, BStage bstage, byte type, int state)
	{
		if ((type == Body.SET
				|| type == Body.ENABLE)
				&& (state & Body.DEAD) == Body.DEAD)
		{
			if (isState(WIN))
			{
				return;
			}
			
			long currenttime = bstage.getCurrentTime();
			eventHandler.setPosition(currenttime, currenttime, this, getPosition());
			eventHandler.state(currenttime, currenttime + 3000, this, Body.ENABLE, Body.INVISIBLE);

			for (int i=0;i<getOption(OPTION_FIRE);i++)
			{
				BodyItem body = new BodyItem();
				body.setOption(BodyItem.OPTION_TYPE, BodyItem.ITEM_FIRE);
				throwItem(currenttime, body, bstage, eventHandler);
			}
			for (int i=0;i<getOption(OPTION_SPEED);i++)
			{
				BodyItem body = new BodyItem();
				body.setOption(BodyItem.OPTION_TYPE, BodyItem.ITEM_SPEED);
				throwItem(currenttime, body, bstage, eventHandler);
			}
			for (int i=0;i<getOption(OPTION_BOMS);i++)
			{
				BodyItem body = new BodyItem();
				body.setOption(BodyItem.OPTION_TYPE, BodyItem.ITEM_BOMS);
				throwItem(currenttime, body, bstage, eventHandler);
			}
			if ((getOption(OPTION_WEAPON) & WEAPON_KICK) == WEAPON_KICK)
			{
				BodyItem body = new BodyItem();
				body.setOption(BodyItem.OPTION_TYPE, BodyItem.ITEM_KICK);
				throwItem(currenttime, body, bstage, eventHandler);
			}
			if ((getOption(OPTION_WEAPON) & WEAPON_GLOVE) == WEAPON_GLOVE)
			{
				BodyItem body = new BodyItem();
				body.setOption(BodyItem.OPTION_TYPE, BodyItem.ITEM_GLOVE);
				throwItem(currenttime, body, bstage, eventHandler);
			}
			if ((getOption(OPTION_WEAPON) & WEAPON_PUNCH) == WEAPON_PUNCH)
			{
				BodyItem body = new BodyItem();
				body.setOption(BodyItem.OPTION_TYPE, BodyItem.ITEM_PUNCH);
				throwItem(currenttime, body, bstage, eventHandler);
			}
			
			Body[] bs = bstage.getBodysHasScore();
			int num = 0;
			for (int i=0;i<bs.length;i++)
			{
				if (!(bs[i].isState(Body.DEAD)))
				{
					num++;
					if (num > 1)
					{
						break;
					}
				}
			}
			BodyManager manager = ((BStageArena)bstage).getManager();
			if (num <= 1)
			{
				eventHandler.action(currenttime, currenttime+1000, manager, manager.getPosition(), BodyManager.ACTION_GAMESET);
			}
		}
	}
	private void throwItem(long currenttime, BodyItem item, BStage bstage, EventHandler eventHandler)
	{
		item.disableState(LIFE);
		item.setPosition(getPosition());
		BVector2f sub = new BVector2f(((BStageArena)bstage).getRandomeFreePoint());
		if (bstage.isIntersectOther(sub, item.getPolygon(), COLLIDABLE))
		    return;

		BVector2f pos = new BVector2f(sub);
		sub.sub(getPosition());
		sub.scale(2f);
		item.setVelocity(sub);
		
		eventHandler.add(currenttime, currenttime, item);
		eventHandler.state(currenttime, currenttime + 1000, item, ENABLE, Body.LIFE);
		eventHandler.setPosition(currenttime, currenttime + 500, item, pos);
		eventHandler.state(currenttime, currenttime + BStageArena.ITEM_SPAWN_INTERVAL, item, ENABLE, DEAD);
		eventHandler.delete(currenttime, currenttime + BStageArena.ITEM_SPAWN_INTERVAL + 500, item);
	}
	public void eventAdd(EventHandler eventHandler, BStage bstage)
	{
		long currenttime = bstage.getCurrentTime();
		//eventHandler.sound(currenttime ,currenttime, BStageArena.SOUND_LOGIN);
		if (bstage.getBodysHasScore().length == 2)
		{
			BodyManager manager = ((BStageArena)bstage).getManager();
			eventHandler.action(currenttime, currenttime+1000, manager, manager.getPosition(), BodyManager.ACTION_RESET);
			//eventHandler.stageAction(currenttime, currenttime + 1000, BStageArena.ACTION_RESET);
		}
	}
	public void eventPush(EventHandler eventHandler, BStage bstage, Body target)
	{
		if (isState(Body.DEAD) || !isState(Body.LIFE))
			return;

		if (target.isState(COLLIDABLE)
				&& target.getType() == BStageArena.BOM1
				&& (getOption(BodyMan.OPTION_WEAPON) & BodyMan.WEAPON_KICK) == BodyMan.WEAPON_KICK )
		{
			long currenttime = bstage.getCurrentTime();
			BVector2f v = new BVector2f(target.getPosition());
			v.sub(getPosition());
			if (Math.abs(v.x) >= Math.abs(v.y))
			{
				v.y = 0;
			}else
			{
				v.x = 0;
			}
			v.normalize();
			v.scale(getMaxSpeed());
			eventHandler.move(currenttime, currenttime, target, target.getPosition(), v, v);
		}
	}
	public void eventRemove(EventHandler eventHandler, BStage bstage)
	{
		//eventHandler.sound(bstage.getCurrentTime(),bstage.getCurrentTime(), BStageArena.SOUND_LOGOUT);
	}
	public void eventCommand(EventHandler eventHandler, BStage bstage, String command)
	{
		if (command.length() != 0)
		{
			long currenttime = bstage.getCurrentTime();
			if (command.charAt(0) != '/')
			{
				//eventHandler.sound(currenttime,currenttime,BStageArena.SOUND_CHAT);
				eventHandler.message(currenttime, currenttime, this, DataManager.MESSAGE_CHAT,command);
			}else if (command.length() > 6 && command.substring(0,6).equals("/note "))
			{
				SimpleDateFormat sdf = new SimpleDateFormat();
				String note = "["+this.getName()+"]{"+ sdf.format(new Date()) +")"+command.substring(6, command.length());
				eventHandler.message(currenttime, currenttime, this,DataManager.MESSAGE_NOTE, note);
			}
		}
	}
	public void preAdd(BStage bstage)
	{
		setPosition(bstage.getRandomeSpawnPoint());
	}
	public boolean isCheckIntersectWith(Body target)
	{
		if (target.getType() == BStageArena.ITEM
			|| target.getType() == BStageArena.FIRE1
			)
			return true;
		else
			return false;
	}
}
