/*
 * Copyright 2005 Takao Sumitomo. All rights reserved.
 * 
 * Redistribution and use in source and binary forms,
 * with or without modification, are permitted provided 
 * that the following conditions are met:
 *
 *   1. Redistributions of source code must retain the 
 *      above copyright notice, this list of conditions
 *      and the following disclaimer.
 *   2. Redistributions in binary form must reproduce 
 *      the above copyright notice, this list of conditions
 *      and the following disclaimer in the documentation
 *      and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software
 * and documentation are those of the authors and should not be
 * interpreted as representing official policies,
 * either expressed or implied.
 * 
 * @Author: Takao Sumitomo
 */

/*
 * 作成日: 2005/01/18
 */
package bakutan.common;

import java.awt.Graphics;
import java.awt.Image;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

import bakutan.common.event.BodyActionEvent;
import bakutan.common.event.BodyEvent;
import bakutan.common.graphics.CharacterImage;
import bakutan.common.graphics.CharacterImageBundle;
import bakutan.common.graphics.EffectBundle;
import bakutan.common.math.BMath;
import bakutan.common.math.BShape;
import bakutan.common.math.BVector2f;

/**
 * @author cattaka
 */
public abstract class Body
{
    /**
     * ダウン状態を表す定数。
     * isState(int)メソッドに使用する。
     */
	public static final int STATE_DEAD = 1<<1;
	/**
     * 衝突する壁の状態を表す定数。 
     * isState(int)メソッドに使用する。
	 */
    public static final int STATE_COLLIDABLE = 1<<2;
    /**
     * 不可視の状態を表す定数。 
     * isState(int)メソッドに使用する。
     */
	public static final int STATE_INVISIBLE = 1<<3;
    /**
     * 炎を止める壁の状態を表す定数(未使用)。 
     * isState(int)メソッドに使用する。
     */
	public static final int STATE_FIREWALL = 1<<4;
    /**
     * 体力を持つキャラクタのであることを表す定数。 
     * isState(int)メソッドに使用する。
     */
	public static final int STATE_LIFE = 1<<5;
    /**
     * 勝利状態を表す定数(未使用)。 
     * isState(int)メソッドに使用する。
     */
	public static final int STATE_WIN = 1<<6;
    /**
     * マーキング状態を表す定数(未使用)。 
     * isState(int)メソッドに使用する。
     */
	public static final int STATE_MARKER = 1<<7;

	public static final byte SET = 0;
	public static final byte ENABLE = 1;
	public static final byte DISABLE = 2;
    
	private long lastStateChange;
	private String name = "";
	private BVector2f position = new BVector2f();
	private BVector2f velocity = new BVector2f();
	private BVector2f direction = new BVector2f();
	private BShape shape = new BShape();
	private byte type;
	private long lastupdate;
	private BStage bstage;
	
	private int id = -1;
	private int state = 0;
	private int armor = 100;
	private int score = 0;
	private int[] options;
	
	//only for client
	private CharacterImage characterImage;
	private int imageMode = CharacterImage.DOWN;
	private long lastMessageChange;
	private String message = "";
	private boolean animLoop = true;
	private int actionLock = 0;
    private float energy = 0f;
    
    /**
     * 新しいBodyインスタンスを作成する。 
     */
	public Body(byte type, int optionNum)
	{
        this.type = type;
        options = new int[optionNum];
	}
	
    /**
     * このクラスのインスタンスが固有名を持つかどうかを返す。
     * @return 固有名を持つならばtrue、そうでなければfalse。
     */
	public boolean hasName()
	{
		return true;
	}
    /**
     * このクラスのインスタンスが可変サイズの形状データを持つかどうかを返す。
     * @return 可変サイズのBCircleを持つならばtrueをそうでなければfalseを返す。
     */
	public boolean hasVariableBCircle()
	{
		return true;
	}
    public boolean hasState()
    {
        return true;
    }
    public boolean hasArmor()
    {
        return true;
    }
    public boolean hasArmorReport()
    {
        return true;
    }
    public boolean hasOption()
    {
        return true;
    }
    /**
     * このクラスのインスタンスがスコアを持つかどうかを返す。
     * @return スコアを持つならばtrue、そうでないならfalseを返す。
     */
	public boolean hasScore()
	{
		return true;
	}
	
    /**
     * 与えられたインスタンスの内容をこのインスタンスにコピーする。
     * @param body 元となるインスタンス
     */
	public void set(Body body)
	{
		System.arraycopy(body.getOptions(), 0, options, 0, options.length);
		//setOptions(body.getOptions());
        setCharacterImage(body.getCharacterImage());
        setActionLock(body.getActionLock());
        setEnergy(body.getEnergy());
        
        setLastStateChange(body.getLastStateChange());
        setName(body.getName());
        setPosition(body.getPosition());
        setVelocity(body.getVelocity());
        setDirection(body.getDirection());
        getShape().setRadius(body.getShape().getRadius());
        setType(body.getType());
        setBstage(body.getBstage());
        setID(body.getID());
        setState(body.getState());
        setArmor(body.getArmor());
        setScore(body.getScore());
        setCharacterImage(body.getCharacterImage());
        setMessage(body.getLastMessageChange(), body.getMessage());
        setActionLock(body.getActionLock());
        setEnergy(body.getEnergy());
        
        lastupdate = body.lastupdate;
        System.arraycopy(body.options, 0, options, 0, options.length);
        imageMode = body.imageMode;
        animLoop = body.animLoop;
	}
    /**
     * 与えられたインスタンスの内容をこのインスタンスにコピーする。
     * @param body 元となるインスタンス
     */
    public void setWithoutPosition(Body body)
    {
        System.arraycopy(body.getOptions(), 0, options, 0, options.length);
        //setOptions(body.getOptions());
        setCharacterImage(body.getCharacterImage());
        setActionLock(body.getActionLock());
        setEnergy(body.getEnergy());
        
        setLastStateChange(body.getLastStateChange());
        setName(body.getName());
        getShape().setRadius(body.getShape().getRadius());
        setType(body.getType());
        setBstage(body.getBstage());
        setID(body.getID());
        setState(body.getState());
        setArmor(body.getArmor());
        setScore(body.getScore());
        setCharacterImage(body.getCharacterImage());
        setMessage(body.getLastMessageChange(), body.getMessage());
        setActionLock(body.getActionLock());
        setEnergy(body.getEnergy());
        
        lastupdate = body.lastupdate;
        options = new int[body.options.length];
        System.arraycopy(body.options, 0, options, 0, options.length);
        imageMode = body.imageMode;
        animLoop = body.animLoop;
    }
	public long getLastUpdate()
	{
		return this.lastupdate;
	}
	public void setCurrentTime(long currenttime)
	{
		this.lastupdate = currenttime;
	}
	private void checkPosition()
	{
		if (bstage == null)
			return;
		if (position.x > bstage.getSize().x)
			position.x %= bstage.getSize().x;
		else if(position.x < 0)
			position.x = (position.x % bstage.getSize().x) + bstage.getSize().x;

		if (position.y > bstage.getSize().y)
			position.y %= bstage.getSize().y;
		else if(position.y < 0)
			position.y = (position.y % bstage.getSize().y) + bstage.getSize().y;
	}
    public boolean isLoopPosition()
    {
        return false;
    }
	public void update(long currenttime)
	{
        long dl = currenttime - lastupdate;
        float d = (float)(currenttime - lastupdate) * 0.001f;
		this.lastupdate = currenttime;
		position.x += velocity.x * d;
		position.y += velocity.y * d;
		
        if (isLoopPosition())
            checkPosition();
        
        energy += getEnergyCharge() * d;
        if (energy > getMaxEnergy())
            energy = getMaxEnergy();
        
        actionLock -= dl;
        if (actionLock < 0)
            actionLock = 0;
    }
	public long getLastStateChange()
	{
		return lastStateChange;
	}
	public void setLastStateChange(long lastStateChange)
	{
		this.lastStateChange = lastStateChange;
	}
	public void setVelocity(BVector2f velocity)
	{
		this.velocity.set(velocity);
	}
    /**
     * 速度のBVector2fの実体を取得する。
     * @return 速度を表すBVector2f実態
     */
	public BVector2f getVelocity()
	{
		return velocity;
	}
	public void setPosition(BVector2f position)
	{
		this.position.set(position);
		checkPosition();
	}
    /**
     * 位置のBVector2fの実態を取得する。
     * @return 位置のBVector2fの実体
     */
	public BVector2f getPosition()
	{
		return position;
	}
    /**
     * 名前を取得する。
     * @return このインスタンスの名前
     */
	public String getName()
	{
		return name;
	}
	public void setName(String name)
	{
		this.name = name;
	}
    /**
     * このインスタンスの形状を表すBShapeの実体を返す。
     * @return 形状を表すBShapeの実体
     */
	public BShape getShape()
	{
		return shape;
	}
	public void setShape(BShape polygon)
	{
		this.shape = polygon;
	}
    /**
     * このBodyのタイプIDを返す。
     * @return タイプID
     */
	public byte getType() {
		return type;
	}
	public void setType(byte image) {
		this.type = image;
	}
    /**
     * このインスタンスのBStage上の固有IDを返す。
     * @return BStage上の固有ID
     */
	public int getID() {
		return id;
	}
	public void setID(int id) {
		this.id = id;
	}
    /**
     * このインスタンスの状態を表すint値を返す。通常は使用しない。
     * @return インスタンスの状態を表すint値
     * @see #getState() getState()
     * @see #isState(int) isState(int)
     */
	public int getState()
	{
		return this.state;
	}

    /**
     * このインスタンスの状態を表すint値を設定する。通常は使用しない。
     * @param state 設定するインスタンスの状態
     * @see #enableState(int) enableState(int)
     * @see #disableState(int) disableState(int)
     */
	public void setState(int state)
	{
		this.state = state;
	}

    /**
     * このインスタンスの指定した状態のみtrueに設定する。
     * @param state 設定する状態のID
     * @see #STATE_COLLIDABLE STATE_COLLIDABLE
     * @see #STATE_DEAD STATE_DEAD
     * @see #STATE_FIREWALL STATE_FIREWALL
     * @see #STATE_LIFE STATE_LIFE
     * @see #STATE_INVISIBLE STATE_INVISIBLE 
     */
	public void enableState(int state)
	{
		this.state |= state;
	}
    /**
     * このインスタンスの指定した状態のみfalseに設定する。
     * @param state 設定する状態のID
     * @see #STATE_COLLIDABLE STATE_COLLIDABLE
     * @see #STATE_DEAD STATE_DEAD
     * @see #STATE_FIREWALL STATE_FIREWALL
     * @see #STATE_LIFE STATE_LIFE
     * @see #STATE_INVISIBLE STATE_INVISIBLE 
     */
	public void disableState(int state)
	{
		this.state = (this.state & ~state);
	}
    
	/**
     * 指定した状態の値を返す。
     * @param state 欲しい状態のID
     * @return 指定した状態の値
     * @see #STATE_COLLIDABLE STATE_COLLIDABLE
     * @see #STATE_DEAD STATE_DEAD
     * @see #STATE_FIREWALL STATE_FIREWALL
     * @see #STATE_LIFE STATE_LIFE
     * @see #STATE_INVISIBLE STATE_INVISIBLE 
	 */
    public boolean isState(int state)
	{
		return (this.state & state) == state;
	}
    
    /**
     * このインスタンスのオプションを取得する。
     * @return オプションを表すint配列の実体
     */
	public int[] getOptions()
	{
		return options;
	}
//	protected void setOptions(int[] options)
//	{
//		this.options = options;
//	}
	public void setOption(int type, int value)
	{
		this.options[type] = value;
	}
    /**
     * このインスタンスの指定したtypeの値を取得する。
     * 引数typeの値については各サブクラスのOPTION_???フィールドを参照。
     * @param type 得ようとするtypeのID
     * @return 指定したオプションの値
     */
	public int getOption(int type)
	{
		return this.options[type];
	}
    /**
     * このインスタンスの装甲を取得する。
     * @return インスタンスの装甲ポイント
     */
	public int getArmor()
	{
		return armor;
	}
    /**
     * このインスタンスのスコアを取得する。
     * @return スコア
     */
	public int getScore()
	{
		return score;
	}
	public void setScore(int score)
	{
		this.score = score;
	}
	public void setArmor(int armor)
	{
		this.armor = armor;
	}
	public void drawCharacter(Graphics g , long currenttime, CharacterImageBundle cib, EffectBundle eb)
	{
		if (characterImage == null)
			return;
		
		int offsetx = characterImage.getWidth() / 2;
		int offsety = characterImage.getHeight() / 2;
		//int x = -offset.width + (int) (getPosition().x * DataManager.BLOCKSIZE - offsetx);
		//int y = -offset.height + (int) (getPosition().y * DataManager.BLOCKSIZE - offsety);
        int x = -offsetx;
        int y = -offsety;
		int n = ((int)(currenttime - getLastStateChange()))/DataManager.ANIMINTERVAL;

		Image img = characterImage.getImages(this, n);
		g.drawImage(img, x, y, null);
	}
	public String getMessage()
	{
		return message;
	}
	public void setMessage(long lastMessageChange, String message)
	{
		this.lastMessageChange = lastMessageChange;
		this.message = message;
	}
	public long getLastMessageChange()
	{
		return lastMessageChange;
	}
	public CharacterImage getCharacterImage()
	{
		return characterImage;
	}
	public void setCharacterImage(CharacterImage characterImage)
	{
		this.characterImage = characterImage;
	}
	public BStage getBstage()
	{
		return bstage;
	}
	public void setBstage(BStage bstage)
	{
		this.bstage = bstage;
	}
    /**
     * このインスタンスが出せる最高速度を返す。
     * @return このインスタンスが出せる最高速度
     */
	public float getMaxSpeed()
	{
		return 0;
	}
    
    /**
     * このインスタンスの方向を表すBVector2fの実体を返す。
     * @return このインスタンスの方向を表すBVector2fの実体
     */
	public BVector2f getDirection()
	{
		return direction;
	}
	public void setDirection(BVector2f direction)
	{
		this.direction.set(direction);
        if (this.direction.isZeroVector())
        {
            this.direction.set(0,1);
        }
        else
        {
            this.direction.normalize();
        }
	}
    abstract public byte[] getCheckCollidType();
    abstract public byte[] getCheckIntersectType();
	public boolean isCheckCollidWith(Body target)
	{
		return target.isState(STATE_COLLIDABLE);
	}
	public boolean isCheckIntersectWith(Body target)
	{
		return true;
	}
	public boolean doPushReport()
	{
		return true;
	}
	public void actionForClient(BodyEvent bodyEvent, ClientBody mybody, CharacterImageBundle cib, EffectBundle eb, BUserInterface ui, long currenttime, BVector2f bufferCenter)
	{
	}
	public void eventCollisition(EventHandler eventHandler, BStage bstage, Body target)
	{
	}
	public void eventPush(EventHandler eventHandler, BStage bstage, Body target)
	{
	}
	public boolean eventAction(EventHandler eventHandler, BStage bstage, BVector2f position, BodyActionEvent event)
	{
        return false;
	}
	public void eventOptionChange(EventHandler eventHandler, BStage bstage, int type, int value)
	{
	}
    public void eventArmorChange(EventHandler eventHandler, BStage bstage, Body from, int type, int value)
    {
    }
    public void eventChangeOthersArmor(EventHandler eventHandler, BStage bstage, Body to, int type, int value)
    {
    }
	public void eventStateChange(EventHandler eventHandler, BStage bstage, byte type, int state)
	{
	}
	public void eventAdd(EventHandler eventHandler, BStage bstage)
	{
	}
	public void eventRemove(EventHandler eventHandler, BStage bstage)
	{
	}
	public void eventCommand(EventHandler eventHandler, BStage bstage, String command)
	{
	}
	public void preAdd(BStage bstage)
	{
	}
	public boolean isAnimLoop()
	{
		return animLoop;
	}
	public void setAnimLoop(boolean animLoop)
	{
		this.animLoop = animLoop;
	}
    public short getActionLockTime(byte action)
    {
        return 0;
    }
    public float getEnergyConsumption(byte action)
    {
        return 0f;
    }
    public float getEnergyCharge()
    {
        return 0f;
    }
    public float getMaxEnergy()
    {
        return 0f;
    }

    public int getActionLock()
    {
        return actionLock;
    }

    public void setActionLock(int actionLock)
    {
        this.actionLock = actionLock;
    }

    public float getEnergy()
    {
        return energy;
    }

    public void setEnergy(float energy)
    {
        this.energy = energy;
    }
    public void addActionLockEnergy(float energy, int actionLock)
    {
        this.actionLock += actionLock;
        this.energy += energy;
    }
//    public boolean isActionableForClient(byte action)
//    {
//        return isActionableForClient(action, getEnergy(), getActionLock());
//    }
    public boolean isActionableForClient(byte action, float energy, int actionlock)
    {
        boolean flag = true;
        if (energy < getEnergyConsumption(action))
        {
            flag = false;
        }
        if (actionlock > 0)
        {
            flag = false;
        }
        return flag;
    }
//    public boolean isInSite(Dimension offset, Dimension size)
//    {
//        BVector2f pos = getPosition();
//        int cx = (int)(pos.x * DataManager.BLOCKSIZE);
//        int cy = (int)(pos.y * DataManager.BLOCKSIZE);
//        int csize = (int)(getShape().getRadius() * DataManager.BLOCKSIZE);
//        if (cx + csize < offset.width || offset.width + size.width < cx - csize)
//        {
//            if (cy + csize < offset.height || offset.height + size.height < cy - csize)
//            {
//                return false;
//            }
//        }
//        return true;
//    }
    public boolean isInSite(BVector2f pos, BShape shape)
    {
        return BMath.isIntersect(pos, shape, getPosition(), getShape());
    }
    public void outputData(DataManager dm, DataOutputStream out) throws IOException
    {
        //out.writeByte(getType());
        if (hasName())
            out.writeUTF(getName());
        out.writeShort(getID());
        if (hasState())
            out.writeShort(getState());
        if (hasArmor())
            out.writeShort(getArmor());
        if (hasOption())
        {
            int[] bs = getOptions();
            //out.writeByte(bs.length);
            for (int i=0;i<bs.length;i++)
            {
                out.writeShort((short)bs[i]);
            }
        }
        dm.outputData(out, getPosition());
        dm.outputData(out, getVelocity());
        //outputData(out, getDirection());
        out.writeByte(DataManager.convBVectorToByte(getDirection()));

        if (hasScore())
            out.writeInt(getScore());
        if (hasVariableBCircle())
            dm.outputData(out,getShape());
    }
    public void inputData(DataManager dm, DataInputStream in) throws IOException
    {
        if (hasName())
            setName(in.readUTF());
        setID(in.readShort());
        if (hasState())
            setState(in.readShort());
        if (hasArmor())
            setArmor(in.readShort());
        if (hasOption())
        {
            //byte l = in.readByte();
            int[] bs = new int[options.length];
            for (int i=0;i<bs.length;i++)
            {
                bs[i] = in.readShort();
            }
            System.arraycopy(bs, 0, options, 0, options.length);
        }
        BVector2f temp;
        temp = dm.inputBVector2d(in);
        if (Float.isNaN(temp.x) || Float.isNaN(temp.y))
        {
            throw new IOException("Broken data");
        }
        getPosition().set(temp);
        temp = dm.inputBVector2d(in);
        if (Float.isNaN(temp.x) || Float.isNaN(temp.y))
        {
            throw new IOException("Broken data");
        }
        getVelocity().set(temp);
        //temp = inputBVector2d(in);
        temp = DataManager.convByteToBVector(in.readByte());
        if (Float.isNaN(temp.x) || Float.isNaN(temp.y))
        {
            throw new IOException("Broken data");
        }
        getDirection().set(temp);

        if (hasScore())
            setScore(in.readInt());
        if (hasVariableBCircle())
            setShape(dm.inputBShape(in));
    }
}
