/*
 * 作成日: 2006/12/27
 */
package bakutan.common.ai;

import bakutan.common.Body;
import bakutan.common.ClientBody;
import bakutan.common.math.BShape;
import bakutan.common.math.BVector2f;

/**
 * @author cattaka
 *
 * Body(キャラクタ)の情報を得るためのクラス。
 * 各メソッドによりキャラクタの情報を得ることができる。
 */

public class BodyInfoForAI
{
    private Body body;
    public BodyInfoForAI(Body body)
    {
        super();
        this.body = body;
    }
    /**
     * 速度を取得する。
     * @return 速度を表すBVector2f
     */
    public BVector2f getVelocity()
    {
        return new BVector2f(body.getVelocity());
    }
    /**
     * 位置を取得する。
     * @return 位置を表すBVector2f
     */
    public BVector2f getPosition()
    {
        return new BVector2f(body.getPosition());
    }
    /**
     * このインスタンスの方向を表すBVector2fを返す。
     * @return このインスタンスの方向を表すBVector2f
     */
    public BVector2f getDirection()
    {
        return new BVector2f(body.getDirection());
    }
    /**
     * 名前を取得する。
     * @return 名前
     */
    public String getName()
    {
        return body.getName();
    }
    /**
     * このインスタンスの形状を表すBShapeを返す。
     * @return 形状を表すBShape
     */
    public BShape getShape()
    {
        return new BShape(body.getShape().getRadius());
    }
    /**
     * このBodyのタイプIDを返す。
     * @return このBodyのタイプID
     * @see bakutan.game.deathmatch.BStageDM#TYPE_UNKNOWN TYPE_UNKNOWN
     * @see bakutan.game.deathmatch.BStageDM#TYPE_WALL1 TYPE_WALL1
     * @see bakutan.game.deathmatch.BStageDM#TYPE_PLAYER1 TYPE_MAN1
     * @see bakutan.game.deathmatch.BStageDM#TYPE_MISSILE1 TYPE_MISSILE1
     * @see bakutan.game.deathmatch.BStageDM#TYPE_FIRE1 TYPE_FIRE1
     * @see bakutan.game.deathmatch.BStageDM#TYPE_ITEM TYPE_ITEM
     * @see bakutan.game.deathmatch.BStageDM#TYPE_SOFTWALL TYPE_SOFTWALL
     * @see bakutan.game.deathmatch.BStageDM#TYPE_BOMB1 TYPE_BOM1
     * @see bakutan.game.deathmatch.BStageDM#TYPE_SLUG1 TYPE_SLUG1
     * @see bakutan.game.deathmatch.BStageDM#TYPE_HOMING1 TYPE_HOMING1
     */
    public byte getType() {
        return body.getType();
    }
    /**
     * このインスタンスのBStage上の固有IDを返す。
     * @return BStage上の固有ID
     */
    public int getID() {
        return body.getID();
    }
    /**
     * 指定した状態の値を返す。
     * @param state 欲しい状態のID
     * @return 指定した状態の値
     * @see bakutan.common.Body#STATE_COLLIDABLE STATE_COLLIDABLE
     * @see bakutan.common.Body#STATE_DEAD STATE_DEAD
     * @see bakutan.common.Body#STATE_FIREWALL STATE_FIREWALL
     * @see bakutan.common.Body#STATE_LIFE STATE_LIFE
     * @see bakutan.common.Body#STATE_INVISIBLE STATE_INVISIBLE 
     */
    public boolean isState(int state)
    {
        return body.isState(state);
    }
    /**
     * このインスタンスの指定したtypeの値を取得する。
     * 引数typeの値については各サブクラスのOPTION_???フィールドを参照。
     * @param type 得ようとするtypeのID
     * @return 指定したオプションの値
     * @see bakutan.game.deathmatch.BodyPlayer#OPTION_FIRE
     * @see bakutan.game.deathmatch.BodyPlayer#OPTION_CHARGE
     * @see bakutan.game.deathmatch.BodyPlayer#OPTION_SPEED
     * @see bakutan.game.deathmatch.BodyMissile#OPTION_FIRE
     * @see bakutan.game.deathmatch.BodyFire#OPTION_SIZE
     * @see bakutan.game.deathmatch.BodyBomb#OPTION_FIRE
     * @see bakutan.game.deathmatch.BodySlug#OPTION_FIRE
     * @see bakutan.game.deathmatch.BodyHoming#OPTION_FIRE
     */
    public int getOption(int type)
    {
        return body.getOption(type);
    }
    /**
     * このインスタンスの装甲ポイントを取得する。
     * @return インスタンスの装甲ポイント
     */
    public int getArmor()
    {
        return body.getArmor();
    }
    /**
     * このインスタンスのスコアを取得する。
     * @return スコア
     */
    public int getScore()
    {
        return body.getScore();
    }
    /**
     * 単位時間(1秒)当たりのエネルギーの回復量を得る。
     * @return 単位時間(1秒)当たりのエネルギーの回復量
     */
    public float getAmmoCharge()
    {
        return body.getEnergyCharge();
    }
//    @Override
//    public boolean equals(Object obj)
//    {
//        BodyInfoForAI bifa = (BodyInfoForAI)obj;
//        return this.getID() == bifa.getID();
//    }
    
    /**
     * 指定されたアクションに対するディレイを返す。
     * @param action 対象となるアクション
     * @return ディレイ
     */
    public short getActionLockTime(byte action)
    {
        return body.getActionLockTime(action);
    }
    /**
     * 指定されたアクションに対するエネルギーの消費を返す。
     * @param action 対象となるアクション
     * @return エネルギーの消費
     */
    public float getAmmoConsumption(byte action)
    {
        return body.getEnergyConsumption(action);
    }
    /**
     * 満タン時のエネルギーの最大値を返す。
     * @return 満タン時のエネルギーの最大値
     */
    public float getMaxAmmo()
    {
        return body.getMaxEnergy();
    }
    /**
     * プレイヤーの時、最高速度を返す。
     * @return プレイヤーの時、最高速度。
     */
    public float getMaxSpeed()
    {
        return body.getMaxSpeed();
    }

    BVector2f getRawPosition()
    {
        return body.getPosition();
    }
    /**
     * 単位時間(1秒)当たりのエネルギーの装填量を返す。
     * @return 単位時間当たりのエネルギーの装填量
     */
    public float getEnergyCharge()
    {
        return body.getEnergyCharge();
    }
    /**
     * 指定されたアクションのエネルギー消費を返す。
     * @param action 指定するアクション
     * @return エネルギー消費
     */
    public float getEnergyConsumption(byte action)
    {
        return body.getEnergyConsumption(action);
    }
    /**
     * エネルギーの最大値を返す。
     * @return エネルギーの最大値
     */
    public float getMaxEnergy()
    {
        return body.getMaxEnergy();
    }
    /**
     * 可能なアクションの数を返す。
     * @return 可能なアクションの数
     */
    public byte getActionNum()
    {
        if (body instanceof ClientBody)
            return ((ClientBody)body).getActionNum();
        return 0;
    }
}
