/*
 * Copyright 2005 Takao Sumitomo. All rights reserved.
 * 
 * Redistribution and use in source and binary forms,
 * with or without modification, are permitted provided 
 * that the following conditions are met:
 *
 *   1. Redistributions of source code must retain the 
 *      above copyright notice, this list of conditions
 *      and the following disclaimer.
 *   2. Redistributions in binary form must reproduce 
 *      the above copyright notice, this list of conditions
 *      and the following disclaimer in the documentation
 *      and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software
 * and documentation are those of the authors and should not be
 * interpreted as representing official policies,
 * either expressed or implied.
 * 
 * @Author: Takao Sumitomo
 */

/*
 * 作成日: 2005/01/18
 */
package bakutan.common.graphics;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.GraphicsConfiguration;
import java.awt.Image;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.io.InputStream;

import javax.imageio.ImageIO;

import bakutan.common.Body;
import bakutan.common.math.BVector2f;

/**
 * Four Direction ver
 * @author cattaka
 */
public class CharacterImageFD implements CharacterImage
{
	private BufferedImage[][] imgs;
	private int width = 32;
	private int height = 32;
	private int transparency;
    
	public CharacterImageFD(GraphicsConfiguration gc, int transparency, InputStream in)
	throws IOException
	{
		BufferedImage img = ImageIO.read(in);
        this.transparency = transparency;
		createImage(gc, img);
	}
	private void createImage(GraphicsConfiguration gc, BufferedImage img)
	{
		Color bgcolor;
		bgcolor = new Color(img.getRGB(0,0));
		imgs = new BufferedImage[12][2];
		for (int i=0;i<12;i++)
		{
			int y = i * height;
            for (int j=0;j<2;j++)
            {
                int x = j * width;
                if (gc != null)
                    imgs[i][j] = gc.createCompatibleImage(width,height,transparency);
                else
                    imgs[i][j] = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB_PRE);
                Graphics g = imgs[i][j].getGraphics();
                g.drawImage(img, 0, 0, width, height, x, y, x+width, y+height,null);
                g.dispose();
            }
            //imgs[i] = new BufferedImage[]{img.getSubimage(0,y,width,height),img.getSubimage(width,y,width,height)};
		}
	}
	public Image[] getImages(int state)
	{
		return imgs[state];
	}
	public Image getImages(Body body, int count)
	{
		if (body.isState(Body.STATE_WIN))
		{
			if (count < imgs[WIN].length)
			{
				return imgs[WIN][count];
			}else
			{
				return imgs[WINSTOP][count % imgs[WINSTOP].length];
			}
		}
		else if (body.isState(Body.STATE_DEAD))
		{
			if (count < imgs[DEAD].length)
			{
				return imgs[DEAD][count];
			}else
			{
				return imgs[DEADSTOP][count % imgs[DEADSTOP].length];
			}
		}else
		{
			BVector2f vec = body.getDirection();
			int imageMode = DOWN;
			if (vec.isZeroVector())
			{
				imageMode = WIN;
			}
			else if (!body.getVelocity().isZeroVector())
			{
                int dir = getDirNum(vec);
                switch(dir)
                {
                    case 0:
                        imageMode = CharacterImageFD.MOVERIGHT;
                        break;
                    case 1:
                        imageMode = CharacterImageFD.MOVEUP;
                        break;
                    case 2:
                        imageMode = CharacterImageFD.MOVELEFT;
                        break;
                    case 3:
                        imageMode = CharacterImageFD.MOVEDOWN;
                        break;
                    default:
                        imageMode = CharacterImageFD.WIN;
                        break;
                }
			}
			else
			{
                int dir = getDirNum(vec);
                switch(dir)
                {
                    case 0:
                        imageMode = CharacterImageFD.RIGHT;
                        break;
                    case 1:
                        imageMode = CharacterImageFD.UP;
                        break;
                    case 2:
                        imageMode = CharacterImageFD.LEFT;
                        break;
                    case 3:
                        imageMode = CharacterImageFD.DOWN;
                        break;
                    default:
                        imageMode = CharacterImageFD.WIN;
                        break;
                }
			}
			return imgs[imageMode][count % imgs[imageMode].length];
		}
	}
	public int getHeight()
	{
		return height;
	}
	public int getWidth()
	{
		return width;
	}
    private int getDirNum(BVector2f direction)
    {
        BVector2f v = new BVector2f(direction);
        if (v.isZeroVector())
            return 0;
        v.normalize();
        float l = v.getLength();
        if (l == 0)
            return 3;
        double rad = Math.acos(v.x/l);
        if (v.y > 0)
            rad = 2 * Math.PI - rad;
        rad += (2 * Math.PI) / 8;
        int n = (int)(rad / (2 * Math.PI) * 4);
        if (n > 3)
            n = 0;
        return n;
    }
    public void dispose()
    {        
        for (int i=0;i<imgs.length;i++)
        {
            for (int j=0;j<imgs[i].length;j++)
            {
                imgs[i][j].flush();
            }
        }
    }
}
