/*
 * Copyright 2005 Takao Sumitomo. All rights reserved.
 * 
 * Redistribution and use in source and binary forms,
 * with or without modification, are permitted provided 
 * that the following conditions are met:
 *
 *   1. Redistributions of source code must retain the 
 *      above copyright notice, this list of conditions
 *      and the following disclaimer.
 *   2. Redistributions in binary form must reproduce 
 *      the above copyright notice, this list of conditions
 *      and the following disclaimer in the documentation
 *      and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software
 * and documentation are those of the authors and should not be
 * interpreted as representing official policies,
 * either expressed or implied.
 * 
 * @Author: Takao Sumitomo
 */

/*
 * 作成日: 2005/01/14
 */
package bakutan.common.math;

/**
 * @author cattaka
 * ２次元ベクトルを表すクラス
 */
public class BVector2f
{
	public float x=0, y=0;

    /**
     * (0,0)で初期化する。
     */
	public BVector2f()
	{
	}

    /**
     * 与えられたx,yでベクトルを初期化する。
     * @param x x値
     * @param y y値
     */
	public BVector2f(float x, float y)
	{
		this.x = x;
		this.y = y;
	}
    /**
     * 与えられたベクトルと同じ値で初期化する。
     * @param v 元となるベクトル
     */
	public BVector2f(BVector2f v)
	{
		this.set(v);
	}
    /**
     * ベクトルの値を設定する
     * @param x x値
     * @param y y値
     */
	public void set(float x,float y)
	{
		this.x = x;
		this.y = y;
	}
    /**
     * 与えられたベクトルの値をこのインスタンスに設定する。
     * @param v 元となるベクトル
     */
	public void set(BVector2f v)
	{
		this.x = v.x;
		this.y = v.y;
	}
    /**
     * 引数vにこのインスタンスの内容をコピーする。
     * @param v コピー先
     */
	public void get(BVector2f v)
	{
		v.x = this.x;
		v.y = this.y;
	}
    /**
     * このベクトルと引数のスカラー倍をとる。
     * @param scale スカラー値
     */
	public void scale(float scale)
	{
		x *= scale;
		y *= scale;
	}
    
    /**
     * 引数vをこのインスタンスに加算する。
     * @param v 加算するベクトル
     */
	public void add(BVector2f v)
	{
		x += v.x;
		y += v.y;
	}
    /**
     * 引数dと引数vのスカラー倍をこのインスタンスに加算する。
     * @param d スカラー値
     * @param v ベクトル
     */
	public void scaleAdd(float d, BVector2f v)
	{
		x += v.x * d;
		y += v.y * d;
	}
    
    /**
     * 引数vをこのインスタンスに減算する。
     * @param v 減算するベクトル
     */
	public void sub(BVector2f v)
	{
		x -= v.x;
		y -= v.y;
	}
    /**
     * 2つのベクトルの内積を返す。
     * @param a
     * @param b
     * @return 内積の値。
     */
	public static float getDot(BVector2f a, BVector2f b)
	{
		return a.x*b.x + a.y*b.y;
	}
    /**
     * 2つのベクトルの外積を返す。
     * @param a
     * @param b
     * @return 外積の値。
     */
    public static float getCross(BVector2f a, BVector2f b)
    {
        return a.x*b.y - a.y*b.x;
    }
    /**
     * 2つのベクトルの距離を返す。
     * @param a
     * @param b
     * @return 2つのベクトルの距離。
     */
	public static float getDistance(BVector2f a, BVector2f b)
	{
		return (float)Math.sqrt((a.x-b.x)*(a.x-b.x) + (a.y-b.y)*(a.y-b.y));
	}
    /**
     * 2つのベクトルの角度(ラジアン値)を返す。ベクトルaからベクトルbへの反時計回りの角度となる。
     * @param a
     * @param b
     * @return 2つのベクトルの角度(ラジアン値)
     */
	public static double getAngle(BVector2f a, BVector2f b)
    {
        float s = a.getLength() * b.getLength();
        double rad = Math.acos(BVector2f.getDot(a,b)/s);
        float cross = BVector2f.getCross(a,b);
        if (cross >= 0)
            return rad;
        else
            return -rad;
    }
    /**
     * このベクトルの長さを返す。
     * @return ベクトルの長さ
     */
	public float getLength()
	{
		float d = (float)Math.sqrt(x*x + y*y);
		return d;
	}
    /**
     * このベクトルの長さの二乗を返す。
     * @return ベクトルの長さの二乗
     */
    public float getLengthSquare()
    {
        return (x*x + y*y);
    }

    /**
     * ゼロベクトルかどうか調べる。
     * @return ゼロベクトルならtrue、そうでなければfales。
     */
	public boolean isZeroVector()
    {
	    return x == 0 && y == 0;
    }
    /**
     * このベクトルを正規化する。
     * 長さが0のときは(NaN,NaN)になるので注意すること。
     */
	public void normalize()
	{
		float d = (float)Math.sqrt(x*x + y*y);
		if (d > 0)
		{
			x = x/d;
			y = y/d;
		}
        else
        {
            x = Float.NaN;
            y = Float.NaN;
        }
	}
    /**
     * このベクトルを反時計回りに回転させる。
     * @param rad 回転させる角度(ラジアン値)
     */
    public void rotate(double rad)
    {
        float xd = (float)(x*Math.cos(rad) - y*Math.sin(rad));
        float yd = (float)(y*Math.cos(rad) + x*Math.sin(rad));
        x = xd;
        y = yd;
    }
	public String toString()
	{
		return "( "+x+", "+y+")";
	}
    /**
     * ベクトルの比較して一致するならtrueを返す。そうでなければfalseを返す。
     */
	public boolean equals(Object obj)
	{
		BVector2f v = (BVector2f)obj;
		return (this.x == v.x && this.y == v.y);
	}
}
