/*
 * Copyright 2005 Takao Sumitomo. All rights reserved.
 * 
 * Redistribution and use in source and binary forms,
 * with or without modification, are permitted provided 
 * that the following conditions are met:
 *
 *   1. Redistributions of source code must retain the 
 *      above copyright notice, this list of conditions
 *      and the following disclaimer.
 *   2. Redistributions in binary form must reproduce 
 *      the above copyright notice, this list of conditions
 *      and the following disclaimer in the documentation
 *      and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software
 * and documentation are those of the authors and should not be
 * interpreted as representing official policies,
 * either expressed or implied.
 * 
 * @Author: Takao Sumitomo
 */

package bakutan.common.util;

import java.io.EOFException;
import java.io.IOException;
import java.io.InputStream;
import java.util.zip.DataFormatException;
import java.util.zip.Inflater;

/**
 * @author cattaka
 *
 * GZIPによって圧縮された通信向けのInputStreamを提供します。
 */
public class CompressedInputStream extends InputStream
{
	private InputStream in;
	private Inflater inf = new Inflater();
	private ResetableByteArrayOutputStream buf_in = new ResetableByteArrayOutputStream();
	private ResetableByteArrayInputStream buf_out = new ResetableByteArrayInputStream(new byte[0]);
    private ResetableByteArrayOutputStream tout;
	private byte[] bytebuf;
    private boolean mode_compress = true;
    //buf_in -> (inflat > tout or bytebuf) -> buf_out
    //buf_out == bytebuf || buf_out == tout.buf 
    
	/**
	 * 指定した入力ストリームを利用してCompressedInputStreamを作成します。
	 * CompressedInputStream(out,1024)と同じ動作をします。
	 * @param in
	 * 入力ストリーム
	 */
	public CompressedInputStream(InputStream in)
	{
		this.in = in;
		bytebuf = new byte[1024];
        tout = new ResetableByteArrayOutputStream(1024);
	}
	/**
	 * 指定した入力ストリームを利用してCompressedInputStreamを作成します。
	 * バッファは解凍時に使用されます。
	 * 解凍されるデータがサイズをオーバーするときは一時的に拡張されます。
	 * @param in
	 * 入力ストリーム
	 * @param bufsize
	 * バッファサイズ
	 */
	public CompressedInputStream(InputStream in, int bufsize)
	{
		this.in = in;
		bytebuf = new byte[bufsize];
        tout = new ResetableByteArrayOutputStream(bufsize);
	}
	
	/**
	 * 入力ストリームから1バイト読み込みます。
	 */
	public int read() throws IOException
	{
        synchronized (buf_out)
        {
            int r = buf_out.read();
            if (r == -1)
            {
                fill();
                return read();
            }else
            {
                return r;
            }
        }
	}
	private void fill() throws IOException
	{
        tout.reset();
        int ch1 = in.read();
        int ch2 = in.read();
        if ((ch1 | ch2) < 0)
            throw new EOFException();
        int s = (short)((ch1 << 8) + (ch2 << 0));
        if (s == -1)
        {
            mode_compress = !mode_compress;
            ch1 = in.read();
            ch2 = in.read();
            if ((ch1 | ch2) < 0)
                throw new EOFException();
            s = (short)((ch1 << 8) + (ch2 << 0));
        }
        //TO-DO debug code
        //System.out.println(Thread.currentThread().getName() + "<" + s);
        if (!mode_compress)
        {
            for (int i=0;i<s;i++)
            {
                tout.write(in.read());
            }
            //buf_out = new ResetableByteArrayInputStream(tout.getOriginalDate(), 0, tout.size());
            buf_out.setBuf(tout.getOriginalDate(), 0, tout.size());
            return;
        }
        else
        {
            if (s >= bytebuf.length)
                throw new IOException("Too big packet" + s);
            
            buf_in.reset();
            for (int i=0;i<s;i++)
            {
                buf_in.write(in.read());
            }
            byte[] bs = buf_in.getOriginalDate();
            inf.reset();
            inf.setInput(bs, 0, buf_in.size());
            int ds;
            try
            {
                ds = inf.inflate(bytebuf);
                if (inf.needsInput())
                {
                    //buf_out = new ResetableByteArrayInputStream(bytebuf,0,ds);
                    buf_out.setBuf(bytebuf,0,ds);
                    //System.out.println("<"+MD5Tool.makeString(bytebuf));
                    //System.out.println(Thread.currentThread().getName()+"<"+mode_compress+s+":"+bytebuf[0]);
                }else
                {
                    tout.reset();
                    tout.write(bytebuf,0,ds);
                    do
                    {
                        ds = inf.inflate(bytebuf);
                        tout.write(bytebuf,0,ds);
                    }while(!(inf.needsInput()));
                    buf_out.setBuf(tout.getOriginalDate(), 0, tout.size());
                }
            }catch(DataFormatException e)
            {
                throw new IOException(e.getMessage());
            }
        }
		//System.out.println("decompressed " + s + "to" + ds);
	}
    @Override
    public void close() throws IOException
    {
        in.close();
    }
}
