/*
 * Copyright 2005 Takao Sumitomo. All rights reserved.
 * 
 * Redistribution and use in source and binary forms,
 * with or without modification, are permitted provided 
 * that the following conditions are met:
 *
 *   1. Redistributions of source code must retain the 
 *      above copyright notice, this list of conditions
 *      and the following disclaimer.
 *   2. Redistributions in binary form must reproduce 
 *      the above copyright notice, this list of conditions
 *      and the following disclaimer in the documentation
 *      and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software
 * and documentation are those of the authors and should not be
 * interpreted as representing official policies,
 * either expressed or implied.
 * 
 * @Author: Takao Sumitomo
 */

package bakutan.common.util;

import java.io.IOException;
import java.io.OutputStream;
import java.util.zip.Deflater;

/**
 * @author cattaka
 *
 * GZIPによって圧縮された通信向けのOutputStreamを提供します。
 * このストリームはflush()が呼び出されるまで入力された
 * データをバッファリングし、flush()メソッドが呼び出されたときに
 * バッファリングしたデータを圧縮して送信します。
 * 通常のOutputStreamと違いflush()メソッドが呼び出されるまで
 * データがバッファリングされる点に注意してください。
 */
public class CompressedOutputStream extends OutputStream
{
	private Deflater def;
	private OutputStream out;
	private ResetableByteArrayOutputStream buf;
    private ResetableByteArrayOutputStream tout;
    private boolean mode_compress = true;
    
	private byte[] bytebuf;
	/**
	 * 指定した出力ストリームを利用してCompressedOutputStreamを作成します。
	 * CompressedOutputStream(out,1024)と同じ動作をします。
	 * @param out
	 * 出力ストリーム
	 */
	public CompressedOutputStream(OutputStream out)
	{
        this(out, 1024);
	}
	/**
	 * 指定した出力ストリームとバッファを利用してCompressedOutputStreamを作成します。
	 * バッファはflush()を呼び出されるまでデータを一時的に保存する場所です。
	 * サイズをオーバーする量を書き込んだときは一時的に拡張されます。
	 * @param out
	 * 出力ストリーム
	 * @param bufsize
	 * バッファサイズ
	 */
	public CompressedOutputStream(OutputStream out,int bufsize)
	{
		this.out = out;
		this.bytebuf = new byte[bufsize];
		this.def = new Deflater(Deflater.BEST_SPEED);
        this.buf = new ResetableByteArrayOutputStream(bufsize);
        this.tout = new ResetableByteArrayOutputStream(bufsize);
	}
	/**
	 * 出力ストリームのバッファに1バイトを書き込みます。
	 * この時点では出力ストリームには書き込まれません。
	 */
	public void write(int b) throws IOException
	{
        synchronized (buf)
        {
            if (buf.size() == Short.MAX_VALUE)
                flush();
            buf.write(b);
        }
	}
	
	/**
	 * バッファリングされたデータを圧縮して出力ストリームに出力します。
	 */
	public void flush() throws IOException
	{
        synchronized (buf)
        {
            byte[] org = buf.getOriginalDate();
            int size = buf.size();
            //TO-DO debug code
            //System.out.println(Thread.currentThread().getName() + ">" + size);
            def.reset();
            def.setInput(org,0,size);
            def.finish();
            int s = def.deflate(bytebuf);
            byte[] wbuf;
            if (def.finished())
            {
                wbuf = bytebuf;
            }else
            {
                tout.reset();
                tout.write(bytebuf,0,s);
                do
                {
                    int ds = def.deflate(bytebuf);
                    s += ds;
                    //System.out.println(ds);
                    tout.write(bytebuf,0,ds);
                }while(!(def.finished()));
                //wbuf = tout.toByteArray();
                wbuf = tout.getOriginalDate();
            }

            if (mode_compress ^ s < size)
            {
                short k = -1;
                out.write((k >>>  8) & 0xFF);
                out.write((k >>>  0) & 0xFF);
                mode_compress = !mode_compress;
            }
            
            if (!mode_compress)
            {
                short k = (short)size;
                out.write((k >>>  8) & 0xFF);
                out.write((k >>>  0) & 0xFF);
                out.write(org,0,size);
            }
            else
            {
                short k = (short)s;
                //System.out.println(">"+MD5Tool.makeString(wbuf));
                out.write((k >>>  8) & 0xFF);
                out.write((k >>>  0) & 0xFF);
                out.write(wbuf,0,k);    
            }
            out.flush();
            buf.reset();
            super.flush();
        }
	}
    @Override
    public void close() throws IOException
    {
        out.close();
    }
}
