/*
 * Copyright 2005 Takao Sumitomo. All rights reserved.
 * 
 * Redistribution and use in source and binary forms,
 * with or without modification, are permitted provided 
 * that the following conditions are met:
 *
 *   1. Redistributions of source code must retain the 
 *      above copyright notice, this list of conditions
 *      and the following disclaimer.
 *   2. Redistributions in binary form must reproduce 
 *      the above copyright notice, this list of conditions
 *      and the following disclaimer in the documentation
 *      and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software
 * and documentation are those of the authors and should not be
 * interpreted as representing official policies,
 * either expressed or implied.
 * 
 * @Author: Takao Sumitomo
 */

/*
 * 作成日: 2005/02/11
 */
package bakutan.game.deathmatch;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.GraphicsConfiguration;
import java.awt.Transparency;
import java.awt.geom.Ellipse2D;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.TreeSet;

import javax.sound.midi.InvalidMidiDataException;
import javax.sound.midi.MidiSystem;
import javax.sound.midi.Sequence;

import bakutan.common.BStage;
import bakutan.common.BUserInterface;
import bakutan.common.Body;
import bakutan.common.ClientBody;
import bakutan.common.DataManager;
import bakutan.common.ai.AIInterface;
import bakutan.common.graphics.CharacterImage;
import bakutan.common.graphics.CharacterImageBundle;
import bakutan.common.graphics.CharacterImageFD;
import bakutan.common.graphics.CharacterImageND;
import bakutan.common.math.BMath;
import bakutan.common.math.BShape;
import bakutan.common.math.BVector2f;
import bakutan.common.math.BVector4f;
import bakutan.common.util.SEBundle;
import bakutan.common.util.BSound;
import bakutan.common.util.BGMBundle;

/**
 * @author cattaka
 */
public class BStageDM extends BStage
{
    /**
     * BodyのタイプID、未使用。
     * @see Body#getType() getType()
     */
	public static final byte TYPE_UNKNOWN = 0;
    /**
     * BodyのタイプID、ユーザーの操作するキャラクタを表す。
     * @see Body#getType() getType()
     * @see BodyWall BodyWall
     */
	public static final byte TYPE_WALL1 = 1;
    /**
     * BodyのタイプID、ミサイルを表す。
     * @see Body#getType() getType()
     * @see BodyPlayer BodyMan
     */
	public static final byte TYPE_PLAYER1 = 2;
    /**
     * BodyのタイプID、爆風を表す。
     * @see Body#getType() getType()
     * @see BodyMissile BodyMissile
     */
	public static final byte TYPE_MISSILE1 = 3;
    /**
     * BodyのタイプID、アイテムを表す。
     * @see Body#getType() getType()
     * @see BodyFire BodyFile
     */
    public static final byte TYPE_FIRE1 = 4;
    /**
     * BodyのタイプID、柔らかい壁を表す。
     * @see Body#getType() getType()
     * @see BodyItem BodyItem
     */
    public static final byte TYPE_ITEM = 5;
    /**
     * BodyのタイプID、爆弾を表す。
     * @see Body#getType() getType()
     * @see BodySoftWall BodySoftWall
     */
    public static final byte TYPE_SOFTWALL = 7;
    /**
     * BodyのタイプID、スラグを表す。
     * @see Body#getType() getType()
     * @see BodyBomb BodyBom
     */
    public static final byte TYPE_BOMB1 = 8;
    /**
     * BodyのタイプID、スラグを表す。
     * @see Body#getType() getType()
     * @see BodySlug BodySlug
     */
    public static final byte TYPE_SLUG1 = 9;
    /**
     * BodyのタイプID、ホーミングを表す。
     * @see Body#getType() getType()
     * @see BodyHoming BodyHoming
     */
    public static final byte TYPE_HOMING1 = 10;
    /**
     * BodyのタイプID、クラスター爆弾を表す。
     * @see Body#getType() getType()
     * @see BodyClusterBomb BodyClusterBomb
     */
    public static final byte TYPE_CLUSTERBOMB = 11;

    public static final byte TYPE_EFFECT = 32;

    public static final byte SOUND_LOGIN = 0;
    public static final byte SOUND_LOGOUT = 1;
    public static final byte SOUND_CHAT = 2;
    public static final byte SOUND_DOWN = 3;
    public static final byte SOUND_ITEM = 4;
    public static final byte SOUND_FIRE = 5;
    public static final byte SOUND_MISSILE = 6;
	public static final byte SOUND_BOMB = 7;
    public static final byte SOUND_SLUG = 8;
    public static final byte SOUND_HOMING = 9;
    public static final byte SOUND_BARRAGE = 10;
    public static final byte SOUND_CLUSTERBOMB = 11;
    public static final byte SOUND_PHALANX = 12;
    public static final byte SOUND_AVENGER = 13;
    public static final byte SOUND_FIERCEATTACK = 14;
	
    private TreeSet<Body> sorter = new TreeSet<Body>(new BodyComparator());
	public static final float WALL_WIDTH = 0.50f;
	public static final float PLAYER_WIDTH = 0.3f;
    public static final float MISSILE_WIDTH = 0.2f;
    public static final float BOM_WIDTH = 0.5f;
    public static final float SLUG_WIDTH = 0.5f;
    public static final float HOMING_WIDTH = 0.3f;
	public static final float ITEM_WIDTH = 0.5f;
	public static final float FIRE_WIDTH = 0.5f;
	public static final float FIRE_WIDTH_INC = 0.1f;
    public static final float CLUSTERBOM_WIDTH = 0.4f;

	public static final long ITEM_SPAWN_INTERVAL = 10000;
	public static final long WALL_SPAWN_INTERVAL = 120000;
	public static final float ITEM_SPAWN_RATE = 0.25f;
	
    
	private BVector2f[] freePoint;
    
    @Override
    public String getGameName()
    {
        return "DM";
    }
    @Override
    public byte[] getBodyTypeArray()
    {
        return new byte[]{
                TYPE_WALL1,
                TYPE_PLAYER1,
                TYPE_MISSILE1,
                TYPE_FIRE1,
                TYPE_ITEM,
                TYPE_SOFTWALL,
                TYPE_BOMB1,
                TYPE_SLUG1,
                TYPE_HOMING1,
                TYPE_CLUSTERBOMB
                };
    }
    public AIInterface createDefaultAIInterface()
    {
        return new AIAlgorismicVer2();
    }
	public Body createBodyInstance(byte bodyType)
	{
        Body body;
		if (bodyType == TYPE_WALL1)
		{
			body = new BodyWall();
		}else if (bodyType == TYPE_PLAYER1)
		{
			body = new BodyPlayer();
		}else if (bodyType == TYPE_MISSILE1)
		{
			body = new BodyMissile();
		}else if (bodyType == TYPE_FIRE1)
		{
			body = new BodyFire();
		}else if (bodyType == TYPE_ITEM)
		{
			body = new BodyItem();
		}else if (bodyType == TYPE_SOFTWALL)
		{
			body = new BodySoftWall();
        }else if (bodyType == TYPE_BOMB1)
        {
            body = new BodyBomb();
        }else if (bodyType == TYPE_SLUG1)
        {
            body = new BodySlug();
        }else if (bodyType == TYPE_HOMING1)
        {
            body = new BodyHoming();
        }else if (bodyType == TYPE_CLUSTERBOMB)
        {
            body = new BodyClusterBomb();
		}else
		{
			throw new RuntimeException("Unknown Body type :" + bodyType);
		}
        return body;
	}
	public CharacterImageBundle createCharacterImageBundle(GraphicsConfiguration gc)
	{
		CharacterImageBundle characterImageBundle = new CharacterImageBundle();
		try
		{
			CharacterImage img_wall1 = new CharacterImageFD(gc, Transparency.BITMASK, getClass().getClassLoader().getResourceAsStream("image/wall1.png"));
            CharacterImage img_player1 = new CharacterImageND(gc, getClass().getClassLoader().getResourceAsStream("image/tank1.png"));
			CharacterImage img_missile1 = new CharacterImageND(gc, getClass().getClassLoader().getResourceAsStream("image/missile1.png"));
            CharacterImage img_bom1 = new CharacterImageND(gc, getClass().getClassLoader().getResourceAsStream("image/bom1.png"));
			CharacterImage img_fire1 = new CharacterImageND(gc, getClass().getClassLoader().getResourceAsStream("image/fire.png"));
			CharacterImage img_item = new CharacterImageFD(gc, Transparency.BITMASK, getClass().getClassLoader().getResourceAsStream("image/item_fire.png"));
			CharacterImage img_soft_wall = new CharacterImageFD(gc, Transparency.BITMASK, getClass().getClassLoader().getResourceAsStream("image/wall2.png"));
            CharacterImage img_slug1 = new CharacterImageND(gc, getClass().getClassLoader().getResourceAsStream("image/slug1.png"));
            CharacterImage img_homing1 = new CharacterImageND(gc, getClass().getClassLoader().getResourceAsStream("image/homing1.png"));
            CharacterImage img_effect = new CharacterImageFD(gc, Transparency.TRANSLUCENT, getClass().getClassLoader().getResourceAsStream("image/effect.png"));
			characterImageBundle.putCharacterImage(TYPE_WALL1, img_wall1);
			characterImageBundle.putCharacterImage(TYPE_PLAYER1, img_player1);
			characterImageBundle.putCharacterImage(TYPE_MISSILE1, img_missile1);
			characterImageBundle.putCharacterImage(TYPE_FIRE1, img_fire1);
			characterImageBundle.putCharacterImage(TYPE_ITEM, img_item);
			characterImageBundle.putCharacterImage(TYPE_SOFTWALL, img_soft_wall);
            characterImageBundle.putCharacterImage(TYPE_BOMB1, img_bom1);
            characterImageBundle.putCharacterImage(TYPE_SLUG1, img_slug1);
            characterImageBundle.putCharacterImage(TYPE_HOMING1, img_homing1);
            characterImageBundle.putCharacterImage(TYPE_EFFECT, img_effect);
            characterImageBundle.putCharacterImage(TYPE_CLUSTERBOMB, img_bom1);
			return characterImageBundle;
		}catch(IOException e)
		{
			e.printStackTrace();
			throw new RuntimeException("Can't load CharacterImages");
		}
	}
	public SEBundle createSEBundle()
	{
        ArrayList<URL> al = new ArrayList<URL>();
        al.add(getClass().getClassLoader().getResource("sound/login.wav"));
        al.add(getClass().getClassLoader().getResource("sound/logout.wav"));
        al.add(getClass().getClassLoader().getResource("sound/chat.wav"));
        al.add(getClass().getClassLoader().getResource("sound/down.wav"));
        al.add(getClass().getClassLoader().getResource("sound/item.wav"));
        al.add(getClass().getClassLoader().getResource("sound/fire.wav"));
        al.add(getClass().getClassLoader().getResource("sound/missile.wav"));
        al.add(getClass().getClassLoader().getResource("sound/bomb.wav"));
        al.add(getClass().getClassLoader().getResource("sound/slug.wav"));
        al.add(getClass().getClassLoader().getResource("sound/homing.wav"));
        al.add(getClass().getClassLoader().getResource("sound/barrage.wav"));
        al.add(getClass().getClassLoader().getResource("sound/clusterbomb.wav"));
        al.add(getClass().getClassLoader().getResource("sound/phalanx.wav"));
        al.add(getClass().getClassLoader().getResource("sound/avenger.wav"));
        al.add(getClass().getClassLoader().getResource("sound/fierceattack.wav"));
        
        SEBundle ad = new SEBundle();
        for (byte i = 0;i< al.size();i++)
        {
            URL url;
            url = al.get(i);
            if (url != null)
                ad.putAudioData(i,BSound.loadAudioFile(url));
        }
        return ad;
	}
	public BGMBundle createBGMBundle()
	{
		BGMBundle bgmBundle = new BGMBundle();
		ArrayList<URL> al = new ArrayList<URL>();
		al.add(getClass().getClassLoader().getResource("midi/maoutamsi-battle06.mid"));
		al.add(getClass().getClassLoader().getResource("midi/maoutamsi-battle08.mid"));
		al.add(getClass().getClassLoader().getResource("midi/maoutamsi-battle11.mid"));
		al.add(getClass().getClassLoader().getResource("midi/maoutamsi-battle14.mid"));
		al.add(getClass().getClassLoader().getResource("midi/maoutamsi-battle15.mid"));
		al.add(getClass().getClassLoader().getResource("midi/maoutamsi-battle17.mid"));
		al.add(getClass().getClassLoader().getResource("midi/maoutamsi-battle18.mid"));
		al.add(getClass().getClassLoader().getResource("midi/maoutamsi-battle21.mid"));
		al.add(getClass().getClassLoader().getResource("midi/maoutamsi-battle22.mid"));
		al.add(getClass().getClassLoader().getResource("midi/maoutamsi-battle25.mid"));
		al.add(getClass().getClassLoader().getResource("midi/maoutamsi-battle27.mid"));
		al.add(getClass().getClassLoader().getResource("midi/maoutamsi-battle28.mid"));
		al.add(getClass().getClassLoader().getResource("midi/maoutamsi-boss01.mid"));
		al.add(getClass().getClassLoader().getResource("midi/maoutamsi-boss03.mid"));
		al.add(getClass().getClassLoader().getResource("midi/maoutamsi-ivent-rock01.mid"));
		al.add(getClass().getClassLoader().getResource("midi/maoutamsi-ivent-rock03.mid"));
		al.add(getClass().getClassLoader().getResource("midi/maoutamsi-ivent-rock06.mid"));
		al.add(getClass().getClassLoader().getResource("midi/maoutamsi-ivent-rock08.mid"));
		al.add(getClass().getClassLoader().getResource("midi/maoutamsi-ivent-rock09.mid"));
		al.add(getClass().getClassLoader().getResource("midi/maoutamsi-ivent-rock14.mid"));
		al.add(getClass().getClassLoader().getResource("midi/maoutamsi-ivent-rock17.mid"));
		al.add(getClass().getClassLoader().getResource("midi/maoutamsi-syber01.mid"));
				
		for (byte i = 0;i< al.size();i++)
		{
			URL bgmfileurl = (URL)al.get(i);
			try
			{
                InputStream in = bgmfileurl.openStream();
				Sequence sequence = MidiSystem.getSequence(in);
                in.close();
				bgmBundle.putBGM(i, sequence);
			}catch(InvalidMidiDataException e)
			{
			}catch(IOException e)
			{
			}
		}
		return bgmBundle;
	}

    
    public byte getRemoteClientBodyType()
	{
		return TYPE_PLAYER1;
	}
	public ClientBody createClientBody()
	{
		return new BodyPlayer();
	}
    @Override
    public Body[] getBodysHasScore()
    {
        return getBodysByType(TYPE_PLAYER1);
    }
	@Override
	public boolean isRemoteClientBody(byte type)
	{
	    return type == TYPE_PLAYER1;
	}
	public void drawScore(Graphics bg, long currenttime, Dimension size, Body myBody)
	{
		final Font scoreFont = new Font("Serif",Font.BOLD,20);
		bg.setFont(scoreFont);
		Body[] bodys = getBodysHasScore();
		int x = DataManager.BLOCKSIZE * 4;
		int y = DataManager.BLOCKSIZE * 3;
		int s = bg.getFont().getSize();
		for (int i=0;i<bodys.length;i++)
		{
			if (bodys[i] == myBody)
				bg.setColor(Color.YELLOW);
			else
				bg.setColor(Color.WHITE);
				
			bg.drawString(bodys[i].getName(), x, y);
			bg.drawString(String.valueOf(bodys[i].getScore()), x + DataManager.BLOCKSIZE * 7, y);
			y += bg.getFont().getSize();
		}
	}
	public void drawStatus(Graphics bg, long currenttime, Dimension size, ClientBody myBody, BUserInterface ui)
	{
        final int blockSize = DataManager.BLOCKSIZE;
        final Font smallFont = new Font("Serif",Font.BOLD,blockSize / 2);
		final Font mediumFont = new Font("Serif",Font.BOLD,blockSize * 2 / 3);
        final Font largeFont = new Font("Serif",Font.BOLD,blockSize);
        final Color gaugeColor = new Color(1f,1f,1f,0.25f);
		int fire = myBody.getOption(BodyPlayer.OPTION_FIRE);
		int boms = myBody.getOption(BodyPlayer.OPTION_CHARGE);
		int speed = myBody.getOption(BodyPlayer.OPTION_SPEED);
        int armor = myBody.getArmor();
		int x = 0;//blockSize * 1;
		int y = size.height - blockSize / 4;
		int offset = 0;
		float energy = myBody.getEnergy();
        
        BodyPlayer player = (BodyPlayer)myBody;
        
		String str = "";
		
        bg.setFont(smallFont);
        //Fire
		if (fire >= BodyPlayer.MAX_FIRE)
			bg.setColor(Color.YELLOW);
		else
			bg.setColor(Color.WHITE);
        str = ui.translate("fire");
        offset = (blockSize * 3 - bg.getFontMetrics().stringWidth(str)) / 2;
		bg.drawString(str, x + offset , y);
        str = String.valueOf(fire);
        offset = (blockSize * 3 - bg.getFontMetrics().stringWidth(str)) / 2;
        bg.drawString(str, x + offset, y-blockSize/2);
		x += blockSize * 3;
        //Charge
		if (boms >= BodyPlayer.MAX_CHARGE)
			bg.setColor(Color.YELLOW);
		else
			bg.setColor(Color.WHITE);
        str = ui.translate("charge");
        offset = (blockSize * 3 - bg.getFontMetrics().stringWidth(str)) / 2;
        bg.drawString(str, x + offset, y);
        str = String.valueOf(boms);
        offset = (blockSize * 3 - bg.getFontMetrics().stringWidth(str)) / 2;
        bg.drawString(str, x + offset, y-blockSize/2);
		x += blockSize * 3;
        //Speed
		if (speed >= BodyPlayer.MAX_SPEED)
			bg.setColor(Color.YELLOW);
		else
			bg.setColor(Color.WHITE);
        str = ui.translate("speed");
        offset = (blockSize * 3 - bg.getFontMetrics().stringWidth(str)) / 2;
        bg.drawString(str, x + offset, y);
        str = String.valueOf(speed);
        offset = (blockSize * 3 - bg.getFontMetrics().stringWidth(str)) / 2;
        bg.drawString(str, x + offset, y-blockSize/2);
        x += blockSize * 4;

        //Armor
        bg.setFont(smallFont);
        bg.setColor(gaugeColor);
        int width_armor = (int)((float)(blockSize*2) * armor / BodyPlayer.DEFAULT_ARMOR);
        bg.fillRect(x+blockSize/2,y-blockSize,width_armor,blockSize/2);
        if (armor > 70)
            bg.setColor(Color.GREEN);
        else if (armor > 30)
            bg.setColor(Color.YELLOW);
        else
            bg.setColor(Color.RED);
        str = String.valueOf(armor);
        offset = (blockSize * 3 - bg.getFontMetrics().stringWidth(str)) / 2;
        bg.drawString(str, x + offset, y-blockSize/2);
        bg.setColor(Color.WHITE);
        str = ui.translate("armor");
        offset = (blockSize * 3 - bg.getFontMetrics().stringWidth(str)) / 2;
        bg.drawString(str, x + offset, y);
        x += blockSize * 3;

        //Energy
        bg.setFont(smallFont);
        bg.setColor(gaugeColor);
        int width_energy = (int)((float)(blockSize*2) * energy / myBody.getMaxEnergy());
        bg.fillRect(x+blockSize/2,y-blockSize,width_energy,blockSize/2);
        if (energy > 70)
            bg.setColor(Color.GREEN);
        else if (energy > 30)
            bg.setColor(Color.YELLOW);
        else
            bg.setColor(Color.RED);
        int int_energy = (int)energy;
        str = String.valueOf(int_energy);
        offset = (blockSize * 3 - bg.getFontMetrics().stringWidth(str)) / 2;
        bg.drawString(str, x + offset, y-blockSize/2);
        bg.setColor(Color.WHITE);
        str = ui.translate("energy");
        offset = (blockSize * 3 - bg.getFontMetrics().stringWidth(str)) / 2;
        bg.drawString(str, x + offset, y);
        if (player.showEnergyEmpty > currenttime)
        {
            int[][] ps = createXPolygon(x + blockSize * 3 /2, y-blockSize*2/3);
            bg.setColor(Color.RED);
            bg.fillPolygon(ps[0], ps[1], ps[0].length);
        }
        x += blockSize * 4;
        
        //Weapon
        bg.setColor(Color.WHITE);
        bg.setFont(smallFont);
        str = ui.translate("weapon");
        offset = (blockSize * 3 - bg.getFontMetrics().stringWidth(str)) / 2;
        bg.drawString(str, x + offset, y);
        byte actionMode = myBody.getActionMode();
        if (actionMode == 0)
            str = ui.translate("missile");
        else if (actionMode == 1)
            str = ui.translate("bomb");
        else if (actionMode == 2)
            str = ui.translate("slug");
        else if (actionMode == 3)
            str = ui.translate("homing");
        else if (actionMode == 4)
            str = ui.translate("fierceattack");
        else if (actionMode == 5)
            str = ui.translate("barrage");
        else if (actionMode == 6)
            str = ui.translate("clusterbomb");
        else if (actionMode == 7)
            str = ui.translate("phalanx");
        else if (actionMode == 8)
            str = ui.translate("avenger");
        offset = (blockSize * 3 - bg.getFontMetrics().stringWidth(str)) / 2;
        bg.drawString(str, x + offset, y-blockSize/2);
        x += blockSize * 4;

        //Morale
        bg.setFont(smallFont);
        bg.setColor(gaugeColor);
        int morale = myBody.getOption(BodyPlayer.OPTION_MORALE);
        int morale_mod = morale % 100;
        int morale_num = morale / 100;
        int width_morale = (int)((float)(blockSize*2) * morale_mod / 100f);
        if (morale >= 100)
        {
            bg.fillRect(x+blockSize/2,y-blockSize,blockSize*2,blockSize/2);
        }
        bg.fillRect(x+blockSize/2,y-blockSize,width_morale,blockSize/2);
        if (morale >= 100)
            bg.setColor(Color.BLUE);
        else
            bg.setColor(Color.YELLOW);
        str = String.valueOf(morale_num);
        offset = (blockSize * 3 - bg.getFontMetrics().stringWidth(str)) / 2;
        bg.drawString(str, x + offset, y-blockSize/2);
        bg.setColor(Color.WHITE);
        str = ui.translate("morale");
        offset = (blockSize * 3 - bg.getFontMetrics().stringWidth(str)) / 2;
        bg.drawString(str, x + offset, y);
        if (player.showMoraleEmpty > currenttime)
        {
            int[][] ps = createXPolygon(x + blockSize * 3 /2, y-blockSize*2/3);
            bg.setColor(Color.RED);
            bg.fillPolygon(ps[0], ps[1], ps[0].length);
        }
        x += blockSize * 4;
    }
    private int[][] createXPolygon(int px,int py)
    {
        int blocksize = DataManager.BLOCKSIZE;
        int[][] ps = new int[2][12];
        int w = 4;
        ps[0][0] = px - blocksize + w;
        ps[1][0] = py - blocksize/4; 
        ps[0][1] = px - blocksize;
        ps[1][1] = py - blocksize/4 + w; 
        ps[0][2] = px - w;
        ps[1][2] = py; 
        ps[0][3] = px - blocksize;
        ps[1][3] = py + blocksize/4 - w; 
        ps[0][4] = px - blocksize + w;
        ps[1][4] = py + blocksize/4; 
        ps[0][5] = px;
        ps[1][5] = py + w;
        ps[0][6] = px + blocksize - w;
        ps[1][6] = py + blocksize/4; 
        ps[0][7] = px + blocksize;
        ps[1][7] = py + blocksize/4 - w; 
        ps[0][8] = px + w;
        ps[1][8] = py;
        ps[0][9] = px + blocksize;
        ps[1][9] = py - blocksize/4 + w; 
        ps[0][10] = px + blocksize - w;
        ps[1][10] = py - blocksize/4; 
        ps[0][11] = px;
        ps[1][11] = py - w;
        
        return ps;
    }
	public void drawRadar(Graphics bg, long currenttime, Dimension size, Body myBody, BVector2f center)
	{
		Body[] bs = getBodys();
        BVector2f radarCenter;
        if (myBody != null)
            radarCenter = myBody.getPosition();
        else
            radarCenter = center;
        final float radius = DataManager.DEFAULT_HEIGHT * 2;
        final float scale = 3;
        final BShape radarShape = new BShape(radius);
//      final Color backColor = new Color(127, 127,  127, 32);
//		final Color myBodyColor = new Color(0, 127, 0,191);
//		final Color playerColor = new Color(0, 0, 255,191);
//		final Color bomColor = new Color(255,0,0,191);
//		final Color softWallColor = new Color(255,255,255,127);
//		final Color wallColor = new Color(255,255,255,191);
//		final Color itemColor = new Color(255,255,0,191);

        final Color backColor = new Color(127, 127,  127);
        final Color myBodyColor = new Color(0, 127, 0);
        final Color playerColor = new Color(0, 0, 255);
        final Color bomColor = new Color(255,0,0);
        final Color softWallColor = new Color(255,255,255);
        final Color wallColor = new Color(255,255,255);
        final Color itemColor = new Color(255,255,0);
        
        int r = (int)(radius * scale);
		int cx = size.width - r - DataManager.BLOCKSIZE;
		int cy = size.height - r - DataManager.BLOCKSIZE*2;
        bg.translate(cx,cy);
        final Ellipse2D clipShape = new Ellipse2D.Float(-r, -r, r*2, r*2);
        bg.setColor(backColor);
        bg.drawOval(-r, -r, r*2, r*2);
        bg.drawOval(-r/2, -r/2, r, r);
//        bg.drawLine(-r, 0, r, 0);
//        bg.drawLine( 0,-r, 0, r);
        bg.setClip(clipShape);
        BVector2f vec = new BVector2f();
		for(int i=0;i<bs.length;i++)
		{
            int ms = 8;
            Body body = bs[i];
			switch(body.getType())
			{
			case TYPE_PLAYER1:
				if (myBody != null && body == myBody)
					bg.setColor(myBodyColor);
				else
					bg.setColor(playerColor);
				break;
			case TYPE_ITEM:
//			    continue;
                ms = 4;
			    bg.setColor(itemColor);
				break;
			case TYPE_SOFTWALL:
//			    continue;
                ms = 4;
			    bg.setColor(softWallColor);
                break;
			case TYPE_WALL1:
//                if (body.getShape().getType() != BShape.LINE)
//                    continue;
                bg.setColor(wallColor);
                break;
			default:
                ms = 6;
                bg.setColor(bomColor);
			}
            if (!BMath.isIntersect(radarCenter, radarShape, body.getPosition(), body.getShape()))
                continue;
            vec.set(body.getPosition());
            vec.sub(radarCenter);

            BShape shape = body.getShape();
            int x = (int)((vec.x)*scale); 
            int y = (int)((vec.y)*scale);
            if (shape.getType() == BShape.CIRCLE)
            {
                bg.fillOval(x-ms/2,y-ms/2,ms, ms);
            }
            else if (shape.getType() == BShape.LINE)
            {
                int px1 = x + (int)((shape.getPoint1().x) * scale); 
                int py1 = y + (int)((shape.getPoint1().y) * scale);
                int px2 = x + (int)((shape.getPoint2().x) * scale); 
                int py2 = y + (int)((shape.getPoint2().y) * scale);
                bg.drawLine(px1, py1, px2, py2);
            }
            else if (shape.getType() == BShape.RECTANGLE)
            {
                BVector4f rec = shape.getRectangle();
                int x1 = x + (int)((rec.a) * scale); 
                int y1 = y + (int)((rec.c) * scale);
                int w = (int)((rec.b-rec.a) * scale); 
                int h = (int)((rec.d-rec.c) * scale);
                bg.fillRect(x1, y1, w, h);
            }
        }
        bg.setClip(null);
        bg.translate(-cx,-cy);
	}
	
	/*
	 * -------------------------------------------------------
	 * static methods
	 */
	public Body[] getSortedBodys()
	{
        sorter.clear();
        Iterator<Body> ite = getBodysIterator();
        while (ite.hasNext())
        {
            sorter.add(ite.next());
        }
        Body[] bs = sorter.toArray(new Body[sorter.size()]);
		return bs;
	}

	public void loadBStageFile(URL url)
	throws IOException
	{
		char[][] data;

		data = getChars(url);
			
		float h = data.length;
		float w = 0;
		for (int i=0;i<h;i++)
		{
			if (w < data[i].length)
				w = data[i].length;
		}
		
		setSize(new BVector2f(w, h));
		
		BStageDM bstage = this;
		//bstage.setEventHandler(eventHandler);
		
		ArrayList<BVector2f> spawnPoints = new ArrayList<BVector2f>();
		ArrayList<BVector2f> freePoints = new ArrayList<BVector2f>();
		for (int i=1;i<data.length-1;i++)
		{
			for (int j=1;j<data[i].length-1;j++)
			{
				if (data[i][j] == ' ')
				{
					BVector2f pos = new BVector2f();
					pos.x = ((float)j) + 0.5f;
					pos.y = ((float)i) + 0.5f;
					freePoints.add(pos);
				}
				else if (data[i][j] == 'W')
				{
					Body body = createBodyInstance(TYPE_WALL1);
					body.setID(bstage.getNextID());
					body.getPosition().x = ((float)j) + 0.5f;
					body.getPosition().y = ((float)i) + 0.5f;
					bstage.addBody(body);
				}
				else if (data[i][j] == 'S')
				{
					Body body = createBodyInstance(TYPE_SOFTWALL);
					body.setID(bstage.getNextID());
					body.getPosition().x = ((float)j) + 0.5f;
					body.getPosition().y = ((float)i) + 0.5f;
					bstage.addBody(body);

					BVector2f pos = new BVector2f();
					pos.x = ((float)j) + 0.5f;
					pos.y = ((float)i) + 0.5f;
					freePoints.add(pos);
				}
				else if (data[i][j] == 's')
				{
					BVector2f pos = new BVector2f();
					pos.x = ((float)j) + 0.5f;
					pos.y = ((float)i) + 0.5f;
					spawnPoints.add(pos);
				}else if (data[i][j] == 'I')
				{
					BodyItem body = (BodyItem)createBodyInstance(TYPE_ITEM);
					body.setID(bstage.getNextID());
					body.selectRundom();
					body.getPosition().x = ((float)j) + 0.5f;
					body.getPosition().y = ((float)i) + 0.5f;
					body.setRespawnable(true);
					bstage.addBody(body);
				}
			}
		}
        BVector4f rect = new BVector4f();
        Body wallN = createBodyInstance(TYPE_WALL1);
        wallN.setID(bstage.getNextID());
        wallN.getPosition().set( w/2, 0.5f);
        wallN.getShape().setType(BShape.RECTANGLE);
        wallN.getShape().getRectangle().set(-w/2f, w/2f, -0.5f, 0.5f);
        bstage.addBody(wallN);
        Body wallS = createBodyInstance(TYPE_WALL1);
        wallS.setID(bstage.getNextID());
        wallS.getPosition().set( w/2, h-0.5f);
        wallS.getShape().setType(BShape.RECTANGLE);
        wallS.getShape().getRectangle().set(-w/2f, w/2f, -0.5f, 0.5f);
        bstage.addBody(wallS);
        Body wallW = createBodyInstance(TYPE_WALL1);
        wallW.setID(bstage.getNextID());
        wallW.getPosition().set( 0.5f, h/2);
        wallW.getShape().setType(BShape.RECTANGLE);
        wallW.getShape().getRectangle().set(-0.5f, 0.5f, -h/2f, h/2f);
        bstage.addBody(wallW);
        Body wallE = createBodyInstance(TYPE_WALL1);
        wallE.setID(bstage.getNextID());
        wallE.getPosition().set( w-0.5f, h/2);
        wallE.getShape().setType(BShape.RECTANGLE);
        wallE.getShape().getRectangle().set(-0.5f, 0.5f, -h/2f, h/2f);
        bstage.addBody(wallE);
        

        BVector2f[] sps = new BVector2f[spawnPoints.size()];
		System.arraycopy(spawnPoints.toArray(),0,sps,0,sps.length);
		bstage.setSpawnPoint(sps);
		BVector2f[] fps = new BVector2f[freePoints.size()];
		System.arraycopy(freePoints.toArray(),0,fps,0,fps.length);
		bstage.setFreePoint(fps);
	}
	/*
	private static BStage createBStage(byte[][] data, EventHandler eventHandler)
	{
		int w = data.length;
		int h = 0;
		for (int i=0;i<w;i++)
		{
			if (h > data[i].length)
				h = data[i].length;
		}
		
		BStageDeathMatch bstage = new BStageDeathMatch();
		bstage.setEventHandler(eventHandler);
		
		ArrayList spawnPoints = new ArrayList();
		ArrayList freePoints = new ArrayList();
		for (int i=0;i<data.length;i++)
		{
			for (int j=0;j<data[i].length;j++)
			{
				if (data[i][j] == 0)
				{
					BVector2d pos = new BVector2d();
					pos.x = ((float)j) + 0.5;
					pos.y = ((float)i) + 0.5;
					freePoints.add(pos);
				}
				else if (data[i][j] == 1)
				{
					Body body = new BodyWall();
					body.setID(bstage.getNextID());
					body.getPosition().x = ((float)j) + 0.5;
					body.getPosition().y = ((float)i) + 0.5;
					bstage.addBody(body);
				}
				else if (data[i][j] == 2)
				{
					Body body = new BodySoftWall();
					body.setID(bstage.getNextID());
					body.getPosition().x = ((float)j) + 0.5;
					body.getPosition().y = ((float)i) + 0.5;
					bstage.addBody(body);

					BVector2d pos = new BVector2d();
					pos.x = ((float)j) + 0.5;
					pos.y = ((float)i) + 0.5;
					freePoints.add(pos);
				}
				else if (data[i][j] == 3)
				{
					BVector2d pos = new BVector2d();
					pos.x = ((float)j) + 0.5;
					pos.y = ((float)i) + 0.5;
					spawnPoints.add(pos);
				}else if (data[i][j] == 4)
				{
					//BodyItemSpawner body = new BodyItemSpawner();
					BodyItem body = new BodyItem();
					body.setID(bstage.getNextID());
					body.getPosition().x = ((float)j) + 0.5;
					body.getPosition().y = ((float)i) + 0.5;
					body.setRespawnable(true);
					bstage.addBody(body);
					//long currenttime = bstage.getCurrentTime();
					//bstage.getEventHandler().action(currenttime, currenttime, body, body.getPosition(), BodyItemSpawner.ACTION_SPAWN_SOMETHING);
				}
			}
		}
		BVector2d[] sps = new BVector2d[spawnPoints.size()];
		System.arraycopy(spawnPoints.toArray(),0,sps,0,sps.length);
		bstage.setSpawnPoint(sps);
		BVector2d[] fps = new BVector2d[freePoints.size()];
		System.arraycopy(freePoints.toArray(),0,fps,0,fps.length);
		bstage.setFreePoint(fps);
		return bstage;
	}
	*/
//	private static byte[][] getBytes(URL mapfile)
//	throws IOException
//	{
//		BufferedReader in = new BufferedReader(new InputStreamReader(mapfile.openStream()));
//		String tmp;
//		ArrayList<byte[]> al = new ArrayList<byte[]>();
//		while((tmp = in.readLine()) != null)
//		{
//			byte[] bs = new byte[tmp.length()];
//			for (int i=0;i<tmp.length();i++)
//			{
//				try
//				{
//					bs[i] = Byte.parseByte(tmp.substring(i,i+1));
//				}catch(NumberFormatException e)
//				{
//					bs[i] = -1;
//				}
//			}
//			al.add(bs);
//		}
//		in.close();
//		byte[][] bss = new byte[al.size()][];
//		for (int i=0;i<bss.length;i++)
//		{
//			byte[] bs = (byte[])al.get(i);
//			bss[i] = bs;
//		}
//		return bss;
//	}
    private static char[][] getChars(URL mapfile)
    throws IOException
    {
        BufferedReader in = new BufferedReader(new InputStreamReader(mapfile.openStream()));
        String tmp;
        ArrayList<char[]> al = new ArrayList<char[]>();
        while((tmp = in.readLine()) != null)
        {
            char[] cs = tmp.toCharArray();
            al.add(cs);
        }
        in.close();
        char[][] css = new char[al.size()][];
        for (int i=0;i<css.length;i++)
        {
            char[] cs = al.get(i);
            css[i] = cs;
        }
        return css;
    }
	public BVector2f getRandomeFreePoint()
	{
		return getFreePoint()[(int)(Math.random() * getFreePoint().length)];
	}
	public BVector2f[] getFreePoint()
	{
		return freePoint;
	}
	public void setFreePoint(BVector2f[] freePoint)
	{
		this.freePoint = freePoint;
	}
//	public void eventClientDead(long currenttime, Body body)
//	{
//		getEventHandler().state(currenttime, currenttime + 3000, body, Body.DISABLE, Body.STATE_LIFE | Body.STATE_DEAD);
//		getEventHandler().armor(currenttime,currenttime + 3000, body, body, BodyArmorEvent.SET, BodyPlayer.DEFAULT_ARMOR);
//		getEventHandler().state(currenttime, currenttime + 3000, body, Body.ENABLE, Body.STATE_MARKER);
//		getEventHandler().state(currenttime, currenttime + 6000, body, Body.DISABLE, Body.STATE_MARKER);
//		getEventHandler().state(currenttime, currenttime + 6000, body, Body.ENABLE, Body.STATE_LIFE);
//
//		getEventHandler().option(currenttime, currenttime + 3000, body, BodyPlayer.OPTION_FIRE, 0);
//		getEventHandler().option(currenttime, currenttime + 3000, body, BodyPlayer.OPTION_SPEED, 0);
//		getEventHandler().option(currenttime, currenttime + 3000, body, BodyPlayer.OPTION_CHARGE, 0);
//		getEventHandler().option(currenttime, currenttime + 3000, body, BodyPlayer.OPTION_WEAPON, 0);
//		
//		Body[] bs = getBodysHasScore();
//		int[] scores = new int[bs.length];
//		for (int i=0;i<bs.length;i++)
//		{
//			if (bs[i] == body)
//			{
//			    scores[i] = - bs.length + 1;
//			}
//			else
//			{
//                scores[i] = 1;
//			}
//		}
//		getEventHandler().scoreChange(currenttime,currenttime, body, bs, scores);
//	}
    public int getClientActionHook(Body myBody, int action)
    {
        if (action == BodyPlayer.ACTION_BARRAGE
                || action == BodyPlayer.ACTION_CLUSTERBOMB
                || action == BodyPlayer.ACTION_PHALANX
                || action == BodyPlayer.ACTION_AVENGER)
        {
            if (!myBody.isState(BodyPlayer.STATE_FIERCEATTACK))
                return BodyPlayer.ACTION_FIERCEATTACK;
        }
        return action;
    }
}
