/*
 * Copyright 2005 Takao Sumitomo. All rights reserved.
 * 
 * Redistribution and use in source and binary forms,
 * with or without modification, are permitted provided 
 * that the following conditions are met:
 *
 *   1. Redistributions of source code must retain the 
 *      above copyright notice, this list of conditions
 *      and the following disclaimer.
 *   2. Redistributions in binary form must reproduce 
 *      the above copyright notice, this list of conditions
 *      and the following disclaimer in the documentation
 *      and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software
 * and documentation are those of the authors and should not be
 * interpreted as representing official policies,
 * either expressed or implied.
 * 
 * @Author: Takao Sumitomo
 */

/*
 * 作成日: 2005/02/07
 */
package bakutan.game.deathmatch;

import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.geom.AffineTransform;

import bakutan.common.BStage;
import bakutan.common.BUserInterface;
import bakutan.common.Body;
import bakutan.common.ClientBody;
import bakutan.common.DataManager;
import bakutan.common.EventHandler;
import bakutan.common.event.BodyActionEvent;
import bakutan.common.event.BodyEvent;
import bakutan.common.event.BodyArmorEvent;
import bakutan.common.graphics.CharacterImage;
import bakutan.common.graphics.CharacterImageBundle;
import bakutan.common.graphics.EffectBundle;
import bakutan.common.math.BShape;
import bakutan.common.math.BVector2f;

/**
 * @author cattaka
 */
public class BodyHoming extends Body
{
    /**
     * 火力のオプションを得るための定数。
     * getOption(int)に使用する。
     */
    public static final byte OPTION_FIRE = 0;
    /**
     * ホーミングの軌道修正の間隔(ミリ秒)
     */
	public static final int AIM_INTERVAL = 400;
    /**
     * ホーミングの一回の軌道修正の角度の最大値(ラジアン値)
     */
    public static final float AIM_ANGLE = (float)(Math.PI/9d);
    /**
     * ホーミングの一回の軌道修正の補正値(ラジアン値)
     * 火力最大で40度になる。
     */
    public static final float AIM_ANGLE_INCREASE = (float)(Math.PI/9d/(BodyPlayer.DEFAULT_FIRE + BodyPlayer.MAX_FIRE));
    /**
     * 目標を捕捉できる角度(ラジアン値)。
     */
    public static final float SEARCH_ANGLE = (float)(Math.PI/3d);
    /**
     * 目標を捕捉できる角度の補正値(ラジアン値)。
     * 火力最大で90度になる。
     */
    public static final float SEARCH_ANGLE_INCREASE = (float)(Math.PI/6d/(BodyPlayer.DEFAULT_FIRE + BodyPlayer.MAX_FIRE));
    /**
     * ホーミングの飛翔時間を表す定数、単位はミリ秒。
     */
    public static final int LIFE_TIME = 10000;
    /**
     * ホーミング直撃時のダメージ
     */
    public static final int DAMAGE = 20;
    /**
     * 爆風のダメージ
     */
    public static final int FIRE_DAMAGE = 20;
    /**
     * 爆風の残る時間の係数
     */
    public static final int FIRE_DERAY = 200;

    public static final byte ACTION_BOMB = 0;
    public static final byte ACTION_AIM = 1;
    private Body owner;
    private Body target = null;
	
    private long lastSmokeTime = 0l;
    
	public BodyHoming()
	{
        super(BStageDM.TYPE_HOMING1, 1);
//        setType(BStageThrowBom.TYPE_HOMING1);
//		setOptions(new int[1]);
		setOption(BodyHoming.OPTION_FIRE, getOption(OPTION_FIRE));
		setShape(new BShape(BStageDM.HOMING_WIDTH));
		//enableState(Body.COLLIDABLE);
		enableState(Body.STATE_LIFE);
	}
    public void drawCharacter(Graphics g, long currenttime, CharacterImageBundle cib, EffectBundle eb)
    {
        CharacterImage characterImage = getCharacterImage();
        if (characterImage == null)
            return;
        
        int offsetx = characterImage.getWidth() / 2;
        int offsety = characterImage.getHeight() / 2;
//        int x = -offset.width + (int) (getPosition().x * DataManager.BLOCKSIZE);
//        int y = -offset.height + (int) (getPosition().y * DataManager.BLOCKSIZE);
        int x = 0;
        int y = 0;
        int n = ((int)(currenttime - getLastStateChange()))/DataManager.ANIMINTERVAL;

        Image img = characterImage.getImages(CharacterImage.DOWN)[n%2];

        BVector2f v = new BVector2f(getDirection());
        v.normalize();
        float l = v.getLength();
        double rad;
        if (l > 0)
        {
            rad = Math.acos(v.y/l);
            if (v.x > 0)
                rad = 2 * Math.PI - rad;
        }else
        {
            rad = 0;
        }

        AffineTransform at = new AffineTransform();
        at.setToTranslation(x,y);
        at.rotate(rad);
        at.translate(-offsetx,-offsety);
        
        Graphics2D g2 = (Graphics2D)g;
        g2.drawImage(img, at, null);
        
        if (currenttime - lastSmokeTime >= 200)
        {
            lastSmokeTime = currenttime;
            BVector2f vec = new BVector2f(getVelocity());
            vec.scale(0.5f);
            EffectSmoke smoke = new EffectSmoke(getPosition(), vec, currenttime, cib.getCharacterImage(BStageDM.TYPE_EFFECT));
            eb.putPreEffect(smoke);
        }
    }
	public boolean hasName()
	{
		return false;
	}
	public boolean hasVariableBCircle()
	{
		return false;
	}
	public boolean hasScore()
	{
		return false;
	}
    @Override
    public boolean hasState()
    {
        return false;
    }
    @Override
    public boolean hasArmor()
    {
        return false;
    }
    @Override
    public boolean hasOption()
    {
        return false;
    }
    @Override
    public boolean hasArmorReport()
    {
        return false;
    }
	public void actionForClient(BodyEvent bodyEvent, ClientBody mybody, CharacterImageBundle cib, EffectBundle eb, BUserInterface ui, long currenttime, BVector2f bufferCenter)
	{
        if (bodyEvent instanceof BodyActionEvent)
        {
            BodyActionEvent bae = (BodyActionEvent)bodyEvent;
            if (bae.getAction() == ACTION_BOMB)
            {
                ui.getBAudio().playSE(BStageDM.SOUND_FIRE, 0, bufferCenter, getPosition());
            }
        }
	}
	public boolean eventAction(EventHandler eventHandler, BStage bstage, BVector2f position, BodyActionEvent event)
	{
        long currenttime = bstage.getCurrentTime();
		if (event.getAction() == ACTION_BOMB)
		{
			BVector2f p = new BVector2f(getPosition());
			//p.x = 0.5 + (int)(p.x);
			//p.y = 0.5 + (int)(p.y);
			
			BodyFire fire = new BodyFire();
            fire.setOwner(getOwner());
//			float length = BStageThrowBom.FIRE_WIDTH + BStageThrowBom.FIRE_WIDTH_INC * (float)getOption(OPTION_FIRE);
			fire.setOption(BodyFire.OPTION_SIZE, 0);
//            fire.setPolygon(createBpolygon(length));
            fire.setShape(new BShape(BodyFire.calcRadius(0)));
            fire.setDamage(FIRE_DAMAGE);
			fire.setPosition(p);
			fire.setVelocity(getVelocity());
			float speed = fire.getVelocity().getLength();
            if (speed > 0 && !fire.getVelocity().isZeroVector())
            {
                fire.getVelocity().normalize();
                fire.getVelocity().scale(speed/4);
            }
            
			eventHandler.delete(currenttime, currenttime,this);
			//eventHandler.sound(currenttime,currenttime,BStageThrowBom.SOUND_BOMB);
			eventHandler.add(currenttime, currenttime, fire);

            eventHandler.delete(currenttime, currenttime + FIRE_DERAY, fire);
            return true;
		}
        else if (event.getAction() == ACTION_AIM)
        {
            //eventHandler.action(currenttime, currenttime+AIM_INTERVAL, this, getPosition(), ACTION_AIM);
            eventHandler.smallAction(currenttime, currenttime+AIM_INTERVAL, this, ACTION_AIM);
            float search_angle = SEARCH_ANGLE + SEARCH_ANGLE_INCREASE * getOption(OPTION_FIRE);
            float aim_angle = AIM_ANGLE + AIM_ANGLE_INCREASE * getOption(OPTION_FIRE);
            if (target != null)
            {
                BVector2f vec = new BVector2f(target.getPosition());
                vec.sub(getPosition());
                if (!vec.isZeroVector())
                {
                    double angle = BVector2f.getAngle(vec, getVelocity());
                    if (Math.abs(angle) > search_angle)
                        target = null;
                }
            }
            if (target == null)
            {
                Body[] bs = bstage.getBodysHasScore();
                BVector2f vec = new BVector2f();
                Body selected = null;
                for (int i=0;i<bs.length;i++)
                {
                    if (bs[i] == owner)
                        continue;
                    
                    vec.set(bs[i].getPosition());
                    vec.sub(getPosition());
                    double angle = BVector2f.getAngle(getVelocity(), vec);
                    if (Math.abs(angle) > search_angle)
                    {
                        continue;
                    }
                    if (selected == null)
                    {
                        selected = bs[i];
                    }
                    else
                    {
                        if (BVector2f.getDistance(getPosition(),selected.getPosition())
                                > BVector2f.getDistance(getPosition(), bs[i].getPosition()))
                        {
                            selected = bs[i];
                        }
                    }
                }
                if (selected == null)
                    return false;
                
                target = selected;
            }
            BVector2f v = new BVector2f();
            v.set(target.getPosition());
            v.sub(getPosition());
            BVector2f s = new BVector2f();
            s.set(target.getVelocity());
            s.sub(getVelocity());

            float t = v.getLength() / s.getLength();
            BVector2f vec = new BVector2f();
            vec.set(target.getPosition());
            if (!Float.isNaN(t))
            {
                vec.scaleAdd(t, target.getVelocity());
            }
            
            vec.sub(getPosition());
            double angle = BVector2f.getAngle(getVelocity(), vec);
            if (Double.isNaN(angle))
            {
                return false;
            }
            
            //System.out.println(angle);
            if (angle > aim_angle)
            {
                angle = aim_angle;
            }
            else if (angle < -aim_angle)
            {
                angle = -aim_angle;
            }
            vec.set(getVelocity());
            vec.rotate(angle);
            BVector2f dir = new BVector2f(vec);
            dir.normalize();
            eventHandler.move(currenttime, currenttime, this, getPosition(), vec, dir);
            return true;
        }
        return false;
	}
	public void eventCollisition(EventHandler eventHandler, BStage bstage, Body target)
	{
		if (target.getType() == BStageDM.TYPE_HOMING1)
		{
			BodyHoming bom = (BodyHoming)target;
			if (getOwner() == bom.getOwner())
				return;
		}
		if (target != owner)
		{
			long currenttime = bstage.getCurrentTime();
            if (target.isState(Body.STATE_LIFE))
            {
                eventHandler.armor(currenttime, currenttime, target, this.getOwner(), BodyArmorEvent.SUB, DAMAGE);
            }
            //eventHandler.action(currenttime,currenttime,this,getPosition(),ACTION_BOMB);
            eventHandler.smallAction(currenttime,currenttime,this,ACTION_BOMB);
		}
	}
	
	public void eventStateChange(EventHandler eventHandler, BStage bstage, byte type, int state)
	{
		if ((type == Body.SET || type == Body.ENABLE)
				&& (state & Body.STATE_DEAD) == Body.STATE_DEAD)
		{
			long currenttime = bstage.getCurrentTime();
			//eventHandler.action(currenttime,currenttime,this,getPosition(),ACTION_BOMB);
            eventHandler.smallAction(currenttime,currenttime,this,ACTION_BOMB);
		}
	}
	public void eventPush(EventHandler eventHandler, BStage bstage, Body target)
	{
		if (target.getType() == BStageDM.TYPE_PLAYER1
				&& (target.getOption(BodyPlayer.OPTION_WEAPON) & BodyPlayer.WEAPON_KICK) == BodyPlayer.WEAPON_KICK )
		{
			long currenttime = bstage.getCurrentTime();
			BVector2f v = new BVector2f(getPosition());
			v.sub(target.getPosition());
//            v.x -= target.getPosition().x;
//			v.y -= target.getPosition().y;
//			if (Math.abs(v.x) >= Math.abs(v.y))
//			{
//				v.y = 0;
//			}else
//			{
//				v.x = 0;
//			}
			if (!v.isZeroVector())
            {
                v.normalize();
                v.scale(target.getMaxSpeed());
                eventHandler.move(currenttime, currenttime, this, this.getPosition(), v, v);
            }
		}
	}
    @Override
    public byte[] getCheckCollidType()
    {
        return new byte[]{
                BStageDM.TYPE_WALL1,
                BStageDM.TYPE_SOFTWALL
                };
    }
    @Override
    public byte[] getCheckIntersectType()
    {
        return new byte[]{
                BStageDM.TYPE_WALL1,
                BStageDM.TYPE_PLAYER1,
                BStageDM.TYPE_MISSILE1,
                BStageDM.TYPE_FIRE1,
                BStageDM.TYPE_SOFTWALL,
                BStageDM.TYPE_BOMB1,
                BStageDM.TYPE_SLUG1,
                BStageDM.TYPE_HOMING1,
                BStageDM.TYPE_CLUSTERBOMB
                };
    }
    public boolean isCheckCollidWith(Body target)
    {
        return false;
    }
	public boolean isCheckIntersectWith(Body target)
	{
		if (target.getType() == BStageDM.TYPE_ITEM)
			return false;
		return true;
	}
	public Body getOwner()
	{
		return owner;
	}
	public void setOwner(Body owner)
	{
		this.owner = owner;
	}
    public void setTarget(Body target)
    {
        this.target = target;
    }
    
}
