/*
 * Copyright 2005 Takao Sumitomo. All rights reserved.
 * 
 * Redistribution and use in source and binary forms,
 * with or without modification, are permitted provided 
 * that the following conditions are met:
 *
 *   1. Redistributions of source code must retain the 
 *      above copyright notice, this list of conditions
 *      and the following disclaimer.
 *   2. Redistributions in binary form must reproduce 
 *      the above copyright notice, this list of conditions
 *      and the following disclaimer in the documentation
 *      and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software
 * and documentation are those of the authors and should not be
 * interpreted as representing official policies,
 * either expressed or implied.
 * 
 * @Author: Takao Sumitomo
 */

/*
 * 作成日: 2005/02/07
 */
package bakutan.game.deathmatch;

import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.geom.AffineTransform;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.Date;

import bakutan.common.BStage;
import bakutan.common.BUserInterface;
import bakutan.common.Body;
import bakutan.common.ClientBody;
import bakutan.common.DataManager;
import bakutan.common.EventHandler;
import bakutan.common.event.BodyActionEvent;
import bakutan.common.event.BodyAddEvent;
import bakutan.common.event.BodyArmorEvent;
import bakutan.common.event.BodyDeleteEvent;
import bakutan.common.event.BodyEvent;
import bakutan.common.event.BodyMessageEvent;
import bakutan.common.event.BodyOptionEvent;
import bakutan.common.event.BodyStateEvent;
import bakutan.common.graphics.CharacterImage;
import bakutan.common.graphics.CharacterImageBundle;
import bakutan.common.graphics.EffectBundle;
import bakutan.common.math.BShape;
import bakutan.common.math.BVector2f;

/**
 * @author cattaka
 */
public class BodyPlayer extends ClientBody
{
    /**
     * 火力のオプションを得るための定数。
     * getOption(int)に使用する。
     */
	public static final byte OPTION_FIRE = 0;
    /**
     * 爆弾数のオプションを得るための定数。
     * getOption(int)に使用する。
     */
	public static final byte OPTION_CHARGE = 1;
    /**
     * スピードのオプションを得るための定数。
     * getOption(int)に使用する。
     */
	public static final byte OPTION_SPEED = 2;
    /**
     * 武器のオプションを得るための定数(未使用)。
     * getOption(int)に使用する。
     */
	public static final byte OPTION_WEAPON = 3;
    /**
     * 士気のオプションを得るための定数。
     * getOption(int)に使用する。
     */
    public static final byte OPTION_MORALE = 4;

    /**
     * 強襲モードを表す定数。
     * isState(int)メソッドに使用する。
     */
    public static final int STATE_FIERCEATTACK = 1<<16;
    /**
     * 標準の火力
     */
	public static final int DEFAULT_FIRE = 1;
    /**
     * 標準の爆弾数
     */
	public static final int DEFAULT_CHARGE = 5;
    /**
     * 標準のスピード
     */
	public static final int DEFAULT_SPEED = 16;

    /**
     * 火力の最大値
     */
	public static final int MAX_FIRE = 5;
    /**
     * 爆弾数の最大値
     */
	public static final int MAX_CHARGE = 5;
    /**
     * スピードの最大値
     */
	public static final int MAX_SPEED = 5;
    /**
     * 士気の最大値
     */
    public static final int MAX_MORALE = 1000;
    
    /**
     * 標準の装甲
     */
    public static final int DEFAULT_ARMOR = 100;

    /**
     * 最小ミサイル発射間隔
     */
    public static final int ACTIONLOCK_MISSILE = 200;
    /**
     * 最小爆弾発射間隔
     */
    public static final int ACTIONLOCK_BOMB = 200;
    /**
     * 最小スラグ発射間隔
     */
    public static final int ACTIONLOCK_SLUG = 500;
    /**
     * 最小ホーミング発射間隔
     */
    public static final int ACTIONLOCK_HOMING = 500;
    /**
     * 猛攻モード発動間隔
     */
    public static final int ACTIONLOCK_FIERCEATTACK = 1000;
    /**
     * 最小バラージ間隔
     */
    public static final int ACTIONLOCK_BARRAGE = 1000;
    /**
     * 最小クラスター間隔
     */
    public static final int ACTIONLOCK_CLUSTERBOMB = 1000;
    /**
     * 最小ファランクス間隔
     */
    public static final int ACTIONLOCK_PHALANX = 1000;
    /**
     * 最小アベンジャー間隔
     */
    public static final int ACTIONLOCK_AVENGER = 1000;
    /**
     * エネルギーの最大値
     */
    public static final float MAX_ENERGY = 100f;
    
    public static final float CONSUMPTION_MISSILE = 0.1f * (float)ACTIONLOCK_MISSILE;
    public static final float CONSUMPTION_BOMB = 0.1f * (float)ACTIONLOCK_BOMB;
    public static final float CONSUMPTION_SLUG = 0.1f * (float)ACTIONLOCK_SLUG;
    public static final float CONSUMPTION_HOMING = 0.1f * (float)ACTIONLOCK_HOMING;
    public static final float CONSUMPTION_FIERCEATTACK = 0f;
    public static final float CONSUMPTION_BARRAGE = MAX_ENERGY/2;
    public static final float CONSUMPTION_CLUSTERBOMB = MAX_ENERGY/2;
    public static final float CONSUMPTION_PHALANX = MAX_ENERGY/2;
    public static final float CONSUMPTION_AVENGER = MAX_ENERGY/2;
    
    public static final byte ACTION_MISSILE = 0;
    public static final byte ACTION_BOMB = 1;
    public static final byte ACTION_SLUG = 2;
    public static final byte ACTION_HOMING = 3;
    public static final byte ACTION_FIERCEATTACK = 4;
    public static final byte ACTION_BARRAGE = 5;
    public static final byte ACTION_CLUSTERBOMB = 6;
    public static final byte ACTION_PHALANX = 7;
    public static final byte ACTION_AVENGER = 8;
    public static final int WEAPON_KICK = 1;
    
    public static final float SPEEDSCALE_MISSILE = 1.8f;
    public static final float SPEEDSCALE_BOMB = 1.3f;
    public static final float SPEEDSCALE_SLUG = 1.4f;
    public static final float SPEEDSCALE_HOMING = 1.5f;
	
    long showEnergyEmpty = 0l;
    long showMoraleEmpty = 0l;

    private byte actionMode = 0;
    protected Body lastAttacked = this;
    
	public BodyPlayer()
	{
        super(BStageDM.TYPE_PLAYER1, 5);
        setArmor(DEFAULT_ARMOR);
		setShape(new BShape(BStageDM.PLAYER_WIDTH));
		enableState(Body.STATE_DEAD);
        setEnergy(getMaxEnergy());

        //TODO make sure to comment out this line
        //setOption(OPTION_MORALE, 1000);
    }
    protected BodyPlayer(byte type, int options)
    {
        super(type, options);
        setArmor(DEFAULT_ARMOR);
        setShape(new BShape(BStageDM.PLAYER_WIDTH));
        enableState(Body.STATE_DEAD);
        setEnergy(getMaxEnergy());
    }

    @Override
    public boolean isLoopPosition()
    {
        return true;
    }
	public boolean hasName()
	{
		return true;
	}
	public boolean hasVariableBCircle()
	{
		return false;
	} 
	public boolean hasScore()
	{
		return true;
	}

	public void actionForClient(BodyEvent bodyEvent, ClientBody mybody, CharacterImageBundle cib, EffectBundle eb, BUserInterface ui, long currenttime, BVector2f bufferCenter)
	{
        if (bodyEvent instanceof BodyActionEvent)
        {
            BodyActionEvent bae = (BodyActionEvent)bodyEvent;
            if (bae.getAction() == BodyPlayer.ACTION_BARRAGE
                    || bae.getAction() == BodyPlayer.ACTION_CLUSTERBOMB
                    || bae.getAction() == BodyPlayer.ACTION_PHALANX
                    || bae.getAction() == BodyPlayer.ACTION_AVENGER)
            {
                EffectFlush effect = new EffectFlush(this, currenttime, 40f);
                eb.putPostEffect(effect);
            }
            byte soundType = -1;
            switch (bae.getAction())
            {
                case ACTION_MISSILE :
                    soundType = BStageDM.SOUND_MISSILE;
                    break;
                case ACTION_BOMB :
                    soundType = BStageDM.SOUND_BOMB;
                    break;
                case ACTION_SLUG :
                    soundType = BStageDM.SOUND_SLUG;
                    break;
                case ACTION_HOMING :
                    soundType = BStageDM.SOUND_HOMING;
                    break;
                case ACTION_BARRAGE :
                    soundType = BStageDM.SOUND_BARRAGE;
                    break;
                case ACTION_CLUSTERBOMB :
                    soundType = BStageDM.SOUND_BOMB;
                    break;
                case ACTION_PHALANX :
                    soundType = BStageDM.SOUND_PHALANX;
                    break;
                case ACTION_AVENGER :
                    soundType = BStageDM.SOUND_AVENGER;
                    break;
                case ACTION_FIERCEATTACK :
                    soundType = BStageDM.SOUND_FIERCEATTACK;
                    break;
                default :
                    break;
            }
            if (soundType != -1)
            {
                ui.getBAudio().playSE(soundType, 0);
            }
        }
        else if (bodyEvent instanceof BodyArmorEvent)
        {
            BodyArmorEvent bae = (BodyArmorEvent)bodyEvent;
            if (bae.getType() == BodyArmorEvent.ADD)
            {
                EffectHealth effect = new EffectHealth(this, EffectHealth.ARMOR, currenttime);
                eb.putPostEffect(effect);
                if (this == mybody)
                {
                    ui.outputMessage(ui.translate("armor_is_repaired"));
                }
            }
            else if (bae.getType() == BodyArmorEvent.SUB)
            {
                EffectHealth effect = new EffectHealth(this, EffectHealth.DAMAGE, currenttime);
                eb.putPostEffect(effect);
            }
        }
        else if (bodyEvent instanceof BodyAddEvent)
		{
			ui.getBAudio().playSE(BStageDM.SOUND_LOGIN, 0);
		}
		else if (bodyEvent instanceof BodyDeleteEvent)
		{
			ui.getBAudio().playSE(BStageDM.SOUND_LOGOUT, 0);
		}
		else if (bodyEvent instanceof BodyStateEvent)
		{
			BodyStateEvent bse = (BodyStateEvent)bodyEvent;
			if ((bse.getType() == Body.SET || bse.getType() == Body.ENABLE)
					&& (bse.getState() & Body.STATE_DEAD) == Body.STATE_DEAD)
			{
				ui.getBAudio().playSE(BStageDM.SOUND_DOWN, 0, bufferCenter, getPosition());
				ui.outputMessage(ui.translate("is_down", new String[]{getName()}));
                EffectBlast effect = new EffectBlast(getPosition(), getVelocity(), 50, 10f, currenttime);
                eb.putPreEffect(effect);
			}
            else if ((bse.getType() == Body.SET || bse.getType() == Body.ENABLE)
                    && (bse.getState() & BodyPlayer.STATE_FIERCEATTACK) == BodyPlayer.STATE_FIERCEATTACK)
            {
                EffectFierceAttack effect = new EffectFierceAttack(this, currenttime);
                eb.putPreEffect(effect);
            }
		}
		else if (bodyEvent instanceof BodyMessageEvent)
		{
			BodyMessageEvent bme = (BodyMessageEvent)bodyEvent;
			if (bme.getType() == DataManager.MESSAGE_CHAT)
			{
				ui.getBAudio().playSE(BStageDM.SOUND_CHAT, 0);
			}
		}
        else if (bodyEvent instanceof BodyOptionEvent)
        {
            BodyOptionEvent boe = (BodyOptionEvent)bodyEvent;
            byte type = boe.getType();
            if (type == OPTION_FIRE
                    || type == OPTION_CHARGE
                    || type == OPTION_SPEED
                    )
            {
                if (getOption(type) < boe.getValue())
                {
                    ui.getBAudio().playSE(BStageDM.SOUND_ITEM, 0, bufferCenter, getPosition());
                    EffectHealth effect = new EffectHealth(this, EffectHealth.ITEM, currenttime);
                    eb.putPostEffect(effect);
                }
                if (this == mybody)
                {
                    if (type == BodyPlayer.OPTION_FIRE)
                    {
                        ui.outputMessage(ui.translate("fire_is_rised"));
                    }
                    else if (type == BodyPlayer.OPTION_CHARGE)
                    {
                        ui.outputMessage(ui.translate("charge_is_rised"));
                    }
                    else if (type == BodyPlayer.OPTION_SPEED)
                    {
                        ui.outputMessage(ui.translate("speed_is_rised"));
                    }
                }
            }
        }
	}
//    public void drawCharacter(Graphics g, Dimension offset, long currenttime)
//    {
//        if(!isState(LIFE))
//        {
//            int n = ((int)(currenttime - getLastStateChange()))/(DataManager.ANIMINTERVAL/2);
//            if ((n%2) == 0)
//                return;
//        }
//        super.drawCharacter(g, offset, currenttime);
//    }
	public void drawCharacter(Graphics g, long currenttime, CharacterImageBundle cib, EffectBundle eb)
	{
		if(!isState(STATE_LIFE))
		{
			int n = ((int)(currenttime - getLastStateChange()))/(DataManager.ANIMINTERVAL/2);
			if ((n%2) == 0)
				return;
		}
        CharacterImage characterImage = getCharacterImage();
        if (characterImage == null)
            return;
        
        int offsetx = characterImage.getWidth() / 2;
        int offsety = characterImage.getHeight() / 2;
        int x = 0;
        int y = 0;
        int n = ((int)(currenttime - getLastStateChange()))/DataManager.ANIMINTERVAL;

        Image[] imgs;
        if (isState(Body.STATE_WIN))
        {
            imgs = characterImage.getImages(CharacterImage.WIN);
        }
        else if (isState(Body.STATE_DEAD))
        {
            imgs = characterImage.getImages(CharacterImage.DEAD);
        }
        else if (!getVelocity().isZeroVector())
        {
            imgs = characterImage.getImages(CharacterImage.MOVEDOWN);
        }
        else
        {
            imgs = characterImage.getImages(CharacterImage.DOWN);
        }

        Image img = imgs[n%2];

        BVector2f v = new BVector2f(getDirection());
        v.normalize();
        float l = v.getLength();
        double rad;
        if (l > 0)
        {
            rad = Math.acos(v.y/l);
            if (v.x > 0)
                rad = 2 * Math.PI - rad;
        }else
        {
            rad = 0;
        }

        AffineTransform at = new AffineTransform();
        at.setToTranslation(x,y);
        at.rotate(rad);
        at.translate(-offsetx,-offsety);
        
        Graphics2D g2 = (Graphics2D)g;
        g2.drawImage(img, at, null);
    }
	public boolean eventAction(EventHandler eventHandler, BStage bstage, BVector2f position, BodyActionEvent event)
	{
		if (isState(Body.STATE_DEAD) || !isState(Body.STATE_LIFE))
			return false;
		
		if (event.getAction() == ACTION_MISSILE)
		{
            long currenttime = bstage.getCurrentTime();
			
			BVector2f vec = new BVector2f();
            vec.set(event.getTarget());
            vec.sub(event.getPosition());
            float distance = vec.getLength();
            if (distance == 0)
            {
                return false;
            }
            if (distance > BodyMissile.SHOOTING_RANGE)
                distance = BodyMissile.SHOOTING_RANGE;
            long interval = (long)(distance / (getMaxSpeed() * SPEEDSCALE_MISSILE) * 1000f);
            vec.normalize();
            vec.scale(getMaxSpeed() * SPEEDSCALE_MISSILE);
            
            BodyMissile body = new BodyMissile();
            body.setOwner(this);
            //body.setID(bstage.getNextID());
            body.setPosition(getPosition());
            body.setOption(BodyMissile.OPTION_FIRE, getOption(OPTION_FIRE) + DEFAULT_FIRE);
			body.setVelocity(vec);
			body.setDirection(vec);
			
			eventHandler.add(currenttime,currenttime,body);
            eventHandler.smallAction(currenttime, currenttime + interval, body, BodyMissile.ACTION_BOMB);
            return true;
		}
		else if (event.getAction() == ACTION_BOMB)
		{
            long currenttime = bstage.getCurrentTime();

			BodyBomb body = new BodyBomb();
            body.setOwner(this);
            //body.setID(bstage.getNextID());
			body.setPosition(getPosition());
            int size = getOption(OPTION_FIRE) + DEFAULT_FIRE + 4;
			body.setOption(BodyBomb.OPTION_FIRE, size);
            BVector2f vec = new BVector2f();
            vec.set(event.getTarget());
            vec.sub(event.getPosition());
            float distance = vec.getLength();
            if (distance > 0)
            {
                vec.normalize();
                body.setDirection(vec);
                vec.scale(getMaxSpeed() * SPEEDSCALE_BOMB);
                body.setVelocity(vec);
            }
            else
            {
                body.setVelocity(vec);
                body.setDirection(getDirection());
            }

			eventHandler.add(currenttime,currenttime,body);
            long interval = (long)(distance/(getMaxSpeed()*SPEEDSCALE_BOMB)*1000l);
            eventHandler.smallAction(currenttime, currenttime + interval, body, BodyBomb.ACTION_STOP);
            eventHandler.smallAction(currenttime, currenttime + 3000, body, BodyBomb.ACTION_BOMB);
            return true;
		}
        else if (event.getAction() == ACTION_SLUG)
        {
            long currenttime = bstage.getCurrentTime();
            
            BodySlug body = new BodySlug();
            body.setOwner(this);
            //body.setID(bstage.getNextID());
            body.setPosition(getPosition());
            body.setArmor(BodySlug.SLUG_ARMOR + getOption(OPTION_FIRE)*BodySlug.SLUG_ARMOR_INCREASE);
            body.setOption(BodyMissile.OPTION_FIRE, getOption(OPTION_FIRE) + DEFAULT_FIRE);
            body.putHit(this);
            BVector2f vec = new BVector2f();
            vec.set(event.getTarget());
            vec.sub(event.getPosition());
            float distance = vec.getLength();
            if (distance == 0)
            {
                vec.set(getDirection());
            }

            vec.normalize();
            body.setDirection(vec);
            vec.scale(getMaxSpeed()*SPEEDSCALE_SLUG);
            body.setVelocity(vec);
            
            eventHandler.add(currenttime,currenttime,body);
            eventHandler.smallAction(currenttime,currenttime + BodySlug.LIFE_TIME,body,BodySlug.ACTION_BOMB);
            return true;
        }
        else if (event.getAction() == ACTION_HOMING)
        {
            long currenttime = bstage.getCurrentTime();
            
            BodyHoming body = new BodyHoming();
            body.setOwner(this);
            //body.setID(bstage.getNextID());
            body.setPosition(getPosition());
            body.setOption(BodyHoming.OPTION_FIRE, getOption(OPTION_FIRE) + DEFAULT_FIRE);
            BVector2f vec = new BVector2f();
            vec.set(event.getTarget());
            vec.sub(event.getPosition());
            float distance = vec.getLength();
            if (distance == 0)
            {
                vec.set(getDirection());
                distance = vec.getLength();
            }

            vec.normalize();
            body.setDirection(vec);
            vec.scale(getMaxSpeed() * SPEEDSCALE_HOMING);
            body.setVelocity(vec);
            
            //body.setVelocity(getVelocity());
            Body[] bs = bstage.getBodysHasScore();
            Body aim_target = null;
            float d_max = Float.MAX_VALUE;
            for(int i=0;i<bs.length;i++)
            {
                if (bs[i] == this)
                    continue;
                float d = BVector2f.getDistance(event.getTarget(),bs[i].getPosition());
                if (d_max > d)
                {
                    d_max = d;
                    aim_target = bs[i];
                }
            }
            if (aim_target != null && BVector2f.getDistance(getPosition(),aim_target.getPosition()) <= 5f);
            {
                body.setTarget(aim_target);
            }
            
            eventHandler.add(currenttime,currenttime,body);
            eventHandler.smallAction(currenttime, currenttime + BodyHoming.AIM_INTERVAL, body, BodyHoming.ACTION_AIM);
            eventHandler.smallAction(currenttime, currenttime + 10000, body, BodyHoming.ACTION_BOMB);
            return true;
        }
        else if (event.getAction() == ACTION_FIERCEATTACK)
        {
            if (getOption(BodyPlayer.OPTION_MORALE) < 100 || isState(BodyPlayer.STATE_FIERCEATTACK))
            {
                return false;
            }
            long currenttime = bstage.getCurrentTime();
            eventHandler.state(currenttime, currenttime, this, Body.ENABLE, BodyPlayer.STATE_FIERCEATTACK);
            eventHandler.state(currenttime, currenttime+30000, this, Body.DISABLE, BodyPlayer.STATE_FIERCEATTACK);
            eventHandler.option(currenttime, currenttime + 30000,this, BodyPlayer.OPTION_MORALE, 0);
            return true;
        }
        else if (event.getAction() == ACTION_BARRAGE)
        {
            if (getOption(BodyPlayer.OPTION_MORALE) < 100 || !isState(BodyPlayer.STATE_FIERCEATTACK))
            {
                return false;
            }
            long currenttime = bstage.getCurrentTime();
            BVector2f dir = new BVector2f(event.getTarget());
            dir.sub(getPosition());
            if (dir.isZeroVector())
            {
                dir.set(getDirection());
                if (dir.isZeroVector())
                    return false;
            }
            dir.normalize();
            Body[][] missiles = new Body[24][];
            for (int i=0;i<24;i++)
            {
                missiles[i] = new Body[2];
            }
            
            double angle = Math.PI / 6;
            double d = Math.PI / 144;
            BVector2f vec = new BVector2f();
            vec.set(dir);
            vec.rotate(-angle-d);
            for (int i=0;i<24;i++)
            {
                if (i == 12)
                {
                    vec.set(dir);
                    vec.rotate(-angle-d);
                }
                //float speedscale = SPEEDSCALE_MISSILE * (2f - (float)i/24f);
                float speedscale = SPEEDSCALE_MISSILE;
                BodyMissile missile = new BodyMissile();
                missile.setOwner(this);
                missile.setArmor(BodyMissile.FIRE_DAMAGE+1);
                //missile.setID(bstage.getNextID());
                missile.setPosition(getPosition());
                missile.setOption(BodyMissile.OPTION_FIRE, getOption(OPTION_FIRE) + DEFAULT_FIRE);
                missile.setDirection(vec);
                missile.setVelocity(vec);
                missile.getVelocity().scale(getMaxSpeed() * speedscale);
//                missile.getPosition().scaleAdd(3f, vec);
                if (i%2 == 0)
                    missile.getPosition().scaleAdd(3f, vec);
                else
                    missile.getPosition().scaleAdd(4f, vec);
                missiles[i][0] = missile;
                //eventHandler.add(currenttime,currenttime + 100 * i,missile);
                eventHandler.smallAction(currenttime, currenttime + BodyMissile.LIFE_TIME, missile, BodyMissile.ACTION_BOMB);
                
                vec.rotate(angle/ 6);
            }
            
            vec.set(dir);
            vec.rotate(angle+d);
            for (int i=0;i<24;i++)
            {
                if (i == 12)
                {
                    vec.set(dir);
                    vec.rotate(angle+d);// * (1d/2d + 1d/24d));
                }
                //float speedscale = SPEEDSCALE_MISSILE * (2f - (float)i/24f);
                float speedscale = SPEEDSCALE_MISSILE;
                BodyMissile missile = new BodyMissile();
                missile.setOwner(this);
                missile.setArmor(BodyMissile.FIRE_DAMAGE+1);
                //missile.setID(bstage.getNextID());
                missile.setPosition(getPosition());
                missile.setOption(BodyMissile.OPTION_FIRE, getOption(OPTION_FIRE) + DEFAULT_FIRE);
                missile.setDirection(vec);
                missile.setVelocity(vec);
                missile.getVelocity().scale(getMaxSpeed() * speedscale);
                if (i%2 == 0)
                    missile.getPosition().scaleAdd(3f, vec);
                else
                    missile.getPosition().scaleAdd(4f, vec);
                missiles[i][1] = missile;
                //eventHandler.add(currenttime,currenttime + 100 * i,missile);
                eventHandler.smallAction(currenttime, currenttime + BodyMissile.LIFE_TIME, missile, BodyMissile.ACTION_BOMB);
                
                vec.rotate(-angle/6);
            }
            for (int i=0;i<missiles.length;i++)
            {
                eventHandler.multiAdd(currenttime,currenttime + i*50,missiles[i]);
            }
            int morale = getOption(BodyPlayer.OPTION_MORALE) - 100;
            eventHandler.option(currenttime,currenttime,this, BodyPlayer.OPTION_MORALE, morale);
            return true;

//            if (getOption(BodyPlayer.OPTION_MORALE) < 100 || !isState(BodyPlayer.STATE_FIERCEATTACK))
//            {
//                return false;
//            }
//            long currenttime = bstage.getCurrentTime();
//            BVector2f dir = new BVector2f(event.getTarget());
//            dir.sub(getPosition());
//            if (dir.isZeroVector())
//            {
//                dir.set(getDirection());
//                if (dir.isZeroVector())
//                    return false;
//            }
//            dir.normalize();
//            BVector2f vec = new BVector2f();
//            vec.set(dir);
//            Body[][] missiles = new Body[12][];
//            missiles[0] = new Body[28];
//            for (int i=1;i<12;i++)
//            {
//                missiles[i] = new Body[4];
//            }
//            for (int i=0;i<12;i++)
//            {
//                BodyMissile missile = new BodyMissile();
//                missile.setOwner(this);
//                missile.setID(bstage.getNextID());
//                missile.setPosition(getPosition());
//                missile.setOption(BodyMissile.OPTION_FIRE, getOption(OPTION_FIRE) + DEFAULT_FIRE);
//                missile.setDirection(vec);
//                missile.setVelocity(vec);
//                missile.getVelocity().scale(getMaxSpeed() * SPEEDSCALE_HOMING);
//                missile.getPosition().scaleAdd(3f, vec);
//                missiles[i][0] = missile;
//                //eventHandler.add(currenttime,currenttime + 100 * i,missile);
//                eventHandler.smallAction(currenttime, currenttime + BodyMissile.LIFE_TIME, missile, BodyMissile.ACTION_BOMB);
//                
//                vec.rotate(Math.PI / 12);
//            }
//            vec.set(dir);
//            vec.scale(-1);
//            for (int i=0;i<12;i++)
//            {
//                BodyMissile missile = new BodyMissile();
//                missile.setOwner(this);
//                missile.setID(bstage.getNextID());
//                missile.setPosition(getPosition());
//                missile.setOption(BodyMissile.OPTION_FIRE, getOption(OPTION_FIRE) + DEFAULT_FIRE);
//                missile.setDirection(vec);
//                missile.setVelocity(vec);
//                missile.getVelocity().scale(getMaxSpeed() * SPEEDSCALE_MISSILE);
//                missile.getPosition().scaleAdd(3f, vec);
//                missiles[i][1] = missile;
//                //eventHandler.add(currenttime,currenttime + 100 * i,missile);
//                eventHandler.smallAction(currenttime, currenttime + BodyMissile.LIFE_TIME, missile, BodyMissile.ACTION_BOMB);
//                
//                vec.rotate(Math.PI / 12);
//            }
//            vec.set(dir);
//            for (int i=0;i<12;i++)
//            {
//                BodyMissile missile = new BodyMissile();
//                missile.setOwner(this);
//                missile.setID(bstage.getNextID());
//                missile.setPosition(getPosition());
//                missile.setOption(BodyMissile.OPTION_FIRE, getOption(OPTION_FIRE) + DEFAULT_FIRE);
//                missile.setDirection(vec);
//                missile.setVelocity(vec);
//                missile.getVelocity().scale(getMaxSpeed() * SPEEDSCALE_MISSILE);
//                missile.getPosition().scaleAdd(3f, vec);
//                missiles[i][2] = missile;
//                //eventHandler.add(currenttime,currenttime + 100 * i,missile);
//                eventHandler.smallAction(currenttime, currenttime + BodyMissile.LIFE_TIME, missile, BodyMissile.ACTION_BOMB);
//                
//                vec.rotate(-Math.PI / 12);
//            }
//            vec.set(dir);
//            vec.scale(-1);
//            for (int i=0;i<12;i++)
//            {
//                BodyMissile missile = new BodyMissile();
//                missile.setOwner(this);
//                missile.setID(bstage.getNextID());
//                missile.setPosition(getPosition());
//                missile.setOption(BodyMissile.OPTION_FIRE, getOption(OPTION_FIRE) + DEFAULT_FIRE);
//                missile.setDirection(vec);
//                missile.setVelocity(vec);
//                missile.getVelocity().scale(getMaxSpeed() * SPEEDSCALE_MISSILE);
//                missile.getPosition().scaleAdd(3f, vec);
//                missiles[i][3] = missile;
//                //eventHandler.add(currenttime,currenttime + 100 * i,missile);
////                eventHandler.action(currenttime, currenttime + BodyMissile.LIFE_TIME, missile, missile.getPosition(), BodyMissile.ACTION_BOMB);
//                eventHandler.smallAction(currenttime, currenttime + BodyMissile.LIFE_TIME, missile, BodyMissile.ACTION_BOMB);
//                
//                vec.rotate(-Math.PI / 12);
//            }
//            for (int i=0;i<24;i++)
//            {
//                BodyMissile missile = new BodyMissile();
//                missile.setOwner(this);
//                missile.setID(bstage.getNextID());
//                missile.setPosition(getPosition());
//                missile.setOption(BodyMissile.OPTION_FIRE, getOption(OPTION_FIRE) + DEFAULT_FIRE);
//                missile.setDirection(vec);
//                missile.setVelocity(vec);
//                missile.getVelocity().scale(getMaxSpeed() * SPEEDSCALE_MISSILE);
//                missile.getPosition().scaleAdd(3f, vec);
//                int deray = 200 * (i%2);
//
//                missiles[0][i+4] = missile;
//                //eventHandler.add(currenttime,currenttime+deray,missile);
////                eventHandler.action(currenttime, currenttime + BodyMissile.LIFE_TIME, missile, missile.getPosition(), BodyMissile.ACTION_BOMB);
//                eventHandler.smallAction(currenttime, currenttime + BodyMissile.LIFE_TIME, missile, BodyMissile.ACTION_BOMB);
//                
//                vec.rotate(-Math.PI / 12);
//            }
//            for (int i=0;i<12;i++)
//            {
//                eventHandler.multiAdd(currenttime,currenttime + i*100,missiles[i]);
//            }
//            int morale = getOption(BodyPlayer.OPTION_MORALE) - 100;
//            eventHandler.option(currenttime,currenttime,this, BodyPlayer.OPTION_MORALE, morale);
//            return true;

        }
        else if(event.getAction() == ACTION_CLUSTERBOMB)
        {
            if (getOption(BodyPlayer.OPTION_MORALE) < 100 || !isState(BodyPlayer.STATE_FIERCEATTACK))
            {
                return false;
            }
            long currenttime = bstage.getCurrentTime();

            BodyClusterBomb body = new BodyClusterBomb();
            body.setOwner(this);
            //body.setID(bstage.getNextID());
            body.setPosition(getPosition());
            body.setOption(BodyClusterBomb.OPTION_FIRE, (getOption(OPTION_FIRE) + DEFAULT_FIRE)*2);
            body.setOption(BodyClusterBomb.OPTION_MODE, BodyClusterBomb.MODE_DIFFUSION);
            BVector2f vec = new BVector2f();
            vec.set(event.getTarget());
            vec.sub(event.getPosition());
            float distance = vec.getLength();
            if (distance > 0)
            {
                vec.normalize();
                body.setDirection(vec);
                vec.scale(getMaxSpeed() * SPEEDSCALE_BOMB);
                body.setVelocity(vec);
            }
            else
            {
                body.setVelocity(vec);
                body.setDirection(getDirection());
            }

            eventHandler.add(currenttime,currenttime,body);
            long interval = (long)(distance/(getMaxSpeed()*SPEEDSCALE_BOMB)*1000l);
            eventHandler.smallAction(currenttime, currenttime + interval, body, BodyClusterBomb.ACTION_STOP);
            eventHandler.smallAction(currenttime, currenttime + 3000, body, BodyClusterBomb.ACTION_BOMB);

            int morale = getOption(BodyPlayer.OPTION_MORALE) - 100;
            eventHandler.option(currenttime,currenttime,this, BodyPlayer.OPTION_MORALE, morale);
            return true;
        }
        else if (event.getAction() == ACTION_PHALANX)
        {
            if (getOption(BodyPlayer.OPTION_MORALE) < 100 || !isState(BodyPlayer.STATE_FIERCEATTACK))
            {
                return false;
            }
            long currenttime = bstage.getCurrentTime();
            BVector2f dir = new BVector2f(event.getTarget());
            dir.sub(getPosition());
            if (dir.isZeroVector())
            {
                dir.set(getDirection());
                if (dir.isZeroVector())
                    return false;
            }
            dir.normalize();
            BVector2f dirV = new BVector2f(dir);
            dirV.rotate(Math.PI/2);
            BVector2f vec = new BVector2f();
            vec.set(dir);
            for (int i=0;i<=12;i++)
            {
                BodySlug slug = new BodySlug();
                slug.setOwner(this);
                //slug.setID(bstage.getNextID());
                slug.setPosition(getPosition());
                int p = i;
                if (p > 6)
                    p = 12 - p;
                slug.getPosition().scaleAdd(BStageDM.SLUG_WIDTH*2*p, dirV);
                slug.setArmor(2*(BodySlug.SLUG_ARMOR + getOption(OPTION_FIRE)*BodySlug.SLUG_ARMOR_INCREASE));
                slug.setOption(BodySlug.OPTION_FIRE, getOption(OPTION_FIRE) + DEFAULT_FIRE);
                slug.setDirection(vec);
                slug.setVelocity(vec);
                slug.getVelocity().scale(getMaxSpeed() * SPEEDSCALE_SLUG);
                //slug.getPosition().scaleAdd(1.5f, vec);
                slug.putHit(this);
                eventHandler.add(currenttime,currenttime + 100 * i,slug);
                //eventHandler.action(currenttime,currenttime + BodySlug.LIFE_TIME,slug,slug.getPosition(),BodySlug.ACTION_BOMB);
                eventHandler.smallAction(currenttime,currenttime + BodySlug.LIFE_TIME,slug,BodySlug.ACTION_BOMB);
            }
            dirV.scale(-1f);
            for (int i=1;i<12;i++)
            {
                BodySlug slug = new BodySlug();
                slug.setOwner(this);
                //slug.setID(bstage.getNextID());
                slug.setPosition(getPosition());
                int p = i;
                if (p > 6)
                    p = 12 - p;
                slug.getPosition().scaleAdd(BStageDM.SLUG_WIDTH*2*p, dirV);
                slug.setArmor(10*(BodySlug.SLUG_ARMOR + getOption(OPTION_FIRE)*BodySlug.SLUG_ARMOR_INCREASE));
                slug.setOption(BodySlug.OPTION_FIRE, getOption(OPTION_FIRE) + DEFAULT_FIRE);
                slug.setDirection(vec);
                slug.setVelocity(vec);
                slug.getVelocity().scale(getMaxSpeed() * SPEEDSCALE_SLUG);
                //slug.getPosition().scaleAdd(1.5f, vec);
                slug.putHit(this);
                eventHandler.add(currenttime,currenttime + 100 * i,slug);
                //eventHandler.action(currenttime,currenttime + BodySlug.LIFE_TIME,slug,slug.getPosition(),BodySlug.ACTION_BOMB);
                eventHandler.smallAction(currenttime,currenttime + BodySlug.LIFE_TIME*3, slug,BodySlug.ACTION_BOMB);
            }
            int morale = getOption(BodyPlayer.OPTION_MORALE) - 100;
            eventHandler.option(currenttime,currenttime,this, BodyPlayer.OPTION_MORALE, morale);
            return true;
        }
        else if (event.getAction() == ACTION_AVENGER)
        {
            if (getOption(BodyPlayer.OPTION_MORALE) < 100 || !isState(BodyPlayer.STATE_FIERCEATTACK))
            {
                return false;
            }
            long currenttime = bstage.getCurrentTime();
            BVector2f dir = new BVector2f(event.getTarget());
            dir.sub(getPosition());
            if (dir.isZeroVector())
            {
                dir.set(getDirection());
                if (dir.isZeroVector())
                    return false;
            }
            dir.normalize();
            BVector2f vec = new BVector2f();
            vec.set(dir);
            BodyHoming[] bs1 = new BodyHoming[18];
            BodyHoming[] bs2 = new BodyHoming[18];
            for (int i=0;i<36;i++)
            {
                BodyHoming homing = new BodyHoming();
                homing.setOwner(this);
                //homing.setID(bstage.getNextID());
                homing.setArmor(BodyHoming.FIRE_DAMAGE+1);
                homing.setPosition(getPosition());
                homing.setOption(BodyHoming.OPTION_FIRE, getOption(OPTION_FIRE) + DEFAULT_FIRE);
                homing.setDirection(vec);
                homing.setVelocity(vec);
                homing.getVelocity().scale(getMaxSpeed() * SPEEDSCALE_HOMING);
                homing.getPosition().scaleAdd(3f, vec);
                if (i%2 == 0)
                {
                    bs1[i/2] = homing;
                }
                else
                {
                    bs2[i/2] = homing;
                }
//                int deray = 300 * (i%2);
//                eventHandler.add(currenttime,currenttime + deray,homing);
//                eventHandler.smallAction(currenttime,currenttime + deray,homing, BodyHoming.ACTION_AIM);
//                eventHandler.smallAction(currenttime,currenttime+BodyHoming.LIFE_TIME, homing, BodyHoming.ACTION_BOMB);
                vec.rotate(Math.PI / 18);
            }
            eventHandler.multiAdd(currenttime,currenttime,bs1);
            eventHandler.multiAdd(currenttime,currenttime+400,bs2);
            for (int i=0;i<18;i++)
            {
                eventHandler.smallAction(currenttime,currenttime,bs1[i], BodyHoming.ACTION_AIM);
                eventHandler.smallAction(currenttime,currenttime+BodyHoming.LIFE_TIME, bs1[i], BodyHoming.ACTION_BOMB);
                eventHandler.smallAction(currenttime,currenttime+400,bs2[i], BodyHoming.ACTION_AIM);
                eventHandler.smallAction(currenttime,currenttime+BodyHoming.LIFE_TIME, bs2[i], BodyHoming.ACTION_BOMB);
            }
            int morale = getOption(BodyPlayer.OPTION_MORALE) - 100;
            eventHandler.option(currenttime,currenttime,this, BodyPlayer.OPTION_MORALE, morale);
            return true;
        }
        return false;
	}
	public float getMaxSpeed()
	{
		return 0.25f * (getOption(BodyPlayer.OPTION_SPEED) + DEFAULT_SPEED);
	}
	public void eventCollisition(EventHandler eventHandler, BStage bstage, Body target)
	{
	}
	public void eventStateChange(EventHandler eventHandler, BStage bstage, byte type, int state)
	{
		if ((type == Body.SET
				|| type == Body.ENABLE)
				&& (state & Body.STATE_DEAD) == Body.STATE_DEAD)
		{
			long currenttime = bstage.getCurrentTime();
			eventHandler.setPosition(currenttime, currenttime, this, getPosition());
			eventHandler.setPosition(currenttime, currenttime + 3000, this, bstage.getRandomeSpawnPoint());

			for (int i=0;i<getOption(OPTION_FIRE);i++)
			{
				BVector2f pos = ((BStageDM)bstage).getRandomeFreePoint();
				BodyItem body = new BodyItem();
				body.setOption(BodyItem.OPTION_TYPE, BodyItem.ITEM_FIRE);
				body.setPosition(pos);
				if (!(bstage.isIntersectOther(body,STATE_COLLIDABLE)))
				{
					eventHandler.add(currenttime, currenttime, body);
					eventHandler.state(currenttime, currenttime + BStageDM.ITEM_SPAWN_INTERVAL, body, ENABLE, STATE_DEAD);
					eventHandler.delete(currenttime, currenttime + BStageDM.ITEM_SPAWN_INTERVAL + 500, body);
				}
			}
			for (int i=0;i<getOption(OPTION_SPEED);i++)
			{
				BodyItem body = new BodyItem();
				body.setOption(BodyItem.OPTION_TYPE, BodyItem.ITEM_SPEED);
				body.setPosition(((BStageDM)bstage).getRandomeFreePoint());
				if (!(bstage.isIntersectOther(body,STATE_COLLIDABLE)))
				{
					eventHandler.add(currenttime, currenttime, body);
					eventHandler.state(currenttime, currenttime + BStageDM.ITEM_SPAWN_INTERVAL, body, ENABLE, STATE_DEAD);
					eventHandler.delete(currenttime, currenttime + BStageDM.ITEM_SPAWN_INTERVAL + 500, body);
				}
			}
			for (int i=0;i<getOption(OPTION_CHARGE);i++)
			{
				BodyItem body = new BodyItem();
				body.setOption(BodyItem.OPTION_TYPE, BodyItem.ITEM_CHARGE);
				body.setPosition(((BStageDM)bstage).getRandomeFreePoint());
				if (!(bstage.isIntersectOther(body,STATE_COLLIDABLE)))
				{
					eventHandler.add(currenttime, currenttime, body);
					eventHandler.state(currenttime, currenttime + BStageDM.ITEM_SPAWN_INTERVAL, body, ENABLE, STATE_DEAD);
					eventHandler.delete(currenttime, currenttime + BStageDM.ITEM_SPAWN_INTERVAL + 500, body);
				}
			}
			if ((getOption(OPTION_WEAPON) & WEAPON_KICK) == WEAPON_KICK)
			{
				BodyItem body = new BodyItem();
				body.setOption(BodyItem.OPTION_TYPE, BodyItem.ITEM_WEAPON);
				body.setWeapon(WEAPON_KICK);
				body.setPosition(((BStageDM)bstage).getRandomeFreePoint());
				if (!(bstage.isIntersectOther(body,STATE_COLLIDABLE)))
				{
					eventHandler.add(currenttime, currenttime, body);
					eventHandler.state(currenttime, currenttime + BStageDM.ITEM_SPAWN_INTERVAL, body, ENABLE, STATE_DEAD);
					eventHandler.delete(currenttime, currenttime + BStageDM.ITEM_SPAWN_INTERVAL + 500, body);
				}
			}
            
            //Respawn
            eventHandler.state(currenttime, currenttime + 3000, this, Body.DISABLE, Body.STATE_LIFE | Body.STATE_DEAD);
            eventHandler.armor(currenttime,currenttime + 3000, this, this, BodyArmorEvent.SET, BodyPlayer.DEFAULT_ARMOR);
            eventHandler.state(currenttime, currenttime + 3000, this, Body.ENABLE, Body.STATE_MARKER);
            eventHandler.state(currenttime, currenttime + 6000, this, Body.DISABLE, Body.STATE_MARKER);
            eventHandler.state(currenttime, currenttime + 6000, this, Body.ENABLE, Body.STATE_LIFE);
            eventHandler.option(currenttime, currenttime + 3000, this, BodyPlayer.OPTION_FIRE, 0);
            eventHandler.option(currenttime, currenttime + 3000, this, BodyPlayer.OPTION_SPEED, 0);
            eventHandler.option(currenttime, currenttime + 3000, this, BodyPlayer.OPTION_CHARGE, 0);
            eventHandler.option(currenttime, currenttime + 3000, this, BodyPlayer.OPTION_WEAPON, 0);
            
            //Score
            if (lastAttacked == this)
            {
                eventHandler.scoreChange(currenttime,currenttime, this, new Body[]{this}, new int[]{-1});
            }
            else
            {
                eventHandler.scoreChange(currenttime,currenttime, this, new Body[]{lastAttacked}, new int[]{1});
            }
		}
	}
    
	public void eventAdd(EventHandler eventHandler, BStage bstage)
	{
		long currenttime = bstage.getCurrentTime();
		//eventHandler.sound(currenttime,currenttime, BStageThrowBom.SOUND_LOGIN);
		eventHandler.state(currenttime, currenttime, this, Body.ENABLE, Body.STATE_MARKER);
		eventHandler.state(currenttime, currenttime, this, Body.DISABLE, Body.STATE_DEAD);
		eventHandler.state(currenttime, currenttime + 3000, this, Body.DISABLE, Body.STATE_MARKER);
		eventHandler.state(currenttime, currenttime + 3000, this, Body.ENABLE, Body.STATE_LIFE);
	}
	public void eventRemove(EventHandler eventHandler, BStage bstage)
	{
		//eventHandler.sound(bstage.getCurrentTime(),bstage.getCurrentTime(), BStageThrowBom.SOUND_LOGOUT);
	}
	public void eventCommand(EventHandler eventHandler, BStage bstage, String command)
	{
		if (command.length() != 0)
		{
			long currenttime = bstage.getCurrentTime();
			if (command.charAt(0) != '/')
			{
				//eventHandler.sound(currenttime,currenttime,BStageThrowBom.SOUND_CHAT);
				eventHandler.message(currenttime, currenttime, this, DataManager.MESSAGE_CHAT,command);
			}else if (command.length() > 6 && command.substring(0,6).equals("/note "))
			{
				SimpleDateFormat sdf = new SimpleDateFormat();
				String note = "["+this.getName()+"]{"+ sdf.format(new Date()) +")"+command.substring(6, command.length());
				eventHandler.message(currenttime, currenttime, this,DataManager.MESSAGE_NOTE, note);
			}
		}
	}
    
	@Override
    public void eventArmorChange(EventHandler eventHandler, BStage bstage, Body from, int type, int value)
    {
        super.eventArmorChange(eventHandler, bstage, from, type, value);
        //remove this comment out
        if (from.getType() == BStageDM.TYPE_PLAYER1 && from != this)
        {
            if (from.hasScore())
                lastAttacked = from;
            long currenttime = bstage.getCurrentTime();
            if (type == BodyArmorEvent.SUB && !this.isState(BodyPlayer.STATE_FIERCEATTACK))
            {
                int morale = this.getOption(BodyPlayer.OPTION_MORALE) + value;
                if (morale > MAX_MORALE)
                    morale = MAX_MORALE;
                if (morale < 0)
                    morale = 0;
                eventHandler.option(currenttime, currenttime, this, BodyPlayer.OPTION_MORALE, morale);
            }
        }
    }
    @Override
    public void eventChangeOthersArmor(EventHandler eventHandler, BStage bstage, Body to, int type, int value)
    {
        super.eventChangeOthersArmor(eventHandler, bstage, to, type, value);
        if (to.getType() == BStageDM.TYPE_PLAYER1 && to != this)
        {
            long currenttime = bstage.getCurrentTime();
            if (type == BodyArmorEvent.SUB && !this.isState(BodyPlayer.STATE_FIERCEATTACK))
            {
                int morale = this.getOption(BodyPlayer.OPTION_MORALE) + value; 
                if (morale > MAX_MORALE)
                    morale = MAX_MORALE;
                if (morale < 0)
                    morale = 0;
                eventHandler.option(currenttime, currenttime, this, BodyPlayer.OPTION_MORALE, morale);
            }
        }
    }
    public void preAdd(BStage bstage)
	{
		setPosition(bstage.getRandomeSpawnPoint());
	}
	public boolean isCheckIntersectWith(Body target)
	{
	    return true;
	}
    @Override
    public byte[] getCheckCollidType()
    {
        return new byte[]{
                BStageDM.TYPE_WALL1,
                BStageDM.TYPE_SOFTWALL
                };
    }
    @Override
    public byte[] getCheckIntersectType()
    {
        return null;
    }
    @Override
    public boolean isCheckCollidWith(Body target)
    {
        if (target.getType() == BStageDM.TYPE_SOFTWALL
                || target.getType() == BStageDM.TYPE_WALL1)
            return true;
        else
            return false;
    }
    @Override
    public short getActionLockTime(byte action)
    {
        if (action == ACTION_MISSILE)
            return ACTIONLOCK_MISSILE;
        else if (action == BodyPlayer.ACTION_BOMB)
            return ACTIONLOCK_BOMB;
        else if (action == BodyPlayer.ACTION_SLUG)
            return ACTIONLOCK_SLUG;
        else if (action == BodyPlayer.ACTION_HOMING)
            return ACTIONLOCK_HOMING;
        else if (action == BodyPlayer.ACTION_FIERCEATTACK)
            return ACTIONLOCK_FIERCEATTACK;
        else if (action == BodyPlayer.ACTION_BARRAGE)
            return ACTIONLOCK_BARRAGE;
        else if (action == BodyPlayer.ACTION_CLUSTERBOMB)
            return ACTIONLOCK_CLUSTERBOMB;
        else if (action == BodyPlayer.ACTION_PHALANX)
            return ACTIONLOCK_PHALANX;
        else if (action == BodyPlayer.ACTION_AVENGER)
            return ACTIONLOCK_AVENGER;
        return 0;
    }
    @Override
    public float getEnergyCharge()
    {
        return (float)(super.getOption(OPTION_CHARGE) + DEFAULT_CHARGE) * 10f;
    }
    @Override
    public float getEnergyConsumption(byte action)
    {
        if (action == ACTION_MISSILE)
            return CONSUMPTION_MISSILE;
        else if (action == ACTION_BOMB)
            return CONSUMPTION_BOMB;
        else if (action == ACTION_SLUG)
            return CONSUMPTION_SLUG;
        else if (action == ACTION_HOMING)
            return CONSUMPTION_HOMING;
        else if (action == ACTION_FIERCEATTACK)
            return CONSUMPTION_FIERCEATTACK;
        else if (action == ACTION_BARRAGE)
            return CONSUMPTION_BARRAGE;
        else if (action == ACTION_CLUSTERBOMB)
            return CONSUMPTION_CLUSTERBOMB;
        else if (action == ACTION_PHALANX)
            return CONSUMPTION_PHALANX;
        else if (action == ACTION_AVENGER)
            return CONSUMPTION_AVENGER;
        return super.getEnergyConsumption(action);
    }
    @Override
    public float getMaxEnergy()
    {
        return MAX_ENERGY;
    }
    @Override
    public boolean isActionableForClient(byte action, float energy, int actionlock)
    {
        boolean flag = true;
        if (energy < getEnergyConsumption(action))
        {
            showEnergyEmpty = getLastUpdate() + 500l;
            flag = false;
        }
        if (actionlock > 0)
        {
            flag = false;
        }
        if (!isState(Body.STATE_LIFE))
        {
            flag = false;
        }
        if (action == ACTION_FIERCEATTACK
                || action == ACTION_BARRAGE
                || action == ACTION_CLUSTERBOMB
                || action == ACTION_PHALANX
                || action == ACTION_AVENGER
                )
        {
            if (getOption(OPTION_MORALE) < 100)
            {
                showMoraleEmpty = getLastUpdate() + 500l;
                flag = false;
            }
        }
        return flag;
    }
    @Override
    public void outputData(DataManager dm, DataOutputStream out) throws IOException
    {
        super.outputData(dm, out);
        out.writeInt(getActionLock());
        out.writeFloat(getEnergy());
    }
    @Override
    public void inputData(DataManager dm, DataInputStream in) throws IOException
    {
        super.inputData(dm, in);
        setActionLock(in.readInt());
        setEnergy(in.readFloat());
    }
    @Override
    public void changeActionMode(int change)
    {
        int t = actionMode + change;
        t = t % 9;
        if (t < 0)
            t = 9 + t;
        actionMode = (byte)t;
    }
    @Override
    public byte getActionMode()
    {
        return actionMode;
    }
    @Override
    public void setActionMode(byte mode)
    {
        this.actionMode = mode;;
    }
    @Override
    public byte getActionNum()
    {
        return 9;
    }
}