/*
 * Copyright 2005 Takao Sumitomo. All rights reserved.
 * 
 * Redistribution and use in source and binary forms,
 * with or without modification, are permitted provided 
 * that the following conditions are met:
 *
 *   1. Redistributions of source code must retain the 
 *      above copyright notice, this list of conditions
 *      and the following disclaimer.
 *   2. Redistributions in binary form must reproduce 
 *      the above copyright notice, this list of conditions
 *      and the following disclaimer in the documentation
 *      and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * The views and conclusions contained in the software
 * and documentation are those of the authors and should not be
 * interpreted as representing official policies,
 * either expressed or implied.
 * 
 * @Author: Takao Sumitomo
 */
/*
 * 作成日: 2005/01/18
 *
 */
package bakutan.server;

import java.io.DataInputStream;
import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import org.xml.sax.SAXParseException;

import bakutan.common.BPickIntersection;
import bakutan.common.BPickResult;
import bakutan.common.BStage;
import bakutan.common.BUserInterface;
import bakutan.common.Body;
import bakutan.common.BodyState;
import bakutan.common.DataManager;
import bakutan.common.EventHandler;
import bakutan.common.InvalidStructureException;
import bakutan.common.SimpleUserInterface;
import bakutan.common.XMLConverter;
import bakutan.common.event.BodyActionEvent;
import bakutan.common.event.BodyActionLockEnergyEvent;
import bakutan.common.event.BodyAddEvent;
import bakutan.common.event.BodyClientIDEvent;
import bakutan.common.event.BodyCollisionEvent;
import bakutan.common.event.BodyCommandEvent;
import bakutan.common.event.BodyDeleteEvent;
import bakutan.common.event.BodyEvent;
import bakutan.common.event.BodyArmorEvent;
import bakutan.common.event.BodyKnock;
import bakutan.common.event.BodyMessageEvent;
import bakutan.common.event.BodyMoveEvent;
import bakutan.common.event.BodyMultiAddEvent;
import bakutan.common.event.BodyOptionEvent;
import bakutan.common.event.BodyPushEvent;
import bakutan.common.event.BodyScoreChangeEvent;
import bakutan.common.event.BodyScoreEvent;
import bakutan.common.event.BodySetPotisionEvent;
import bakutan.common.event.BodySmallActionEvent;
import bakutan.common.event.BodyStageActionEvent;
import bakutan.common.event.BodyStageMessageEvent;
import bakutan.common.event.BodyStateEvent;
import bakutan.common.event.GameSetEvent;
import bakutan.common.event.StageEvent;
import bakutan.common.math.BMath;
import bakutan.common.math.BVector2f;
import bakutan.common.serverinfo.BStageState;
import bakutan.common.serverinfo.SmallBStageState;
import bakutan.common.util.BakutanClock;
import bakutan.common.util.PacketManager;
import bakutan.server.util.NetworkTool;

/**
 * @author cattaka
 */
public class BakutanServer
{
    public static final int STATE_NEW = 0;
    public static final int STATE_GAME = 1;
    public static final int STATE_GAMESET = 2;

    public static final int CONCURRENT_EVENT_LIMIT = 10000;
    private BStage bstage;
    private HashMap<Body, RemoteClient> remoteClientsMap = new HashMap<Body, RemoteClient>();
    private DataManager dataManager;
    private PacketManager packetManager;
    private int debugLevel = 0;
    private BodyEvent lastBodyEvent;
    private BodyMessageEvent loginMessageEvent;
    private BodyMessageEvent noteMessageEvent;
    private long timelimit = DataManager.DEFAULT_TIMELIMIT;
    private UserDataHolder userDataHolder;
    private boolean checkCheat = false;
    private int state;
    private BUserInterface userInterface;
    
    public BakutanServer(BStageInfo bsi, URL mapFile, DataManager dataManager, UserDataHolder userDataHolder, BUserInterface userInterface) throws IOException
    {
        this.state = STATE_NEW;
        this.dataManager = dataManager;
        this.userDataHolder = userDataHolder;

        this.packetManager = new PacketManager(dataManager);
        this.userInterface = userInterface;
        
        Class bstageClass;
        try
        {
            bstageClass = Class.forName(bsi.getBstageClassName());
        }
        catch (ClassNotFoundException e)
        {
            throw new IOException("unknown BStage class:"+bsi.getBstageClassName());
        }

        if (!(BStage.class.isAssignableFrom(bstageClass)))
        {
            throw new IOException("invalid BStage class");
        }
        BStage bstage;
        try
        {
            bstage = (BStage) bstageClass.newInstance();
        }
        catch (IllegalAccessException e)
        {
            throw new IOException("Could not create BStage");
        }
        catch (InstantiationException e)
        {
            throw new IOException("Could not create BStage");
        }

        EventHandler eventHandler = new ServerEventHandler();
        bstage.setEventHandler(eventHandler);
        bstage.loadBStageFile(mapFile);

        this.bstage = bstage;
//        this.port = bsi.getPort();
//        this.timeout = bsi.getTimeout();
        this.timelimit = bsi.getTimeLimit();
        bstage.setMaxClients(bsi.getMaxClients());
        setLoginMessage(bsi.getLoginMessage());
        setCheckCheat(bsi.isCheckCheat());
    }

    public void startServer()
    {
        // waiting access thread
        systemMessage("Server started");

        //main process
        EventHandler eventHandler = bstage.getEventHandler();

        long currenttime = System.currentTimeMillis();
        long gameSetTime = currenttime + timelimit;
        {
            StageEvent gamesetEvent = new StageEvent(new Body[0], new GameSetEvent(timelimit), currenttime, currenttime);
            eventHandler.getEventThread().putStageEvent(gamesetEvent);
        }
//      long tl = 0;
//        while (tl < timelimit)
//        {
//            StageEvent gamesetEvent = new StageEvent(new Body[0],
//                    new GameSetEvent(tl), currenttime, gameSetTime - tl);
//            tl += 60000;
//            eventHandler.getEventThread().putStageEvent(gamesetEvent);
//        }
        
        HashSet<Body> forCollision = new HashSet<Body>();

        int count = 0;
        //long lastGC = currenttime;
        synchronized (this)
        {
            this.state = STATE_GAME;
        }
        while (true)
        {
            StageEvent stageEvent = bstage.getEventThread().waitForNextEvent();
            if (stageEvent.getGenerationTime() > currenttime)
            {
                currenttime = stageEvent.getGenerationTime();
                bstage.updateAll(currenttime);
                count = 0;
            }
            else
            {
                count++;
                if (count > CONCURRENT_EVENT_LIMIT)
                {
                    systemMessage("Event overflow.");
                    stageEvent = new StageEvent(new Body[0], new GameSetEvent(0), currenttime, currenttime);
                }
            }
            BodyEvent be = stageEvent.getBodyEvent();
            if (4 < this.debugLevel)
                systemMessage(bstage.getCurrentTime() + be.toString(), 4);
            lastBodyEvent = be;
            
            //debug code about gc
            //System.out.println(stageEvent.getBodyEvent());
//            if (currenttime-lastGC > 5000)
//            {
//                StageEvent se = stageEvent;
//                int[] dist = eventHandler.getEventThread().calcEventDistribution();
//                int num = 0;
//                for (int i=0;i<dist.length;i++)
//                {
//                    num += dist[i];
//                    if (dist[i] != 0)
//                        System.out.print(""+i+":"+dist[i] +", ");
//                }
//                System.out.println();
//                System.out.println(eventHandler.getEventThread().getSize()+" : "+num);
//                lastGC = currenttime;
//                System.gc();
//            }

            if (be.getEventID() == DataManager.EVENT_COLLISION)
            {
                BodyCollisionEvent bce = (BodyCollisionEvent) be;
                BPickResult pickResult = bce.getBPickResult();
                BPickIntersection pickIntersection = pickResult.removeFirst();
                Body b1 = stageEvent.getTargets()[0];
                Body b2 = pickIntersection.getBody();
                boolean collidFlag = false;
                
                if (b1.getBstage() == null)
                {
                    pickResult.dispose();
                    continue;
                }
                if (stageEvent.getCreateTime() >= b1.getLastStateChange())
                {
                    if (bstage.isRemoteClientBody(b1.getType()) && b1.isCheckCollidWith(b2))
                    {
                        inputReflectMoveEvent(currenttime, eventHandler, b1, b2);
                    }
                    else
                    {
                        inputBodyCollisionEvent(currenttime, stageEvent.getCreateTime(), bstage, b1, pickResult);
                    }
                    if (b2.getBstage() != null && stageEvent.getCreateTime() >= b2.getLastStateChange())
                    {
                        if (bstage.isRemoteClientBody(b2.getType()) && b2.isCheckCollidWith(b1))
                        {
                            inputReflectMoveEvent(currenttime, eventHandler, b2, b1);
                        }
                            
                        b1.eventCollisition(eventHandler, bstage, b2);
                        b2.eventCollisition(eventHandler, bstage, b1);
                        if (5 <this.debugLevel)
                            systemMessage("" + b1 + b2 + stageEvent.getBodyEvent(), 5);
                    }
                }
//                if (stageEvent.getCreateTime() >= b1.getLastStateChange()
//                        && stageEvent.getCreateTime() >= b2.getLastStateChange()
//                        && b1.getBstage() != null
//                        && b2.getBstage() != null)
//                {
//                    collidFlag = true;
//                    if (false)//(isCheckCheat())
//                    {
//                        Body tg = null;
//                        BVector2f m = null;
//                        BPickResult bpr = bce.getBPickResult();
//                        BPickIntersection bpi = null;
//                        if (bpr.getSize() > 0)
//                            bpi = bce.getBPickResult().removeFirst();
//
//                        if (b1.getType() == bstage.getRemoteClientBodyType()
//                                && bce.getType() == BodyCollisionEvent.COLLISION
//                                && b2.isState(Body.STATE_COLLIDABLE)
//                                && !BMath.isIntersect(b1.getPosition(), b1
//                                        .getCircle(), b2.getPosition(), b2
//                                        .getCircle()))
//                        {
//                            tg = b1;
//                            m = bpi.getNormal2();
//                        }
//                        else if (b2.getType() == bstage
//                                .getRemoteClientBodyType()
//                                && bce.getType() == BodyCollisionEvent.COLLISION
//                                && b1.isState(Body.STATE_COLLIDABLE)
//                                && !BMath.isIntersect(b1.getPosition(), b1
//                                        .getCircle(), b2.getPosition(), b2
//                                        .getCircle()))
//                        {
//                            tg = b2;
//                            m = bpi.getNormal1();
//                        }
//                        if (tg != null)
//                        {
//                            float scale = BVector2f.getDot(m, tg.getVelocity());
//                            m.scale(-scale);
//                            scale = tg.getVelocity().getLength();
//                            m.add(tg.getVelocity());
//                            if (m.getLength() > (scale / 2))
//                            {
//                                m.normalize();
//                                m.scale(scale);
//                            }
//                            else if (m.getLength() < (scale / 1000))
//                            {
//                                m.set(0, 0);
//                            }
//
//                            eventHandler.move(currenttime, currenttime, tg,
//                                    new BVector2f(tg.getPosition()), m,
//                                    new BVector2f(tg.getDirection()));
//                        }
//                    }
//
//                    b1.eventCollisition(eventHandler, bstage, b2);
//                    b2.eventCollisition(eventHandler, bstage, b1);
//                    systemMessage("" + currenttime + stageEvent.getTargets()[0]
//                            + stageEvent.getTargets()[1]
//                            + stageEvent.getBodyEvent(), 5);
//                }
//
//                if (stageEvent.getCreateTime() >= b1.getLastStateChange() && b1.getBstage() != null)
//                {
//                    BPickResult bpr  = bstage.pickIntersection(b1, b2);
//                    inputCollisionEvent(currenttime, bpr, b1, bstage);
//                }
//                if (stageEvent.getCreateTime() >= b2.getLastStateChange() && b2.getBstage() != null)
//                {
//                    BPickResult bpr = bstage.pickIntersection(b2, b1);
//                    inputCollisionEvent(currenttime, bpr, b2, bstage);
//                }
//                if(collidFlag)
//                {
//                    b1.setLastStateChange(currenttime);
//                    b2.setLastStateChange(currenttime);
//                }
                
                continue;
            }
            else if (be.getEventID() == DataManager.EVENT_CLIENTMOVE)
            {
                Body body = stageEvent.getTargets()[0];
                BodyClientMoveEvent bcme = (BodyClientMoveEvent) be;
                BodyMoveEvent bme = bcme.getBodyMoveEvent();

                if (isCheckCheat())
                {
                    RemoteClient rc = bcme.getRemoteClient();

                    if (body.getMaxSpeed() < bme.getVelocity().getLength())
                    {
                        systemMessage(body.getName() + " has doubt of cheating.");
                        bme.getVelocity().normalize();
                        bme.getVelocity().scale(body.getMaxSpeed());
                    }

                    //float d;
                    BVector2f op = new BVector2f();
                    float rag = 0;
                    if (!body.getVelocity().isZeroVector())
                    {
                        BMath.getClosestPoint(op, body.getPosition(), body
                                .getVelocity(), bme.getPosition());

                        BVector2f dir = new BVector2f();
                        dir.set(op);
                        dir.sub(body.getPosition());
                        BVector2f vel = new BVector2f(body.getVelocity());
                        vel.normalize();
                        rag = BVector2f.getDot(dir, vel) / body.getVelocity().getLength();
                    }
                    else
                    {
                        op.set(body.getPosition());
                    }
                    float rag2 = BVector2f.getDistance(op, bme.getPosition())
                            / body.getMaxSpeed();
                    if (rag2 > body.getMaxSpeed() * DataManager.RAG_ATTENUATION)
                        rag += rag2;
                    //System.out.println(rag);

                    float d = (float) (currenttime - rc.lastMove) * 0.001f;
                    rc.lastMove = currenttime;
                    if (d < 0)
                    {
                        d = 0;
                    }
                    else if (d > DataManager.RAG_DELAY
                            / DataManager.RAG_ATTENUATION)
                    {
                        d = DataManager.RAG_DELAY / DataManager.RAG_ATTENUATION;
                    }
                    d *= DataManager.RAG_ATTENUATION;

                    if (rc.rag < 0)
                    {
                        rc.rag += d;
                        if (rc.rag > 0)
                            rc.rag = 0;
                    }
                    else
                    {
                        rc.rag -= d;
                        if (rc.rag < 0)
                            rc.rag = 0;
                    }
                    //System.out.println("M : " + stageEvent.getCreateTime() +
                    // " : " + currenttime + bme.getPosition() +
                    // bme.getVelocity());
                    //System.out.println("" + rag + " : " + rc.rag );

                    rc.rag += rag;

                    if (Math.abs(rc.rag) > DataManager.RAG_DELAY)
                    {
                        //rc.lock = currenttime + (long)(rc.rag * 1000);
                        if (rc.rag > DataManager.RAG_DELAY)
                            rc.rag = DataManager.RAG_DELAY;
                        else if (rc.rag < -DataManager.RAG_DELAY)
                            rc.rag = -DataManager.RAG_DELAY;
                        StageEvent se = new StageEvent(stageEvent.getTargets(),
                                new BodySetPotisionEvent(body.getPosition()),
                                stageEvent.getCreateTime(), stageEvent
                                        .getGenerationTime());
                        //StageEvent se = new
                        // StageEvent(stageEvent.getTargets(), new
                        // BodySetPotisionEvent(bme.getPosition()),
                        // stageEvent.getCreateTime(),
                        // stageEvent.getGenerationTime());
                        eventHandler.getEventThread().putStageEvent(se);
                    }
                    else
                    {
                        StageEvent se = new StageEvent(stageEvent.getTargets(),
                                bme, stageEvent.getCreateTime(), stageEvent
                                        .getGenerationTime());
                        eventHandler.getEventThread().putStageEvent(se);
                    }
                }
                else
                {
                    StageEvent se = new StageEvent(stageEvent.getTargets(),
                            bme, stageEvent.getCreateTime(), stageEvent.getGenerationTime());
                    eventHandler.getEventThread().putStageEvent(se);
                }
            }
            else if (be.getEventID() == DataManager.EVENT_CLIENTACTION)
            {
                Body body = stageEvent.getTargets()[0];
                BodyClientActionEvent bcae = (BodyClientActionEvent) be;
                BodyActionEvent bae = bcae.getBodyActionEvent();
                byte action = bae.getAction();
                if (body.getBstage() == null)
                    continue;
                
                if (isCheckCheat())
                {
                    float dist = BVector2f.getDistance(body.getPosition(), bae
                            .getPosition());
                    if (dist >= body.getMaxSpeed() * DataManager.RAG_ATTENUATION)
                    {
                        continue;
                    }
                }

                if (body.getActionLock() > body.getActionLockTime(action)
                        || body.getEnergy() < -body.getEnergyConsumption(action)
                        || !body.eventAction(eventHandler, bstage, bae.getPosition(), bae))
                {
                    BodyActionLockEnergyEvent balae = new BodyActionLockEnergyEvent(action, body.getEnergyConsumption(action), (short)-body.getActionLockTime(action));
                    StageEvent se = new StageEvent(stageEvent.getTargets(), balae, stageEvent.getCreateTime(), stageEvent.getGenerationTime());
                    bcae.getRemoteClient().sendPacket(packetManager.makePacket(se));
                }
                else
                {
                    body.addActionLockEnergy(-body.getEnergyConsumption(action), body.getActionLockTime(action));
                    StageEvent se = new StageEvent(stageEvent.getTargets(),
                            bae, stageEvent.getCreateTime(), stageEvent
                                    .getGenerationTime());
                    sendAll(packetManager.makePacket(se));
                }
            }
            else if (be.getEventID() == DataManager.EVENT_CLIENTADD)
            {
                BodyClientAddEvent bae = (BodyClientAddEvent) be;
                if (bae.getMode() == BodyClientAddEvent.REQUEST_LOGIN)
                {
                    RemoteClient remoteClient = bae.getRemoteClient();
                    if (getRemoteClientsNum() >= bstage.getMaxClients())
                    {
                        remoteClient.sendPacket(new byte[]{DataManager.AUTH_TOO_MANY_CLIENTS});
                        systemMessage("Too many clients:" + remoteClient.getInetAddress());
                        continue;
                    }
                    else
                    {
                        remoteClient.sendPacket(new byte[]{DataManager.AUTH_SUCCEED});
                    }
                    
                    Body body = bstage.createClientBody();
                    body.setName(remoteClient.getUserInfo().getNickname());
                    body.setID(bstage.getNextID());
                    body.setCurrentTime(currenttime);
                    body.setLastStateChange(currenttime);

                    body.preAdd(bstage);
                    bstage.eventClientPreAdd(currenttime, body);
                    bstage.addBody(body);
                    body.eventAdd(eventHandler, bstage);
                    bstage.eventClientPostAdd(currenttime, body);

                    StageEvent tse = new StageEvent(new Body[0], new BodyAddEvent(body), currenttime, currenttime);
                    sendAll(packetManager.makePacket(tse));
                    remoteClient.setBody(body);
                    putRemoteClient(body, remoteClient);

                    //remoteClient.sendPacket(packetManager.makePacket(body.getID()), false);
                    remoteClient.sendPacket(packetManager.makePacket(bstage), false);
                    remoteClient.sendPacket(packetManager.makePacket(bstage.getBodys()), false);
                    {
                        StageEvent clientIDse = new StageEvent(new Body[0],new BodyClientIDEvent(body.getID()), currenttime, currenttime);
                        remoteClient.sendPacket(packetManager.makePacket(clientIDse));
                    }
                    byte[] stageMessageTypes = bstage.getStageMessageTypes();
                    for (int i = 0; i < stageMessageTypes.length; i++)
                    {
                        String tmp = bstage.getStageMessage(stageMessageTypes[i]);
                        StageEvent se = new StageEvent(new Body[0],new BodyStageMessageEvent(stageMessageTypes[i],tmp), currenttime, currenttime);
                        remoteClient.sendPacket(packetManager.makePacket(se), false);
                    }
                    if (loginMessageEvent != null)
                    {
                        StageEvent se = new StageEvent(new Body[0],loginMessageEvent,currenttime,currenttime);
                        remoteClient.sendPacket(packetManager.makePacket(se), false);
                    }
                    if (noteMessageEvent != null)
                    {
                        StageEvent se = new StageEvent(new Body[0],noteMessageEvent,currenttime,currenttime);
                        remoteClient.sendPacket(packetManager.makePacket(se), false);
                    }
                    systemMessage(body.getName()
                            + "(id="
                            + remoteClient.getAccountID()
                            + ") login from "
                            + remoteClient.getInetAddress());

                    long tmplong = gameSetTime - currenttime;
                    StageEvent gamesetEvent = new StageEvent(new Body[0],new GameSetEvent(tmplong), currenttime, currenttime);
                    //System.out.println("check");
                    remoteClient.sendPacket(packetManager.makePacket(gamesetEvent), true);

                    inputBodyCollisionEvent(currenttime, bstage, body);
                    if (5 <this.debugLevel)
                        systemMessage("" + currenttime + stageEvent.getBodyEvent() + body);
                }
                else if (bae.getMode() == BodyClientAddEvent.REQUEST_STATE)
                {
                    RemoteClient[] rcs = getRemoteClients();
                    BStageState bss = new BStageState();
                    bss.setMaxClients(bstage.getMaxClients());
                    bss.setProtocol(DataManager.getCurrentProtocolVersion());
                    if (loginMessageEvent != null)
                    {
                        bss.setLoginMessage((loginMessageEvent).getMessage());
                    }
                    if (noteMessageEvent != null)
                    {
                        bss.setNoteMessage(noteMessageEvent.getMessage());
                    }
                    for (int i = 0; i < rcs.length; i++)
                    {
                        Body body = rcs[i].getBody();
                        if (body != null)
                            bss.addBodyState(new BodyState(body.getName(), body.getScore(), ""));
                    }
                    RemoteClient rc = bae.getRemoteClient();
                    rc.sendPacket(packetManager.makePacket(bss));

                    systemMessage("BStageState is requested from "
                            + bae.getRemoteClient().getInetAddress());
                }
                else if (bae.getMode() == BodyClientAddEvent.REQUEST_SMALLSTATE)
                {
                    RemoteClient[] rcs = getRemoteClients();
                    SmallBStageState bss = new SmallBStageState();
                    bss.setProtocol(DataManager.getCurrentProtocolVersion());
                    
                    RemoteClient rc = bae.getRemoteClient();
                    rc.sendPacket(packetManager.makePacket(bss));

                    systemMessage("SmallBStageState is requested from "
                            + bae.getRemoteClient().getInetAddress());
                }
                continue;
            }
            else if (be.getEventID() == DataManager.EVENT_ADD)
            {
                BodyAddEvent bae = (BodyAddEvent) be;
                Body body = (Body) bae.getBody();
                body.setID(bstage.getNextID());
                body.setCurrentTime(currenttime);
                body.setLastStateChange(currenttime);

                body.preAdd(bstage);
                bstage.addBody(body);
                body.eventAdd(eventHandler, bstage);

                //do not move this line to first of block
                sendAll(packetManager.makePacket(stageEvent));

                inputBodyCollisionEvent(currenttime, bstage, body);
                if (5 <this.debugLevel)
                    systemMessage("" + currenttime + stageEvent.getBodyEvent() + body + "\n");
            }
            else if (be.getEventID() == DataManager.EVENT_MULTIADD)
            {
                BodyMultiAddEvent bmae = (BodyMultiAddEvent) be;
                Body[] bodys = bmae.getBodys();
                for (int i=0;i<bodys.length;i++)
                {
                    Body body = bodys[i];
                    body.setID(bstage.getNextID());
                    //System.out.println(""+body+body.getID());
                    body.setCurrentTime(currenttime);
                    body.setLastStateChange(currenttime);

                    body.preAdd(bstage);
                    bstage.addBody(body);
                    body.eventAdd(eventHandler, bstage);
                    
                    inputBodyCollisionEvent(currenttime, bstage, body);
                    
                    if (5 <this.debugLevel)
                        systemMessage("" + currenttime + stageEvent.getBodyEvent() + body);
                }
                //do not move this line to first of block
                sendAll(packetManager.makePacket(stageEvent));
            }
            else if (be.getEventID() == DataManager.EVENT_DELETE)
            {
                BodyDeleteEvent bde = (BodyDeleteEvent) be;
                Body body = stageEvent.getTargets()[0];
                if (body.getBstage() == null)
                    continue;
                sendAll(packetManager.makePacket(stageEvent));

                //body.setLastStateChange(currenttime);
                if (!(bstage.removeBody(body)))
                {
                    systemMessage("error at BodyDeleteEvent:Couldn't delete body.");
                }
                body.eventRemove(eventHandler, bstage);
                //System.out.println(eventHandler.getEventThread().getLength());
                bstage.getEventThread().deleteEventAbout(body);
//                RemoteClient rc = (RemoteClient) getRemoteClient(body);
//                if (rc != null)
//                {
//                    removeRemoteClient(body);
//                    systemMessage(body.getName() + "(id=" + rc.getAccountID() + ") logouted. Score " + body.getScore());
//                }
            }
            else if (be.getEventID() == DataManager.EVENT_CLIENTREMOVE)
            {
                BodyClientRemoveEvent bcre = (BodyClientRemoveEvent)be;
                RemoteClient rc = bcre.getRemoteClient();
                Body body = rc.getBody();
                if (body != null)
                {
                    removeRemoteClient(body);
                    systemMessage(body.getName() + "(id=" + rc.getAccountID() + ") logout. Score " + body.getScore());
                    eventHandler.delete(currenttime, currenttime, body);
                }
            }
            else if (be.getEventID() == DataManager.EVENT_MOVE)
            {
                BodyMoveEvent bme = (BodyMoveEvent) be;
                Body body = stageEvent.getTargets()[0];
                if (body.getBstage() == null)
                    continue;
                sendAll(packetManager.makePacket(stageEvent));

                body.setVelocity(bme.getVelocity());
                body.setDirection(bme.getDirection());
                body.setCurrentTime(currenttime);
                body.setLastStateChange(currenttime);

                forCollision.clear();
                bstage.pickIntersection(body, forCollision);
                body.setPosition(bme.getPosition());
                inputBodyCollisionEvent(currenttime, bstage, body, forCollision);
            }
            else if (be.getEventID() == DataManager.EVENT_SETPOTISION)
            {
                BodySetPotisionEvent bse = (BodySetPotisionEvent) be;
                Body body = stageEvent.getTargets()[0];
                if (body.getBstage() == null)
                    continue;
                sendAll(packetManager.makePacket(stageEvent));

                body.setCurrentTime(currenttime);
                body.setLastStateChange(currenttime);
                body.setVelocity(new BVector2f(0, 0));
                body.setPosition(bse.getPosition());

                forCollision.clear();
                bstage.pickIntersection(body, forCollision);
                inputBodyCollisionEvent(currenttime, bstage, body, forCollision);
            }
            else if (be.getEventID() == DataManager.EVENT_RECHECKINTERSECT)
            {
                Body body = stageEvent.getTargets()[0];
                forCollision.clear();
                bstage.pickIntersection(body, forCollision);
                Iterator<Body> ite = forCollision.iterator();
                while (ite.hasNext())
                {
                    Body b = ite.next();
                    body.eventCollisition(eventHandler, bstage, b);
                    b.eventCollisition(eventHandler, bstage, body);
                }
            }
            else if (be.getEventID() == DataManager.EVENT_ACTION
                    || be.getEventID() == DataManager.EVENT_SMALLACTION)
            {
                Body body = stageEvent.getTargets()[0];
                if (body.getBstage() == null)
                    continue;
                BodyActionEvent bae;
                if (be.getEventID() == DataManager.EVENT_ACTION)
                    bae = (BodyActionEvent) be;
                else
                {
                    BodySmallActionEvent bsae = (BodySmallActionEvent)be;
                    bae = new BodyActionEvent(body.getPosition(),body.getPosition(),bsae.getAction());
                }
                    
                if (body.eventAction(eventHandler, bstage, bae.getPosition(), bae))
                {
                    byte action = bae.getAction();
                    body.addActionLockEnergy(-body.getEnergyConsumption(action), body.getActionLockTime(action));
                    sendAll(packetManager.makePacket(stageEvent));
                }
            }
            else if (be.getEventID() == DataManager.EVENT_ARMOR)
            {
                BodyArmorEvent bde = (BodyArmorEvent) be;
                Body to = stageEvent.getTargets()[0];
                Body from = stageEvent.getTargets()[1];
                if (from.getBstage() == null || to.getBstage() == null)
                    continue;

                if (to.hasArmorReport())
                {
                    sendAll(packetManager.makePacket(stageEvent));
                }
                if (bde.getType() == BodyArmorEvent.SUB)
                {
                    to.setArmor(to.getArmor() - bde.getArmor());
                }
                else if (bde.getType() == BodyArmorEvent.ADD)
                {
                    to.setArmor(to.getArmor() + bde.getArmor());
                }
                else if (bde.getType() == BodyArmorEvent.SET)
                {
                    to.setArmor(bde.getArmor());
                }
                to.eventArmorChange(eventHandler, bstage, from, bde.getType(), bde.getArmor());
                from.eventChangeOthersArmor(eventHandler, bstage, to, bde.getType(), bde.getArmor());
                if (to.isState(Body.STATE_LIFE))
                {
                    if (!(to.isState(Body.STATE_DEAD)) && to.getArmor() <= 0)
                    {
                        BodyStateEvent bse = new BodyStateEvent(Body.ENABLE, Body.STATE_DEAD);
                        StageEvent sse = new StageEvent(new Body[]{to}, bse, currenttime, currenttime);
                        bstage.getEventThread().putStageEvent(sse);
                    }
                }
            }
            else if (be.getEventID() == DataManager.EVENT_STATE)
            {
                BodyStateEvent bse = (BodyStateEvent) be;
                Body tg = stageEvent.getTargets()[0];
                if (tg.getBstage() == null)
                    continue;
                sendAll(packetManager.makePacket(stageEvent));
                switch (bse.getType())
                {
                    case Body.SET :
                        if (tg.isState(bse.getState()))
                            continue;
                        tg.setState(bse.getState());
                        break;
                    case Body.ENABLE :
                        if (tg.isState(bse.getState()))
                            continue;
                        tg.enableState(bse.getState());
                        break;
                    case Body.DISABLE :
                        if (!(tg.isState(bse.getState())))
                            continue;
                        tg.disableState(bse.getState());
                        break;
                }
//                if ((bse.getType() == Body.SET || bse.getType() == Body.ENABLE)
//                        && (bse.getState() & Body.DEAD) == Body.DEAD)
//                {
//                    if (tg.getType() == bstage.getRemoteClientBodyType())
//                    {
//                        systemMessage(tg.getName() + "(id=" + tg.getID()
//                                + ") dead.", 0);
//                    }
//                }
                tg.eventStateChange(eventHandler, bstage, bse.getType(), bse.getState());
            }
            else if (be.getEventID() == DataManager.EVENT_OPTION)
            {
                BodyOptionEvent boe = (BodyOptionEvent) be;
                Body tg = stageEvent.getTargets()[0];
                if (tg.getBstage() == null)
                    continue;
                sendAll(packetManager.makePacket(stageEvent));
                tg.setOption(boe.getType(), boe.getValue());
                tg.eventOptionChange(eventHandler, bstage, boe.getType(), boe
                        .getValue());
            }
            else if (be.getEventID() == DataManager.EVENT_CHAT)
            {
                BodyCommandEvent bce = (BodyCommandEvent) be;
                Body body = stageEvent.getTargets()[0];
                if (body.getBstage() == null)
                    continue;
                String tmp = bce.getCommand();
                body.eventCommand(eventHandler, bstage, tmp);
            }
            else if (be.getEventID() == DataManager.EVENT_MESSAGE)
            {
                for (int i = 0; i < stageEvent.getTargets().length; i++)
                {
                    if (stageEvent.getTargets()[i].getBstage() == null)
                    {
                        continue;
                    }
                }
                sendAll(packetManager.makePacket(stageEvent));
                BodyMessageEvent bme = (BodyMessageEvent) be;
                if (bme.getType() == DataManager.MESSAGE_LOGIN)
                {
                    this.loginMessageEvent = bme;
                }
                else if (bme.getType() == DataManager.MESSAGE_NOTE)
                {
                    StageEvent nme = new StageEvent(new Body[0], bme,
                            currenttime, currenttime);
                    this.noteMessageEvent = bme;
                    systemMessage(bme.getMessage());
                }
                else if (bme.getType() == DataManager.MESSAGE_CHAT)
                {
                    Body body = stageEvent.getTargets()[0];
                    systemMessage(body.getName() + "(id=" + body.getID() + "):"
                            + bme.getMessage());
                }
            }
            else if (be.getEventID() == DataManager.EVENT_STAGEMESSAGE)
            {
                sendAll(packetManager.makePacket(stageEvent));
                BodyStageMessageEvent bme = (BodyStageMessageEvent) be;

                bstage.setStageMessage(bme.getType(), bme.getMessage());
            }
            else if (be.getEventID() == DataManager.EVENT_SCORE)
            {
                BodyScoreEvent bse = (BodyScoreEvent) be;
                for (int i = 0; i < bse.getScores().length; i++)
                {
                    stageEvent.getTargets()[i].setScore(bse.getScores()[i]);
                }
                sendAll(packetManager.makePacket(stageEvent));
            }
            else if (be.getEventID() == DataManager.EVENT_SCORECHANGE)
            {
                BodyScoreChangeEvent bsce = (BodyScoreChangeEvent) be;
                Body[] bs = stageEvent.getTargets();
                int[] scoreChanges = bsce.getScores();
                int[] scores = new int[bsce.getScores().length];
                Body from = bs[0];
                Body[] bodys = new Body[bs.length - 1];
                System.arraycopy(bs,1,bodys,0,bodys.length);
                RemoteClient fromRC = getRemoteClient(from);
                for (int i = 0; i < bsce.getScores().length; i++)
                {
                    Body body = bodys[i];
                    body.setScore(body.getScore() + scoreChanges[i]);
                    scores[i] = body.getScore();
                    if (fromRC != null)
                    {
                        RemoteClient rc = getRemoteClient(body);
                        if (rc != null)
                        {
                            rc.putScore(fromRC.getAccountID(), scoreChanges[i]);
                        }
                    }
                }
                StageEvent se = new StageEvent(bodys, new BodyScoreEvent(scores),currenttime,currenttime);
                sendAll(packetManager.makePacket(se));
            }
            else if (be.getEventID() == DataManager.EVENT_SOUND)
            {
                sendAll(packetManager.makePacket(stageEvent));
            }
            else if (be.getEventID() == DataManager.EVENT_PUSH)
            {
                BodyPushEvent bpe = (BodyPushEvent) be;
                Body body = stageEvent.getTargets()[0];
                Body target = bstage.getBody(bpe.getTargetID());
                if (target != null)
                {
                    body.eventPush(eventHandler, bstage, target);
                }
            }
            else if (be.getEventID() == DataManager.EVENT_STAGEACTION)
            {
                BodyStageActionEvent bae = (BodyStageActionEvent) be;
                bstage.eventAction(bae.getAction());
            }
            else if (be.getEventID() == DataManager.EVENT_KNOCK)
            {
                Body body = stageEvent.getTargets()[0];
                StageEvent sse = new StageEvent(new Body[0], new BodyKnock(), currenttime, currenttime);
                getRemoteClient(body).sendPacket(packetManager.makePacket(sse));
            }
            else if (be.getEventID() == DataManager.EVENT_GAMESET)
            {
                sendAll(packetManager.makePacket(stageEvent));
                GameSetEvent gameSetEvent = (GameSetEvent) be;
                if (gameSetEvent.getTimeLimit() <= 0)
                {
                    systemMessage("Gameset.");
                    RemoteClient[] rcs = getRemoteClients();
                    for (int i = 0; i < rcs.length; i++)
                    {
                        //rcs[i].removeBStage();
                        systemMessage(rcs[i].getBody().getName() + "(id=" + rcs[i].getBody().getID()
                                + ") Score " + rcs[i].getBody().getScore());
                    }
                    synchronized (this)
                    {
                        this.state = STATE_GAMESET;
                    }
                    return;
                }
                else
                {
                    gameSetTime = currenttime + gameSetEvent.getTimeLimit();
                    long w = gameSetEvent.getTimeLimit();
                    long gst = 0;
                    if (w > 6000)
                    {
                        gst = w - 6000;
                        w = 6000;
                    }
                    GameSetEvent gse = new GameSetEvent(gst);
                    StageEvent se = new StageEvent(new Body[0], gse, currenttime, currenttime + w);
                    eventHandler.getEventThread().putStageEvent(se);
                }
            }
            else
            {
                userInterface.outputMessage("Unknown Event" + be);
            }
            if (debugLevel >= 5)
            {
                System.out.print("" + currenttime + stageEvent.getBodyEvent());
                for (int i = 0; i < stageEvent.getTargets().length; i++)
                {
                    System.out.print(stageEvent.getTargets()[i]);
                }
                System.out.print("\n");
            }
        }
    }
    private void inputReflectMoveEvent(long currenttime, EventHandler eh, Body b1, Body b2)
    {
        BVector2f d1 = new BVector2f();
        BVector2f d2 = new BVector2f();
        if (BMath.calcNormal(d1,d2,b1.getPosition(), b1.getShape(), b2.getPosition(), b2.getShape()))
        {
            BVector2f vel = new BVector2f(b1.getVelocity());
            float l = -BVector2f.getDot(vel, d1);
            vel.scaleAdd(l, d1);
            //if (!vel.equals(b1.getVelocity()))
            eh.move(currenttime,currenttime,b1,b1.getPosition(), vel,b1.getDirection());
            //System.out.println(""+vel.equals(b1.getVelocity())+":" + vel.getLength());
        }
    }
    private void inputBodyCollisionEvent(long currenttime, BStage stage, Body target)
    {
        BPickResult pickResult = bstage.estimateIntersection(target);
        while (pickResult.getSize() > 0)
        {
            BPickIntersection bpi = pickResult.getFirst();
            BodyCollisionEvent bci = new BodyCollisionEvent(pickResult);
            long t = currenttime + (long) (bpi.getDistance() * 1000f);
            StageEvent se = new StageEvent(new Body[]{target}, bci, currenttime, t);
            bstage.getEventThread().putStageEvent(se);
            break;
        }
    }
    private void inputBodyCollisionEvent(long currenttime, BStage stage, Body target, Set<Body> ignore)
    {
        BPickResult pickResult = bstage.estimateIntersection(target, ignore);
        while (pickResult.getSize() > 0)
        {
            BPickIntersection bpi = pickResult.getFirst();
            BodyCollisionEvent bci = new BodyCollisionEvent(pickResult);
            long t = currenttime + (long) (bpi.getDistance() * 1000f);
            StageEvent se = new StageEvent(new Body[]{target}, bci, currenttime, t);
            bstage.getEventThread().putStageEvent(se);
            break;
        }
    }
    private void inputBodyCollisionEvent(long currenttime, long createtime, BStage stage, Body target, BPickResult pickResult)
    {
        while(pickResult.getSize() > 0)
        {
              BPickIntersection bpi = pickResult.getFirst();
              long t = createtime + (long) (bpi.getDistance() * 1000f);
              Body body = bpi.getBody();
              if (body.getBstage() != null && body.getLastStateChange() <= t)
              {
                  BodyCollisionEvent bci = new BodyCollisionEvent(pickResult);
                  StageEvent se = new StageEvent(new Body[]{target}, bci, createtime, t);
                  bstage.getEventThread().putStageEvent(se);
                  return;
              }
              pickResult.removeFirst();
        }
    }

    public void putStageEvent(StageEvent be)
    {
        bstage.getEventThread().putStageEvent(be);
    }
    public void putGameSetEvent()
    {
        long currenttime = BakutanClock.getCurrenttime();
        StageEvent gamesetEvent = new StageEvent(new Body[0],new GameSetEvent(0), currenttime, currenttime);
        putStageEvent(gamesetEvent);
    }
    
    private void putRemoteClient(Body body, RemoteClient rc)
    {
        remoteClientsMap.put(body, rc);
    }

    private void removeRemoteClient(Body body)
    {
        remoteClientsMap.remove(body);
    }

    private RemoteClient getRemoteClient(Body body)
    {
        return remoteClientsMap.get(body);
    }

    public int getRemoteClientsNum()
    {
        return remoteClientsMap.size();
    }

    public RemoteClient[] getRemoteClients()
    {
        RemoteClient[] rss = new RemoteClient[remoteClientsMap.values().size()];
        System.arraycopy(remoteClientsMap.values().toArray(), 0, rss, 0, rss.length);
        return rss;
    }

    private StageEvent createBodyScoreEvent(long createTime, long generationTime)
    {
        Body[] bs = bstage.getBodysHasScore();
        int[] scores = new int[bs.length];
        for (int i = 0; i < bs.length; i++)
        {
            scores[i] = bs[i].getScore();
        }
        return new StageEvent(bs, new BodyScoreEvent(scores), createTime,
                generationTime);
        /*
         * AbstractRemoteClient[] rcs = getRemoteClients(); Body[] bs = new
         * Body[rcs.length]; int[] scores = new int[rcs.length]; for (int i=0;i
         * <rcs.length;i++) { bs[i] = rcs[i].getBody(); scores[i] =
         * rcs[i].getScore(); } return new StageEvent(bs,new
         * BodyScoreEvent(scores),createTime,generationTime);
         */
    }

    private void sendAll(byte[] packet)
    {
        Iterator<RemoteClient> ite = remoteClientsMap.values().iterator();
        while (ite.hasNext())
        {
            ite.next().sendPacket(packet);
        }
    }

    public void systemMessage(String message, int debugLevel)
    {
        if (debugLevel <= this.debugLevel)
        {
            //System.out.println(port + ":" + dateFormat.format(new Date()) + " : " + message);
            userInterface.outputMessage(message);
        }
    }
    public void systemMessage(String message)
    {
        //System.out.println(port + ":" + dateFormat.format(new Date()) + " : " + message);
        userInterface.outputMessage(message);
    }

    public void setLoginMessage(String message)
    {
        if (message == null)
            throw new NullPointerException();
        if (message.equals(""))
        {
            this.loginMessageEvent = null;
            return;
        }
        long currenttime = System.currentTimeMillis();
        BodyMessageEvent bme = new BodyMessageEvent(DataManager.MESSAGE_LOGIN,
                message);
        StageEvent se = new StageEvent(new Body[0], bme, currenttime,
                currenttime);
        bstage.getEventThread().putStageEvent(se);
    }

    public static void main(String[] args)
    {
        try
        {
            SimpleUserInterface userInterface = new SimpleUserInterface();
            userInterface.setOutputDate(true);
            File sc = new File("ServerConfigurationInfo.xml");
            if (!(sc.exists()))
            {
                throw new IOException(
                        "ServerConigurationInfo.xml is not found.");
            }
            URL url = sc.toURL();
            ServerConfigurationInfo sci = ServerXMLConverter
                    .createConfigurationInfo(XMLConverter.readXML(url));
            //BStageInfo[] bsis = sci.getBStageInfos();
            ServerProcess serverProcesses;
            BodyMessageEvent noteMessageEvent = null;
            RemoteClient[] remoteClients = new RemoteClient[0];
            
            UserDataHolder userDataHolder;
            if (sci.getDataBaseInfo() != null)
            {
                userDataHolder = new DBUserDataHolder(sci.getDataBaseInfo());
                if (!userDataHolder.initialize())
                {
                    userDataHolder = new NullUserDataHolder();
                }
            }
            else
            {
                userDataHolder = new NullUserDataHolder();
            }
//            if (bsis.length == 0)
//            {
//                userInterface.outputMessage("No BStage info");
//                return;
//            }

            NetworkTool networkTool = NetworkTool.getAvailableNetworkTool();
            BStageInfo bsi = sci.getBStageInfo();
            serverProcesses = new ServerProcess(userInterface, networkTool);
            serverProcesses.init(bsi, userDataHolder);
            serverProcesses.setRegisterServer(sci.isRegisterServer());
            serverProcesses.setUseUPNP(sci.isUseUPNP());
            serverProcesses.setServerInfoManagerAddress(sci.getServerInfoManagerAddress());
            serverProcesses.run();
        }
        catch (InvalidStructureException e)
        {
            e.printStackTrace();
        }
        catch (SAXParseException e)
        {
            e.printStackTrace();
        }
        catch (IOException e)
        {
            e.printStackTrace();
        }
    }

    public boolean isCheckCheat()
    {
        return checkCheat;
    }

    public void setCheckCheat(boolean checkCheat)
    {
        this.checkCheat = checkCheat;
    }

    public void joinRemoteClient(RemoteClient rc)
    {
        rc.joinGame(this);
    }
    public String getNoteMessage()
    {
        if (noteMessageEvent != null)
            return noteMessageEvent.getMessage();
        return null;
    }
    public void setNoteMessage(String message)
    {
        this.noteMessageEvent = new BodyMessageEvent(DataManager.MESSAGE_NOTE, message);
    }
    public BodyEvent inputBodyEvent(DataInputStream in) throws IOException
    {
        return dataManager.inputBodyEvent(in, bstage);
    }
    public int getServerState()
    {
        synchronized (this)
        {
            return this.state;
        }
    }
}


